/*
 * 
 * Created on Jan 17, 2006
 *
 */
package animatlab.kodamat.core;

import java.util.List;
import java.util.Map;
import java.util.Set;

import animatlab.kodamat.core.utils.Variable;

/**
 * 
 * Model of a MDP problem
 * 
 * @author Thomas Degris
 * 
 */
public interface IModel {
	/**
	 * 
	 * @return the name of this model
	 * 
	 */
	public String getName();

	/**
	 * 
	 * @return size of the transition model
	 * 
	 */
	public long getTransitionsModelSize();

	/**
	 * 
	 * @return size of the reward function
	 * 
	 */
	public int getRewardModelSize();

	/**
	 * 
	 * @return the set of actions registered as available for the agent
	 * 
	 */
	public List<String> getActionList();

	/**
	 * 
	 * @return the set of variables available for the agent
	 * 
	 */
	public Set<Variable> getVariableSet();

	/**
	 * 
	 * @return the list of value set for each variable
	 * 
	 */
	public Map<Variable, Set<String>> getValueSets();

	/**
	 * 
	 * @return get the reward maximum that the agent can get/has get
	 * 
	 */
	public double getRewardMax();

	/**
	 * 
	 * The set of value of a variable
	 * 
	 * @param variable
	 *          the variable to get the set of value from
	 * 
	 * @return the set of value of variable
	 * 
	 */
	public Set<String> getValueSet(Variable variable);

	/**
	 * 
	 * @return the size of the state of the agent
	 * 
	 */
	public int getStateSize();
}
