/*
 * 
 * Created on 11 mai 2005
 *
 */
package animatlab.kodamat.core;

import java.util.List;
import java.util.Random;

import animatlab.kodamat.core.utils.Parameters;
import animatlab.kodamat.core.utils.Context;
import animatlab.kodamat.core.utils.VariableOrdering;

/**
 * 
 * Interface to represent a world model
 * 
 * @author Thomas Degris
 * 
 */
public interface IProblemDefinition extends IModel {
	/**
	 * 
	 * @return the legend used in this forest
	 * 
	 */
	public VariableOrdering getVariableOrdering();

	/**
	 * 
	 * Compute the next state from the current state and an action
	 * 
	 * @param random
	 *          the random instance to use to build the next state
	 * @param currentState
	 *          the current state of the agent
	 * @param nextState
	 *          the next state filled by the method
	 * @param action
	 *          the action executed by the agent
	 * 
	 * @return the current reward of the agent
	 * 
	 */
	List<Double> getNextState(Random random, Context currentState,
			Context nextState, String action);

	/**
	 * Build an initial state in the problem
	 * 
	 * @param random
	 *          the random instance to use to build the initial state
	 * 
	 * @return an initial state to start from
	 * 
	 */
	Context getInitialState(Random random);

	/**
	 * 
	 * Test if a state is terminal (only if some conditions have been defined as
	 * terminal)
	 * 
	 * @param state
	 *          the state to test
	 * 
	 * @return true if the state is terminal
	 * 
	 */
	boolean isTerminal(Context state);

	/**
	 * 
	 * @return the parameters of this forest
	 * 
	 */
	public Parameters getParameters();

	/**
	 * 
	 * @return the number of possible state
	 * 
	 */
	long getNumberOfStates();

	/**
	 * 
	 * @return true when the problem is stochastic
	 * 
	 */
	boolean isStochastic();
}
