package animatlab.kodamat.core.adds;

import animatlab.kodamat.core.utils.Context;
import animatlab.kodamat.core.utils.Variable;

/**
 * 
 * Decision node in ADDs
 * 
 * @author Thomas Degris, Tomas Simovic
 * @param <T>
 *          type of the data contained at the leaf
 * 
 */
public class DNode<T> extends ADDNode<T> {
	/**
	 * 
	 * Wildcard that can be used for values
	 * 
	 */
	static public final String VALUE_WILDCARD = "*";

	/**
	 * 
	 * Tested variable
	 * 
	 */
	Variable variable;

	/**
	 * 
	 * Right node of this decision node
	 * 
	 */
	private ADDNode<T> rightNode;

	/**
	 * 
	 * Left node of this decision node
	 * 
	 */
	private ADDNode<T> leftNode;

	/**
	 * 
	 * Constructor
	 * 
	 * @param add
	 *          ADD owning this node
	 * @param variable
	 *          variable tested in this decision node
	 * 
	 */
	public DNode(ADD<T> add, Variable variable) {
		super(add, false);
		this.variable = variable;
	}

	/**
	 * 
	 * Set the left side of this decision node
	 * 
	 * @param node
	 *          destination node
	 * 
	 */
	public void setLeft(ADDNode<T> node) {
		this.leftNode = node;
	}

	/**
	 * 
	 * Set the right side of this decision node
	 * 
	 * @param node
	 *          destination node
	 * 
	 */
	public void setRight(ADDNode<T> node) {
		this.rightNode = node;
	}

	/**
	 * 
	 * Get the left side of this decision node
	 * 
	 * @return leftNode
	 * 
	 */
	public ADDNode<T> getLeft() {
		return leftNode;
	}

	/**
	 * 
	 * Get the right side of this decision node
	 * 
	 * @return leftNode
	 * 
	 */
	public ADDNode<T> getRight() {
		return rightNode;
	}

	/**
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		StringBuilder result = new StringBuilder();
		result.append(childToString(leftNode));
		result.append(" < ");
		result.append(add.indexOf(this));
		result.append(" v:");
		result.append(variable.toString());
		result.append(" > ");
		result.append(childToString(rightNode));
		return result.toString();
	}

	/**
	 * 
	 * Print a child into the result buffer
	 * 
	 * @param childNode
	 *          the child to print
	 * 
	 * @return a string representation of the child
	 * 
	 */
	@SuppressWarnings("unchecked")
	private String childToString(ADDNode<T> childNode) {
		if (childNode == null) {
			return "?";
		}
		if (childNode.terminal) {
			return childNode.toString();
		}
		DNode<T> castedChildNode = (DNode<T>) childNode;
		return String.valueOf(add.indexOf(castedChildNode));
	}

	/**
	 * 
	 * @return the variable tested in this decision node
	 * 
	 */
	public Variable getVariableTested() {
		return variable;
	}

	/**
	 * 
	 * @return the left node if this decision node
	 * 
	 */
	public ADDNode<T> left() {
		return leftNode;
	}

	/**
	 * 
	 * @return the left node if this decision node
	 * 
	 */
	public ADDNode<T> right() {
		return rightNode;
	}

	/**
	 * 
	 * Get a child from an array of value
	 * 
	 * @param values
	 *          the set of values for variable
	 * 
	 * @return the child
	 * 
	 */
	final public ADDNode<T> getChild(Context values) {
		return getChild(values.get(variable));
	}

	/**
	 * 
	 * Get a child from the value of the variable
	 * 
	 * @param string
	 *          the value of the variable
	 * 
	 * @return the child
	 * 
	 */
	public ADDNode<T> getChild(String string) {
		if (string.equals("0")) {
			return rightNode;
		}
		return leftNode;
	}

}
