/**
 * 
 **/
package animatlab.kodamat.core.conditions;

import java.util.Random;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import animatlab.kodamat.core.io.IToXml;
import animatlab.kodamat.core.utils.Context;

/**
 * Interface to represent a tree of conditions on state
 * 
 * @author Thomas Degris
 * 
 */
public interface IStateCondition extends IToXml<Object> {
	/**
	 * 
	 * Node for a condition operator
	 * 
	 */
	static final String NAMESPACE = "cnd";

	/**
	 * Test if this condition is true for a state
	 * 
	 * @param state
	 *          the state to test
	 * 
	 * @return true if the conditions are respected
	 */
	boolean isTrue(Context state);

	/**
	 * Set the state to make this condition true
	 * 
	 * @param random
	 *          randomizer used to generate a random true condition
	 * @param state
	 *          the state to set
	 * 
	 */
	void setToTrue(Random random, Context state);

	/**
	 * @see animatlab.kodamat.core.io.IToXml#toXml(org.w3c.dom.Document,
	 *      animatlab.kodamat.core.io.IToXml.IToXmlPolicy)
	 */
	public Element toXml(Document document, IToXmlPolicy<Object> policy);

}
