/*
 * 
 * Created on 16 aout 2005
 *
 */
package animatlab.kodamat.core.conditions;

import java.util.ArrayList;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

import animatlab.kodamat.core.utils.Context;

/**
 * 
 * Condition using an operator
 * 
 * @author Thomas Degris
 * 
 */
public abstract class OperatorCondition implements IStateCondition {
	/**
	 * 
	 * Node for a condition operator
	 * 
	 */
	static public final String NODE_NAME = NAMESPACE + ":" + "cond";

	/**
	 * 
	 * Node for a condition operator
	 * 
	 */
	static public final String ATTRIBUTE_OPERATOR_NAME = "operator";

	/**
	 * 
	 * Xml name of this operator
	 * 
	 */
	private final String operatorName;

	/**
	 * 
	 * List of arguments for the or operator
	 * 
	 */
	protected final ArrayList<IStateCondition> argSet = new ArrayList<IStateCondition>();

	/**
	 * 
	 * Constructor
	 * 
	 * @param xmlNode
	 *          the xml node to build the condition from
	 * @param operatorName
	 *          xml name of this operator
	 * 
	 */
	public OperatorCondition(Node xmlNode, String operatorName) {
		this(operatorName);
		int index = 0;
		while (index < xmlNode.getChildNodes().getLength()) {
			Node currentNode = xmlNode.getChildNodes().item(index);
			if (currentNode.getNodeType() == Node.ELEMENT_NODE) {
				addArg(VariableCondition.load(currentNode));
			}
			index++;
		}
	}

	/**
	 * 
	 * Add an argument to the operatir
	 * 
	 * @param arg
	 *          the argument to add
	 * 
	 */
	public void addArg(IStateCondition arg) {
		argSet.add(arg);
	}

	/**
	 * 
	 * Constructor
	 * 
	 * @param operatorName
	 *          xml name of this operator
	 * 
	 */
	public OperatorCondition(String operatorName) {
		this.operatorName = operatorName;
	}

	/**
	 * @see animatlab.kodamat.core.conditions.IStateCondition#isTrue(Context)
	 */
	public boolean isTrue(Context state) {
		boolean result = false;
		int index = 0;
		for (IStateCondition condition : argSet) {
			if (index == 0) {
				result = condition.isTrue(state);
			} else {
				result = applyOperator(result, condition.isTrue(state));
			}
			index++;
		}
		return result;
	}

	/**
	 * @see animatlab.kodamat.core.io.IToXml#toXml(org.w3c.dom.Document,
	 *      animatlab.kodamat.core.io.IToXml.IToXmlPolicy)
	 */
	public Element toXml(Document document, IToXmlPolicy<Object> policy) {
		Element operatorNode = document.createElement(NODE_NAME);
		operatorNode.setAttribute(ATTRIBUTE_OPERATOR_NAME, operatorName);
		for (IStateCondition condition : argSet) {
			operatorNode.appendChild(condition.toXml(document, null));
		}
		return operatorNode;
	}

	/**
	 * 
	 * Apply this operator on two arguments
	 * 
	 * @param arg01
	 *          the first argument to apply this operator to
	 * @param arg02
	 *          the second argument to apply this operator to
	 * 
	 * @return the result of the boolean operator
	 * 
	 */
	abstract protected boolean applyOperator(boolean arg01, boolean arg02);

	/**
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		StringBuffer result = new StringBuffer("(");
		int index = 0;
		for (IStateCondition condition : argSet) {
			result.append(condition.toString());
			if (index + 1 < argSet.size()) {
				result.append(operatorName);
			}
			index++;
		}
		result.append(")");
		return result.toString();
	}
}
