/*
 * 
 * Created on 16 aout 2005
 *
 */
package animatlab.kodamat.core.conditions.test;

import java.util.LinkedHashSet;
import java.util.Random;
import java.util.Set;

import junit.framework.TestCase;

import org.w3c.dom.Element;

import animatlab.kodamat.core.conditions.IStateCondition;
import animatlab.kodamat.core.conditions.VariableCondition;
import animatlab.kodamat.core.io.XmlUtils;
import animatlab.kodamat.core.utils.Context;
import animatlab.kodamat.core.utils.Variable;

/**
 * 
 * Test loading/saving functionalities for the Forest
 * 
 * @author Thomas Degris
 * 
 */
public class TestStateCondition extends TestCase {
	/**
	 * 
	 * Randomizer for generating examples
	 * 
	 */
	private Random random = new Random(0);

	/**
	 * 
	 * Size of the state tested
	 * 
	 */
	final static private int STATE_SIZE = 10;

	/**
	 * 
	 * Variable used in the condition
	 * 
	 */
	final static Variable TESTED_VARIABLE_01 = Variable.getVariable("2");

	/**
	 * 
	 * Variable used in the condition
	 * 
	 */
	final static Variable TESTED_VARIABLE_02 = Variable.getVariable("5");

	/**
	 * 
	 * Variable used in the condition
	 * 
	 */
	final static Variable TESTED_VARIABLE_03 = Variable.getVariable("1");

	/**
	 * 
	 * Variable used in the condition
	 * 
	 */
	final static Variable TESTED_VARIABLE_04 = Variable.getVariable("3");

	/**
	 * 
	 * First condition to test
	 * 
	 */
	final static String STATE_CONDITION = "<cnd:cond operator='or'>"
			+ "  <cnd:node variable='" + TESTED_VARIABLE_01 + "' value='1'/>"
			+ "  <cnd:node variable='" + TESTED_VARIABLE_02 + "' value='0'/>"
			+ "  <cnd:cond operator='and'>" + "    <cnd:node variable='"
			+ TESTED_VARIABLE_03 + "' value='0'/>" + "    <cnd:node variable='"
			+ TESTED_VARIABLE_04 + "' value='1'/>" + "  </cnd:cond>" + "</cnd:cond>";

	/**
	 * 
	 * Create the tested condition
	 * 
	 * @param conditionDescription
	 *          description of the condition to create
	 * 
	 * @return the tested condition
	 * 
	 */
	private IStateCondition createCondition(String conditionDescription) {
		// Define the condition
		return VariableCondition.load(XmlUtils.getRoot(conditionDescription));
	}

	/**
	 * 
	 * Check the tested condition on an example
	 * 
	 * @param state
	 *          the example to test the condition on
	 * 
	 * @return true if the condition is verified on this example
	 * 
	 */
	private boolean checkCondition(Context state) {
		boolean expectedResult = (state.get(TESTED_VARIABLE_01).equals("1")
				|| state.get(TESTED_VARIABLE_02).equals("0") || (state.get(
				TESTED_VARIABLE_03).equals("0") && state.get(TESTED_VARIABLE_04)
				.equals("1")));
		return expectedResult;
	}

	/**
	 * 
	 * Test state conditions loading/saving
	 * 
	 */
	public void testStateConditionsLoadingSaving() {
		IStateCondition condition = createCondition(STATE_CONDITION);
		Element conditionSaved = condition
				.toXml(XmlUtils.createXmlDocument(), null);
		condition = VariableCondition.load(conditionSaved);

		Context[] states = Context.createStateSet(STATE_SIZE,
				Context.ValueSet.BINARY);
		for (Context state : states) {
			assertEquals(checkCondition(state), condition.isTrue(state));
		}
	}

	/**
	 * 
	 * Test if the SetToTrue method makes the condition true
	 * 
	 */
	public void testStateConditionSetToTrueSatisfied() {
		IStateCondition condition = createCondition(STATE_CONDITION);
		Context[] states = Context.createStateSet(STATE_SIZE,
				Context.ValueSet.BINARY);
		for (Context state : states) {
			condition.setToTrue(random, state);
			assertEquals(true, condition.isTrue(state));
		}
	}

	/**
	 * 
	 * Check if the set to true returns all the possibilities (easier to debug
	 * than the next test)
	 * 
	 */
	public void testStateConditionSetToTruePossibilitiesDebug() {
		final String conditionDescription = new String("<cnd:cond operator='or'>"
				+ "  <cnd:node variable='0' value='1'/>"
				+ "  <cnd:node variable='1' value='1'/>"
				+ "  <cnd:cond operator='and'>"
				+ "    <cnd:node variable='2' value='1'/>"
				+ "    <cnd:node variable='3' value='1'/>" + "  </cnd:cond>"
				+ "</cnd:cond>");
		final int stateSize = 4;
		Context initialState = new Context();
		for (int index = 0; index < stateSize; index++) {
			initialState.put(Variable.getVariable(String.valueOf(index)), "0");
		}

		Set<Context> possibilitiesSet = new LinkedHashSet<Context>();
		IStateCondition condition = createCondition(conditionDescription);
		for (int index = 0; index < 1000; index++) {
			Context currentState = new Context(initialState);
			condition.setToTrue(random, currentState);
			assertEquals(true, condition.isTrue(currentState));
			possibilitiesSet.add(currentState);
		}
		assertEquals(7, possibilitiesSet.size());
	}
}
