package animatlab.kodamat.core.functions;

import java.util.Set;

import org.w3c.dom.Element;

import animatlab.kodamat.core.utils.Variable;

/**
 * 
 * Factored function partitioning the state space as a set of sub-function
 * 
 * @param <T>
 *          type of the result of the function
 * 
 * @author Thomas Degris
 * 
 */
public interface IPartitionFunction<T> extends IFunction<T> {
	/**
	 * 
	 * Policy to load a function from a xml node
	 * 
	 * @param <T>
	 *          type of the function to load
	 * 
	 */
	public interface LoadXmlPolicy<T> {
		/**
		 * 
		 * Load the function from a xml node
		 * 
		 * @param element
		 *          the xml node to load
		 * @return the function loaded
		 * 
		 */
		public IPartitionFunction<T> loadXml(Element element);
	}

	/**
	 * 
	 * @return the set of partitions of the state space
	 * 
	 */
	Set<IPartition<T>> partitions();

	/**
	 * 
	 * Clone this function
	 * 
	 * @return a clone of this function
	 * 
	 */
	IPartitionFunction<T> clone();

	/**
	 * 
	 * @return the scope of this function
	 * 
	 */
	Set<Variable> scope();
}
