/*
 * 
 * Created on Dec 29, 2005
 *
 */
package animatlab.kodamat.core.io;

import org.w3c.dom.Element;

/**
 * 
 * Policy to convert a node in a decision tree to a xml element and vice-versa
 * 
 * @param <T>
 *          type of the node to convert
 * 
 * @author Thomas Degris
 * 
 */
public interface DataLoadingPolicy<T> {
	/**
	 * 
	 * Convert a xml element to a node in a decision tree
	 * 
	 * @param element
	 *          the xml node to convert
	 * 
	 * @return the data created from the xml node
	 * 
	 */
	T xmlToData(Element element);

	/**
	 * 
	 * Policy to set the node in the tested trees from xml
	 * 
	 */
	public static final DataLoadingPolicy<String> stringLoadingPolicy = new DataLoadingPolicy<String>() {
		/**
		 * @see animatlab.kodamat.core.io.DataLoadingPolicy#xmlToData(org.w3c.dom.Element)
		 */
		public String xmlToData(Element element) {
			return element.getAttribute("data");
		}
	};

	/**
	 * 
	 * Policy to load the sum from a xml tree
	 * 
	 */
	public static final DataLoadingPolicy<Double> doubleLoadingPolicy = new DataLoadingPolicy<Double>() {
		public Double xmlToData(Element element) {
			String stringValue = element.getAttribute("data");
			if (stringValue.length() == 0) {
				return null;
			}
			return Double.valueOf(stringValue);
		}
	};

}
