package animatlab.kodamat.core.io;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;

/**
 * 
 * Used to exec command in the environment
 * 
 * @author Thomas Degris
 * 
 */
public class Exec {

	/**
	 * 
	 * Thread used to output a stream into the log
	 * 
	 */
	static class StreamReader extends Thread {
		/**
		 * 
		 * The stream to read and forward to the log
		 * 
		 */
		private InputStream inputStream;

		/**
		 * 
		 * Log used as output
		 * 
		 */
		private Log log;

		/**
		 * 
		 * Constructor
		 * 
		 * @param inputStream
		 *          input stream to output into the log
		 * @param log
		 *          the log used as output
		 * 
		 */
		StreamReader(InputStream inputStream, Log log) {
			this.inputStream = inputStream;
			this.log = log;
		}

		/**
		 * @see java.lang.Runnable#run()
		 */
		@Override
		public void run() {
			try {
				InputStreamReader inputStreamReader = new InputStreamReader(inputStream);
				BufferedReader bufferedReader = new BufferedReader(inputStreamReader);
				String line = null;
				while ((line = bufferedReader.readLine()) != null) {
					if (log != null) {
						log.out().println(line);
					}
				}
			} catch (IOException ioe) {
				// Nothing to do if something goes wrong : just stop
			}
		}
	}

	/**
	 * 
	 * Execute a command in the environment
	 * 
	 * @param command
	 *          the command to execute
	 * @param log
	 *          the log used as output for the error and output of the process
	 * @return the exit code of the process
	 * 
	 */
	public static int exec(String command, Log log) {
		int exitCode = 0;
		try {
			if (log != null) {
				log.out().println("Running " + command);
			}
			Process process = Runtime.getRuntime().exec(command);
			new StreamReader(process.getErrorStream(), log).start();
			new StreamReader(process.getInputStream(), log).start();
			exitCode = process.waitFor();
		} catch (IOException e) {
			e.printStackTrace();
			System.exit(1);
		} catch (InterruptedException e) {
			e.printStackTrace();
			System.exit(1);
		}
		return exitCode;
	}

	/**
	 * 
	 * Execute a command in the environment
	 * 
	 * @param command
	 *          the command to execute
	 * @return the exit code of the process
	 * 
	 */
	public static int exec(String command) {
		return exec(command, null);
	}

}
