package animatlab.kodamat.core.io;

import animatlab.kodamat.core.io.DotUtils.ImageMap;
import animatlab.kodamat.core.utils.Parameters;

/**
 * 
 * Policy to output a data structure into a dot file
 * 
 * @author Thomas Degris
 * @param <T>
 *          type contained in the data structure
 * 
 */
public interface IToDot<T> {
	/**
	 * 
	 * Policy to output data structure into dot file
	 * 
	 * @param <T>
	 *          type contained in the data structure
	 * 
	 */
	public interface IToDotPolicy<T> {
		/**
		 * 
		 * Convert the data into a dot label
		 * 
		 * @param data
		 *          the data to convert (is different of null)
		 * 
		 * @return a string representing the dot label
		 * 
		 */
		String toDotLabel(T data);
	}

	/**
	 * 
	 * Policy using the toString method
	 * 
	 */
	static public IToDotPolicy<Object> toStringPolicy = new IToDotPolicy<Object>() {
		public String toDotLabel(Object data) {
			return String.valueOf(data);
		}
	};

	/**
	 * 
	 * Policy using the toString method
	 * 
	 */
	static public IToDotPolicy<Double> doublePolicy = new IToDotPolicy<Double>() {
		public String toDotLabel(Double data) {
			return Parameters.format(data);
		}
	};

	/**
	 * 
	 * Output this into a dot file
	 * 
	 * @param output
	 *          the output where to print dot data
	 * @param policy
	 *          the policy to use to print data into dot data
	 * @param imageMap
	 *          map of the image to fill
	 * 
	 */
	void toDot(StringBuilder output, IToDotPolicy<T> policy, ImageMap<T> imageMap);

	/**
	 * 
	 * @return the dot name of this
	 * 
	 */
	String dotName();
}
