/*
 * 
 * Created on Jan 12, 2006
 *
 */
package animatlab.kodamat.core.io;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

/**
 * 
 * Interface used when an instance can be converted to a xml tree
 * 
 * @param <T>
 *          the data to convert
 * 
 * @author Thomas Degris
 * 
 */
public interface IToXml<T> {
	/**
	 * 
	 * The policy used to convert the data
	 * 
	 * @param <T>
	 *          the data to convert
	 * 
	 */
	public interface IToXmlPolicy<T> {
		/**
		 * 
		 * Convert the data into xml data
		 * 
		 * @param parent
		 *          the xml node containing the xml data
		 * @param data
		 *          the data to convert
		 * 
		 */
		public void toXml(Element parent, T data);
	}

	/**
	 * 
	 * Convert this instance into a xml tree
	 * 
	 * @param document
	 *          the document owning the xml tree
	 * @param policy
	 *          policy to use to convert the data
	 * 
	 * @return the root of the xml tree
	 * 
	 */
	public Element toXml(Document document, IToXmlPolicy<T> policy);

	/**
	 * 
	 * Policy to save double data into a xml node
	 * 
	 */
	public final static IToXmlPolicy<Double> doubleToXmlPolicy = new IToXmlPolicy<Double>() {
		public void toXml(Element parent, Double data) {
			parent.setAttribute("data", data.toString());
		}
	};
}
