/*
 * 
 * Created on 21 oct. 2005
 *
 */
package animatlab.kodamat.core.io;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.PrintStream;

/**
 * 
 * Class to set a log output file instead of the standard output
 * 
 * @author Thomas Degris
 * 
 */
public class Log {
	/**
	 * 
	 * Output for logs
	 * 
	 */
	private PrintStream out = null;

	/**
	 * 
	 * Output file where log are printed
	 * 
	 */
	final private File logFile;

	/**
	 * 
	 * The standard output stream
	 * 
	 */
	public static Log sysout = new Log(System.out);

	/**
	 * 
	 * Constructor
	 * 
	 * @param filepath
	 *          the filepath where to output the log
	 * 
	 */
	public Log(String filepath) {
		logFile = new File(filepath);
		String path = logFile.getAbsolutePath();
		new File(path.substring(0, path.lastIndexOf(File.separator))).mkdirs();
		try {
			out = new PrintStream(logFile);
		} catch (FileNotFoundException e) {
			e.printStackTrace();
			System.exit(1);
		}
	}

	/**
	 * 
	 * Constructor to build the log from an already existing print stream
	 * 
	 * @param out
	 * 
	 */
	private Log(PrintStream out) {
		this.out = out;
		logFile = null;
	}

	/**
	 * 
	 * @return the current log output
	 * 
	 */
	final public PrintStream out() {
		return out;
	}

	/**
	 * 
	 * Unset the output
	 * 
	 */
	@Override
	public void finalize() {
		out.flush();
		out.close();
	}

	/**
	 * 
	 * Remove a file/directory
	 * 
	 * @param filepath
	 *          the file to remove
	 * 
	 */
	static final public void removeRecursively(String filepath) {
		File file = new File(filepath);
		if (file.isDirectory()) {
			for (String subfile : file.list()) {
				removeRecursively(filepath + "/" + subfile);
			}
		}
		file.delete();
	}

	/**
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		if (logFile == null) {
			return "?";
		}
		return logFile.getAbsolutePath();
	}
}
