package animatlab.kodamat.core.io;

import java.io.FileWriter;
import java.io.IOException;

/**
 * 
 * Helper to write xml file for results
 * 
 * @author Thomas Degris
 * 
 */
public class ResultsXmlFile {
	/**
	 * 
	 * Log file to log the number of time step before the reward is reached
	 * 
	 */
	private FileWriter output = null;

	/**
	 * 
	 * Name of the root node of the xml file
	 * 
	 */
	private final String rootNodeName;

	/**
	 * 
	 * Constructor
	 * 
	 * @param rootNodeName
	 *          name of the root node
	 * @param filePath
	 *          path of the xml file
	 * 
	 */
	public ResultsXmlFile(String rootNodeName, String filePath) {
		this.rootNodeName = rootNodeName;
		try {
			output = new FileWriter(filePath);
		} catch (IOException e) {
			e.printStackTrace();
			System.exit(1);
		}
		write("<" + rootNodeName + ">\n");
	}

	/**
	 * 
	 * Append a new string in the xml text file
	 * 
	 * @param str
	 *          the string to add
	 * 
	 */
	public void write(String str) {
		try {
			output.write(str);
		} catch (IOException e) {
			e.printStackTrace();
			System.exit(1);
		}
		flush();
	}

	/**
	 * 
	 * Flush the current file
	 * 
	 */
	private void flush() {
		try {
			output.flush();
		} catch (IOException e) {
			e.printStackTrace();
			System.exit(1);
		}
	}

	/**
	 * 
	 * Close the xml file
	 * 
	 */
	public void close() {
		write("</" + rootNodeName + ">");
		flush();
		try {
			output.close();
		} catch (IOException e) {
			e.printStackTrace();
			System.exit(1);
		}
		output = null;
	}
}
