/*
 * 
 * Created on Jan 12, 2006
 *
 */
package animatlab.kodamat.core.io;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.Calendar;

/**
 * 
 * Usefull function to use stamp file
 * 
 * @author Thomas Degris
 * 
 */
public class StampFile {
	/**
	 * 
	 * Suffix added to stamp file
	 * 
	 */
	private static final String FINISHED_SUFFIX = ".end.stamp";

	/**
	 * 
	 * Suffix added to stamp file
	 * 
	 */
	private static final String WORKING_SUFFIX = ".working.stamp";

	/**
	 * 
	 * Check if a file exist
	 * 
	 * @param filepath
	 *          the file to check
	 * 
	 * @return true if the file exist
	 * 
	 */
	static public boolean exists(String filepath) {
		if (filepath == null) {
			return false;
		}
		return (new File(filepath + FINISHED_SUFFIX).canRead() || new File(filepath
				+ WORKING_SUFFIX).canRead());
	}

	/**
	 * 
	 * Remove a stamp file
	 * 
	 * @param filepath
	 *          the file to remove
	 * 
	 */
	static public void remove(String filepath) {
		new File(filepath + WORKING_SUFFIX).delete();
		new File(filepath + FINISHED_SUFFIX).delete();
	}

	/**
	 * 
	 * Create a stamp file
	 * 
	 * @param filepath
	 *          file used as stamp
	 * 
	 */
	static public void touchWorking(String filepath) {
		if (filepath == null) {
			return;
		}
		assert (!new File(filepath + FINISHED_SUFFIX).canRead());
		File workingStamp = touch(filepath, WORKING_SUFFIX);
		workingStamp.deleteOnExit();
	}

	/**
	 * 
	 * Create a stamp file
	 * 
	 * @param filepath
	 *          file used as stamp
	 * 
	 */
	static public void touchFinished(String filepath) {
		if (filepath == null) {
			return;
		}
		remove(filepath);
		touch(filepath, FINISHED_SUFFIX);
	}

	/**
	 * 
	 * Create a stamp file
	 * 
	 * @param filepath
	 *          file used as stamp
	 * @param suffix
	 *          extension of the file to create
	 * 
	 * @return the file corresponding to the stamp file touched
	 * 
	 */
	private static File touch(String filepath, String suffix) {
		FileWriter output;
		File stampFile = new File(filepath + suffix);
		try {
			output = new FileWriter(stampFile);
			output.write(java.net.InetAddress.getLocalHost().getHostName());
			output.write(": ");
			DateFormat dateFormat = new SimpleDateFormat("dd-MM-yy HH'h'mm");
			output.write(dateFormat.format(Calendar.getInstance().getTime()) + "\n");
			output.flush();
			output.close();
		} catch (IOException e) {
			e.printStackTrace();
			System.exit(1);
		}
		return stampFile;
	}
}
