package animatlab.kodamat.core.proba.tests;

import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.Map;
import java.util.Set;

import junit.framework.TestCase;
import animatlab.kodamat.core.proba.Chi2Utils;
import animatlab.kodamat.core.proba.ILearningProbabilities;
import animatlab.kodamat.core.proba.ValueStatistics;

/**
 * 
 * Test Chi2 computation
 * 
 * @author Thomas Degris
 * 
 */
public class TestProba extends TestCase {
	/**
	 * 
	 * Chi2 utils instance
	 * 
	 */
	static private Chi2Utils<String> chi2Utils = new Chi2Utils<String>();

	/**
	 * 
	 * Create a new stat instance for a sample
	 * 
	 * @param className
	 *          the class name of the sample
	 * @param theoData
	 *          the data for the sample
	 * @return the new stat instance
	 * 
	 */
	private ValueStatistics<String> getStats(String[] className, int[] theoData) {
		assert (className.length == theoData.length);
		Map<String, Integer> dataMap = new LinkedHashMap<String, Integer>();
		for (int index = 0; index < className.length; index++) {
			dataMap.put(className[index], theoData[index]);
		}
		ValueStatistics<String> result = new ValueStatistics<String>();
		result.merge(dataMap);
		return result;
	}

	/**
	 * 
	 * Test Chi2 computation between theo vs obs
	 * 
	 */
	public void testComputeChi2TheoObs() {
		String[] weekDays = { "Monday", "Tuesday", "Wednesday", "Thursday",
				"Friday", "Saturday", "Sunday" };

		int[] theoData = { 1, 1, 1, 1, 1, 1, 1 };
		ValueStatistics<String> theoStats = getStats(weekDays, theoData);

		int[] obsData = { 564772, 629408, 609596, 604812, 605280, 450840, 404456 };
		ValueStatistics<String> obsStats = getStats(weekDays, obsData);

		double chi2 = chi2Utils.computeChi2(obsStats, theoStats);
		assertEquals((int) chi2, 85210);
	}

	/**
	 * 
	 * Test Chi2 computation two obs distributions
	 * 
	 */
	public void testComputeChi2ObsObs() {
		String[] genre = { "Male", "Female", "Immature" };

		int[] obs01Data = { 20, 20, 60 };
		ValueStatistics<String> obs01Stats = getStats(genre, obs01Data);
		int[] obs02Data = { 05, 30, 90 };
		ValueStatistics<String> obs02Stats = getStats(genre, obs02Data);

		Set<ILearningProbabilities<String>> set = new LinkedHashSet<ILearningProbabilities<String>>();
		set.add(obs01Stats);
		set.add(obs02Stats);
		double chi2 = chi2Utils.computeChi2(set);
		assertEquals((int) chi2, 14);
	}

}
