package animatlab.kodamat.core.rules;

import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import animatlab.kodamat.core.functions.IPartition;
import animatlab.kodamat.core.io.DataLoadingPolicy;
import animatlab.kodamat.core.utils.Context;
import animatlab.kodamat.core.utils.Variable;

/**
 * 
 * Simple rule
 * 
 * @author Thomas Degris
 * @param <T>
 *          type of the data pointed by this rule
 * 
 */
public class PlainRule<T> extends Rule<T> {
	/**
	 * 
	 * Value pointed by the rule
	 * 
	 */
	private T pointedValue;

	/**
	 * 
	 * Expression composing the rule
	 * 
	 */
	final private Context context = new Context();

	/**
	 * 
	 * Constructor
	 * 
	 * @param pointedValue
	 *          value pointed by the rule
	 * 
	 */
	public PlainRule(T pointedValue) {
		this.pointedValue = pointedValue;
	}

	/**
	 * 
	 * Constructor
	 * 
	 * @param partition
	 *          partition to create this rule
	 * 
	 */
	public PlainRule(IPartition<T> partition) {
		this.pointedValue = partition.get();
		setContext(partition.getContext());
	}

	/**
	 * 
	 * Constructor
	 * 
	 * @param element
	 *          root of the xml tree
	 * @param dataLoadingPolicy
	 *          policy to load the value pointed by the rule
	 * 
	 */
	public PlainRule(Element element, DataLoadingPolicy<T> dataLoadingPolicy) {
		pointedValue = dataLoadingPolicy.xmlToData(element);
		NodeList nodeList = element.getChildNodes();
		for (int index = 0; index < nodeList.getLength(); index++) {
			Node childNode = nodeList.item(index);
			if (childNode.getNodeType() == Node.ELEMENT_NODE) {
				if (childNode.getNodeName().equals(NODE_NAME_CONSTRAINT)) {
					Element constraintElement = (Element) childNode;
					Variable variable = Variable.load(constraintElement);
					String value = constraintElement.getAttribute(ATTRIBUTE_VALUE);
					addConstraint(variable, value);
				}
			}
		}
	}

	/**
	 * 
	 * Add a constraint to the rule
	 * 
	 * @param variable
	 *          variable to test in the constraint
	 * @param value
	 *          value of the variable to test with
	 * 
	 */
	public void addConstraint(Variable variable, String value) {
		resetHashCode();
		context.put(variable, value);
	}

	/**
	 * @see animatlab.kodamat.core.functions.IPartition#getContext()
	 */
	public Context getContext() {
		return context;
	}

	/**
	 * 
	 * Set all the conditions from an existing context set
	 * 
	 * @param context
	 *          the existing context set
	 * 
	 */
	public void setContext(Context context) {
		this.context.putAll(context);
	}

	/**
	 * 
	 * Set the value pointed by this rule
	 * 
	 * @param data
	 *          the data to set
	 * 
	 */
	public void set(T data) {
		pointedValue = data;
	}

	/**
	 * @see animatlab.kodamat.core.io.DataContainer#get()
	 */
	public T get() {
		return pointedValue;
	}
}
