package animatlab.kodamat.core.trees;

/**
 * 
 * Policy used to append two leaves together
 * 
 * @param <T>
 *          type of data contained at the leaves
 * 
 */
public interface AppendPolicy<T> {
	/**
	 * 
	 * Append two leaves together
	 * 
	 * @param originalLeaf
	 *          the original leaf. It MUST NOT be modified.
	 * @param appendedLeaf
	 *          the appended leaf that can be modified in the append operation
	 * 
	 */
	public void append(Leaf<T> originalLeaf, Leaf<T> appendedLeaf);

	/**
	 * 
	 * Policy used to add two trees of double
	 * 
	 */
	final static public AppendPolicy<Double> sumPolicy = new AppendPolicy<Double>() {
		public void append(Leaf<Double> originalLeaf, Leaf<Double> appendedLeaf) {
			if (appendedLeaf.get() == null) {
				appendedLeaf.set(0.0);
			}
			appendedLeaf.set(appendedLeaf.get() + originalLeaf.get());
		}
	};

	/**
	 * 
	 * Policy used to add two trees of double
	 * 
	 */
	final static public AppendPolicy<Double> minusPolicy = new AppendPolicy<Double>() {
		public void append(Leaf<Double> originalLeaf, Leaf<Double> appendedLeaf) {
			if (appendedLeaf.get() == null) {
				appendedLeaf.set(0.0);
			}
			appendedLeaf.set(originalLeaf.get() - appendedLeaf.get());
		}
	};

	/**
	 * 
	 * Policy used to maximize two trees of double
	 * 
	 */
	final static public AppendPolicy<Double> maxPolicy = new AppendPolicy<Double>() {
		public void append(Leaf<Double> originalLeaf, Leaf<Double> appendedLeaf) {
			appendedLeaf.set(Math.max(appendedLeaf.get(), originalLeaf.get()));
		}
	};

	/**
	 * 
	 * Policy used to minimize two trees of double
	 * 
	 */
	final static public AppendPolicy<Double> minPolicy = new AppendPolicy<Double>() {
		public void append(Leaf<Double> originalLeaf, Leaf<Double> appendedLeaf) {
			appendedLeaf.set(Math.min(appendedLeaf.get(), originalLeaf.get()));
		}
	};

}