/*
 * 
 * Created on Dec 30, 2005
 *
 */
package animatlab.kodamat.core.trees;

import animatlab.kodamat.core.utils.Parameters;

/**
 * 
 * Policy used to test if two leaves of a tree are equals
 * 
 * @param <T>
 *          the type of leaf to compare
 * 
 */
public interface EqualPolicy<T> {
	/**
	 * 
	 * Policy to compare doubles
	 * 
	 */
	public class ValueLeafEqualPolicy implements EqualPolicy<Double> {
		/**
		 * 
		 * Epsilon to use in the comparaison
		 * 
		 */
		private double epsilon;

		/**
		 * 
		 * Constructor
		 * 
		 * @param epsilon
		 *          epsilon to use in the comparaison
		 * 
		 */
		public ValueLeafEqualPolicy(double epsilon) {
			this.epsilon = epsilon;
		}

		/**
		 * @see animatlab.kodamat.core.trees.EqualPolicy#equals(animatlab.kodamat.core.trees.Leaf,
		 *      animatlab.kodamat.core.trees.Leaf)
		 */
		public boolean equals(Leaf<Double> leaf01, Leaf<Double> leaf02) {
			assert (leaf01.get() != null);
			assert (leaf02.get() != null);
			double data01 = leaf01.get();
			double data02 = leaf02.get();
			assert (!Double.isNaN(data01));
			assert (!Double.isNaN(data02));
			return ((Math.abs(data01 - data02) < epsilon));
		}
	}

	/**
	 * 
	 * Instance of a ValueLeafEqualPolicy
	 * 
	 */
	public final static EqualPolicy<Double> valueLeafEqualPolicy = new ValueLeafEqualPolicy(
			Parameters.EPSILON);

	/**
	 * 
	 * Tests is two data contained at leaves are equals
	 * 
	 * @param leaf01
	 *          the data to compare
	 * @param leaf02
	 *          the data to compare
	 * @return true if both data are equals
	 * 
	 */
	boolean equals(Leaf<T> leaf01, Leaf<T> leaf02);
}
