package animatlab.kodamat.core.trees;

/**
 * 
 * Policy used to merge two trees
 * 
 * @param <T>
 *          type contained by the trees to merge
 * 
 * @author Thomas Degris
 * 
 */
public class MergePolicy<T> {
	/**
	 * 
	 * Policy to use to merge subtree with trees
	 * 
	 */
	final private AppendPolicy<T> mergePolicy;

	/**
	 * 
	 * Policy to use to extract subtree from trees
	 * 
	 */
	final private ClonePolicy<T> clonePolicy;

	/**
	 * 
	 * Constructor
	 * 
	 * @param mergePolicy
	 *          policy to use to merge subtree with trees
	 * @param clonePolicy
	 *          policy to use to extract subtree from trees
	 * 
	 */
	public MergePolicy(AppendPolicy<T> mergePolicy, ClonePolicy<T> clonePolicy) {
		this.mergePolicy = mergePolicy;
		this.clonePolicy = clonePolicy;
	}

	/**
	 * 
	 * @return the policy used to append nodes
	 * 
	 */
	public AppendPolicy<T> getAppendPolicy() {
		return mergePolicy;
	}

	/**
	 * 
	 * Merge a set of trees
	 * 
	 * @param newTree
	 *          tree to merge with the working tree
	 * @param workingTree
	 *          tree used to merge the new tree (WARNING: this tree is modified)
	 * @return the merged tree
	 * 
	 */
	public Tree<T> merge(Tree<T> newTree, Tree<T> workingTree) {
		if (newTree == null) {
			return workingTree;
		}
		if (workingTree == null) {
			return newTree.clone();
		}
		workingTree.append(newTree, mergePolicy, clonePolicy);
		return workingTree;
	}

	/**
	 * 
	 * Policy used to add two trees of double
	 * 
	 */
	final static public MergePolicy<Double> sumPolicy = new MergePolicy<Double>(
			AppendPolicy.sumPolicy, null);

	/**
	 * 
	 * Policy used to maximize two trees of double
	 * 
	 */
	final static public MergePolicy<Double> maxPolicy = new MergePolicy<Double>(
			AppendPolicy.maxPolicy, null);

	/**
	 * 
	 * Policy used to maximize two trees of double
	 * 
	 */
	final static public MergePolicy<Double> minPolicy = new MergePolicy<Double>(
			AppendPolicy.minPolicy, null);

}
