/*
 * 
 * Created on 22 dec. 2005
 *
 */
package animatlab.kodamat.core.trees.test;

import static animatlab.kodamat.core.io.XmlUtils.getRoot;

import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Set;

import junit.framework.TestCase;

import org.w3c.dom.Element;

import animatlab.kodamat.core.io.DataLoadingPolicy;
import animatlab.kodamat.core.io.IToXml;
import animatlab.kodamat.core.io.XmlUtils;
import animatlab.kodamat.core.trees.AppendPolicy;
import animatlab.kodamat.core.trees.ConvertPolicy;
import animatlab.kodamat.core.trees.DNode;
import animatlab.kodamat.core.trees.EqualPolicy;
import animatlab.kodamat.core.trees.Leaf;
import animatlab.kodamat.core.trees.Tree;
import animatlab.kodamat.core.trees.Vertex;
import animatlab.kodamat.core.utils.Context;
import animatlab.kodamat.core.utils.Variable;

/**
 * 
 * Tests case for decision trees
 * 
 * @author Thomas Degris
 * 
 */
public class TestDecisionTrees extends TestCase {
	/**
	 * 
	 * Test the set/get method of leaf
	 * 
	 */
	public void testData() {
		String test = "Hello World";
		Tree<String> tree = new Tree<String>();
		((Leaf<String>) tree.getRoot()).set(test);
		assertEquals(test, ((Leaf<String>) tree.getRoot()).get());
	}

	/**
	 * 
	 * Policy to convert a string into a double
	 * 
	 */
	static class StringToDoublePolicy extends ConvertPolicy<String, Double> {
		/**
		 * 
		 * Convert from String to Double
		 * 
		 * @param leaf
		 *          the string to convert
		 * @return the double converted
		 * 
		 */
		@Override
		public Double convert(Leaf<String> leaf) {
			if (leaf == null) {
				return -55.0;
			}
			return Double.parseDouble(leaf.get());
		}

	}

	/**
	 * 
	 * Build a tree from a xml definition
	 * 
	 * @param xmlString
	 *          the xml definition of the tree
	 * 
	 * @return the new tree
	 * 
	 */
	private Tree<String> getTree(String xmlString) {
		return new Tree<String>(getRoot(xmlString),
				DataLoadingPolicy.stringLoadingPolicy);
	}

	/**
	 * 
	 * Test expand against xml loading
	 * 
	 */
	public void testXmlLoadingSaving() {
		Tree<String> tree = new Tree<String>();
		DNode<String> dnode = tree.getRoot().replace(Variable.getVariable("4"));
		dnode.addLeaf("0", "Hello");
		dnode = dnode.addDNode("1", Variable.getVariable("3"));
		dnode.addLeaf("0", "world");
		dnode.addLeaf("1", "and");
		dnode.addLeaf("2", "everybody");
		String xmlString = "<dtr:dnode variable='4'>"
				+ "  <dtr:leaf value='0' data='Hello'/>"
				+ "  <dtr:dnode value='1' variable='3'>"
				+ "    <dtr:leaf value='0' data='world'/>"
				+ "    <dtr:leaf value='1' data='and'/>"
				+ "    <dtr:leaf value='2' data='everybody'/>" + "  </dtr:dnode>"
				+ "</dtr:dnode>";
		assertEquals(getTree(xmlString).toString(), tree.toString());

		xmlString = XmlUtils.toXmlString(tree, new IToXml.IToXmlPolicy<String>() {
			public void toXml(Element parent, String data) {
				parent.setAttribute("data", data);
			}
		});
		assertEquals(getTree(xmlString).toString(), tree.toString());

		xmlString = "<dtr:leaf data='Hello'/>";
		tree = new Tree<String>();
		((Leaf<String>) tree.getRoot()).set("Hello");
		assertEquals(getTree(xmlString).toString(), tree.toString());
	}

	/**
	 * 
	 * Test the satisfy method
	 * 
	 */
	public void testStatisfy() {
		String xmlString = "<dtr:dnode variable='4'>"
				+ "  <dtr:leaf value='0' data='Hello'/>"
				+ "  <dtr:dnode value='1' variable='3'>"
				+ "    <dtr:leaf value='0' data='world'/>"
				+ "    <dtr:leaf value='1' data='and'/>"
				+ "    <dtr:leaf value='2' data='everybody'/>" + "  </dtr:dnode>"
				+ "</dtr:dnode>";
		Tree<String> treeTested = getTree(xmlString);
		Leaf<String> leaf = treeTested.getLeaf(Context.getState(",,,1,1", null));
		assertEquals(true, leaf.satisfy(Context.getState("4=1,3=1", null)));
		assertEquals(true, leaf.satisfy(Context.getState(",,,1,1,2", null)));
		assertEquals(false, leaf.satisfy(Context.getState(",,,,0", null)));
	}

	/**
	 * 
	 * Test the remove method on trees
	 * 
	 */
	public void testRemove() {
		String xmlString = "<dtr:dnode variable='4'>"
				+ "  <dtr:leaf value='0' data='Hello'/>"
				+ "  <dtr:dnode value='1' variable='3'>"
				+ "    <dtr:leaf value='0' data='world'/>"
				+ "    <dtr:leaf value='1' data='and'/>"
				+ "    <dtr:leaf value='2' data='everybody'/>" + "  </dtr:dnode>"
				+ "</dtr:dnode>";
		Tree<String> treeTested = getTree(xmlString);
		DNode<String> dnode = (DNode<String>) treeTested.getRoot();
		dnode = (DNode<String>) dnode.getChild("1");
		dnode.remove();
		xmlString = "<dtr:dnode variable='4'>"
				+ "  <dtr:leaf value='0' data='Hello'/>" + "</dtr:dnode>";
		assertEquals(getTree(xmlString).toString(), treeTested.toString());
	}

	/**
	 * 
	 * Test the remove method on trees
	 * 
	 */
	public void testReplaceWithLeaf() {
		String xmlString = "<dtr:dnode variable='4'>"
				+ "  <dtr:leaf value='0' data='Hello'/>"
				+ "  <dtr:dnode value='1' variable='3'>"
				+ "    <dtr:leaf value='0' data='world'/>"
				+ "    <dtr:leaf value='1' data='and'/>"
				+ "    <dtr:leaf value='2' data='everybody'/>" + "  </dtr:dnode>"
				+ "</dtr:dnode>";
		Tree<String> treeTested = getTree(xmlString);
		DNode<String> dnode = (DNode<String>) treeTested.getRoot();
		dnode = (DNode<String>) dnode.getChild("1");
		Leaf<String> leaf = dnode.replaceWithLeaf();
		leaf.set("Bonjour");
		xmlString = "<dtr:dnode variable='4'>"
				+ "  <dtr:leaf value='0' data='Hello'/>"
				+ "  <dtr:leaf value='1' data='Bonjour'/>" + "</dtr:dnode>";
		assertEquals(getTree(xmlString).toString(), treeTested.toString());
	}

	/**
	 * 
	 * Test the replace in a tree of a leaf by an DNode
	 * 
	 */
	public void testReplaceLeafWithDNode() {
		String xmlString = "<dtr:dnode variable='4'>"
				+ "  <dtr:leaf value='0' data='Hello'/>"
				+ "  <dtr:leaf value='1' data='Coucou'/>" + "</dtr:dnode>";
		Tree<String> treeTested = getTree(xmlString);
		Vertex<String> vertex = ((DNode<String>) treeTested.getRoot())
				.getChild("1");
		DNode<String> dnode = vertex.replace(Variable.getVariable("3"));
		dnode.addLeaf("0", "Bonjour");
		dnode.addLeaf("1", "Coucou");

		xmlString = "<dtr:dnode variable='4'>"
				+ "  <dtr:leaf value='0' data='Hello'/>"
				+ "  <dtr:dnode value='1' variable='3'>"
				+ "    <dtr:leaf value='0' data='Bonjour'/>"
				+ "    <dtr:leaf value='1' data='Coucou'/>" + "  </dtr:dnode>"
				+ "</dtr:dnode>";
		assertEquals(getTree(xmlString).toString(), treeTested.toString());
	}

	/**
	 * 
	 * Test the clone method on trees
	 * 
	 */
	public void testClone() {
		String xmlString = "<dtr:dnode variable='4'>"
				+ "  <dtr:leaf value='0' data='Hello'/>"
				+ "  <dtr:dnode value='1' variable='3'>"
				+ "    <dtr:leaf value='0' data='world'/>"
				+ "    <dtr:leaf value='1' data='and'/>"
				+ "    <dtr:leaf value='2' data='everybody'/>" + "  </dtr:dnode>"
				+ "</dtr:dnode>";
		Tree<String> refTree = new Tree<String>(getRoot(xmlString),
				DataLoadingPolicy.stringLoadingPolicy);
		Tree<String> treeClone = refTree.clone();
		assertEquals(refTree.toString(), treeClone.toString());

		Context testsList = new Context();
		testsList.put(Variable.getVariable("4"), "1");
		xmlString = "<dtr:dnode value='1' variable='3'>"
				+ "  <dtr:leaf value='0' data='world'/>"
				+ "  <dtr:leaf value='1' data='and'/>"
				+ "  <dtr:leaf value='2' data='everybody'/>" + "</dtr:dnode>";
		treeClone = refTree.clone(testsList);
		assertEquals(getTree(xmlString).toString(), treeClone.toString());

		testsList.clear();
		testsList.put(Variable.getVariable("3"), "0");
		xmlString = "<dtr:dnode variable='4'>"
				+ "  <dtr:leaf value='0' data='Hello'/>"
				+ "  <dtr:leaf value='1' data='world'/>" + "</dtr:dnode>";
		treeClone = refTree.clone(testsList);
		assertEquals(getTree(xmlString).toString(), treeClone.toString());
	}

	/**
	 * 
	 * Test the extract method on trees
	 * 
	 */
	public void testExtract() {
		String xmlString = "<dtr:dnode variable='4'>"
				+ "  <dtr:leaf value='0' data='Hello'/>"
				+ "  <dtr:dnode value='1' variable='3'>"
				+ "    <dtr:leaf value='0' data='world'/>"
				+ "    <dtr:leaf value='1' data='and'/>"
				+ "    <dtr:leaf value='2' data='everybody'/>" + "  </dtr:dnode>"
				+ "</dtr:dnode>";
		Tree<String> initialTree = new Tree<String>(getRoot(xmlString),
				DataLoadingPolicy.stringLoadingPolicy);
		Tree<String> extractedTree = ((DNode<String>) initialTree.getRoot())
				.getChild("1").extract();

		xmlString = "<dtr:dnode variable='4'>"
				+ "  <dtr:leaf value='0' data='Hello'/>" + "</dtr:dnode>";
		assertEquals(getTree(xmlString).toString(), initialTree.toString());

		xmlString = "<dtr:dnode variable='3'>"
				+ "  <dtr:leaf value='0' data='world'/>"
				+ "  <dtr:leaf value='1' data='and'/>"
				+ "  <dtr:leaf value='2' data='everybody'/>" + "</dtr:dnode>";
		assertEquals(getTree(xmlString).toString(), extractedTree.toString());

	}

	/**
	 * 
	 * Test the replace method on trees
	 * 
	 */
	public void testReplace() {
		String xmlString = "<dtr:dnode variable='4'>"
				+ "  <dtr:leaf value='0' data='Hello'/>"
				+ "  <dtr:leaf value='1' data='Bonjour'/>" + "</dtr:dnode>";
		Tree<String> parentTree = getTree(xmlString);
		xmlString = "<dtr:dnode variable='3'>"
				+ "  <dtr:leaf value='0' data='world'/>"
				+ "  <dtr:leaf value='1' data='and'/>"
				+ "  <dtr:leaf value='2' data='everybody'/>" + "</dtr:dnode>";
		Tree<String> childTree = getTree(xmlString);
		Vertex<String> replaced = ((DNode<String>) parentTree.getRoot())
				.getChild("1");
		parentTree.replace(replaced, childTree);
		xmlString = "<dtr:dnode variable='4'>"
				+ "  <dtr:leaf value='0' data='Hello'/>"
				+ "  <dtr:dnode value='1' variable='3'>"
				+ "    <dtr:leaf value='0' data='world'/>"
				+ "    <dtr:leaf value='1' data='and'/>"
				+ "    <dtr:leaf value='2' data='everybody'/>" + "  </dtr:dnode>"
				+ "</dtr:dnode>";
		assertEquals(getTree(xmlString).toString(), parentTree.toString());
		assertEquals((new Tree<String>()).toString(), childTree.toString());
	}

	/**
	 * 
	 * Test the replace method on a root of a tree
	 * 
	 */
	public void testReplaceRoot() {
		String xmlString = "<dtr:dnode variable='4'>"
				+ "  <dtr:leaf value='0' data='Hello'/>"
				+ "  <dtr:leaf value='1' data='Bonjour'/>" + "</dtr:dnode>";
		Tree<String> parentTree = getTree(xmlString);
		xmlString = "<dtr:dnode variable='3'>"
				+ "  <dtr:leaf value='0' data='world'/>"
				+ "  <dtr:leaf value='1' data='and'/>"
				+ "  <dtr:leaf value='2' data='everybody'/>" + "</dtr:dnode>";
		Tree<String> childTree = getTree(xmlString);
		parentTree.replace(parentTree.getRoot(), childTree);
		assertEquals(getTree(xmlString).toString(), parentTree.toString());
		assertEquals((new Tree<String>()).toString(), childTree.toString());
	}

	/**
	 * 
	 * Test the equal method between two trees
	 * 
	 */
	public void testEqual() {
		String xmlString = "<dtr:dnode variable='4'>"
				+ "  <dtr:leaf value='0' data='Hello'/>"
				+ "  <dtr:dnode value='1' variable='3'>"
				+ "    <dtr:leaf value='0' data='world'/>"
				+ "    <dtr:leaf value='1' data='and'/>"
				+ "    <dtr:leaf value='2' data='everybody'/>" + "  </dtr:dnode>"
				+ "</dtr:dnode>";
		Tree<String> refTree = getTree(xmlString);
		assertEquals(true, refTree.equals(getTree(xmlString)));

		xmlString = "<dtr:dnode variable='4'>"
				+ "  <dtr:leaf value='0' data='Hello'/>"
				+ "  <dtr:dnode value='1' variable='2'>"
				+ "    <dtr:leaf value='0' data='world'/>"
				+ "    <dtr:leaf value='1' data='and'/>"
				+ "    <dtr:leaf value='2' data='everybody'/>" + "  </dtr:dnode>"
				+ "</dtr:dnode>";
		assertEquals(false, refTree.equals(getTree(xmlString)));

		xmlString = "<dtr:dnode variable='4'>"
				+ "  <dtr:leaf value='0' data='Hello'/>"
				+ "  <dtr:dnode value='1' variable='3'>"
				+ "    <dtr:leaf value='0' data='world'/>"
				+ "    <dtr:leaf value='1' data='and'/>"
				+ "    <dtr:leaf value='2' data='tout le monde'/>" + "  </dtr:dnode>"
				+ "</dtr:dnode>";
		assertEquals(false, refTree.equals(getTree(xmlString)));

		xmlString = "<dtr:dnode variable='4'>"
				+ "  <dtr:leaf value='0' data='Hello'/>"
				+ "  <dtr:dnode value='2' variable='3'>"
				+ "    <dtr:leaf value='0' data='world'/>"
				+ "    <dtr:leaf value='1' data='and'/>"
				+ "    <dtr:leaf value='2' data='everybody'/>" + "  </dtr:dnode>"
				+ "</dtr:dnode>";
		assertEquals(false, refTree.equals(getTree(xmlString)));

		xmlString = "<dtr:dnode variable='4'>"
				+ "  <dtr:leaf value='0' data='hello'/>"
				+ "  <dtr:dnode value='1' variable='3'>"
				+ "    <dtr:leaf value='0' data='world'/>"
				+ "    <dtr:leaf value='1' data='and'/>"
				+ "    <dtr:leaf value='2' data='everybody'/>" + "  </dtr:dnode>"
				+ "</dtr:dnode>";
		assertEquals(false, refTree.equals(getTree(xmlString)));
		assertEquals(true, refTree.equals(getTree(xmlString),
				new EqualPolicy<String>() {
					public boolean equals(Leaf<String> leaf01, Leaf<String> leaf02) {
						return leaf01.get().equalsIgnoreCase(leaf02.get());
					}
				}));
	}

	/**
	 * 
	 * Test the simplify method on trees
	 * 
	 */
	public void testSimplify() {
		String xmlString = "<dtr:dnode variable='4'>"
				+ "  <dtr:dnode value='0' variable='3'>"
				+ "    <dtr:leaf value='0' data='world'/>"
				+ "    <dtr:leaf value='1' data='and'/>"
				+ "    <dtr:leaf value='2' data='everybody'/>" + "  </dtr:dnode>"
				+ "  <dtr:dnode value='1' variable='3'>"
				+ "    <dtr:leaf value='0' data='world'/>"
				+ "    <dtr:leaf value='1' data='and'/>"
				+ "    <dtr:leaf value='2' data='everybody'/>" + "  </dtr:dnode>"
				+ "</dtr:dnode>";
		Tree<String> simplifiedTree = getTree(xmlString);
		simplifiedTree.simplify(null);

		xmlString = "<dtr:dnode variable='3'>"
				+ "  <dtr:leaf value='0' data='world'/>"
				+ "  <dtr:leaf value='1' data='and'/>"
				+ "  <dtr:leaf value='2' data='everybody'/>" + "</dtr:dnode>";
		assertEquals(getTree(xmlString).toString(), simplifiedTree.toString());
	}

	/**
	 * 
	 * Test the simplify method on trees containing decision node with only one
	 * child
	 * 
	 */
	public void testSimplifyOneBranch() {
		String xmlString = "<dtr:dnode variable='4'>"
				+ "  <dtr:dnode value='0' variable='3'>"
				+ "    <dtr:leaf value='0' data='world'/>"
				+ "    <dtr:leaf value='1' data='and'/>"
				+ "    <dtr:leaf value='2' data='everybody'/>" + "  </dtr:dnode>"
				+ "</dtr:dnode>";
		Tree<String> simplifiedTree = getTree(xmlString);
		simplifiedTree.simplify(null);

		xmlString = "<dtr:dnode variable='3'>"
				+ "  <dtr:leaf value='0' data='world'/>"
				+ "  <dtr:leaf value='1' data='and'/>"
				+ "  <dtr:leaf value='2' data='everybody'/>" + "</dtr:dnode>";
		assertEquals(getTree(xmlString).toString(), simplifiedTree.toString());
	}

	/**
	 * 
	 * Test the sum method on trees
	 * 
	 */
	public void testSum() {
		String xmlString = "<dtr:dnode variable='4'>"
				+ "  <dtr:leaf value='0' variable='3' data='1'/>"
				+ "  <dtr:dnode value='1' variable='3'>"
				+ "    <dtr:leaf value='0' data='3'/>"
				+ "    <dtr:leaf value='1' data='4'/>" + "  </dtr:dnode>"
				+ "</dtr:dnode>";
		Tree<Double> sum01 = new Tree<Double>(getRoot(xmlString),
				DataLoadingPolicy.doubleLoadingPolicy);

		xmlString = "<dtr:dnode variable='4'>"
				+ "  <dtr:dnode value='0' variable='3'>"
				+ "    <dtr:leaf value='0' data='1'/>"
				+ "    <dtr:leaf value='1' data='2'/>" + "  </dtr:dnode>"
				+ "  <dtr:leaf value='1' variable='3' data='1'/>" + "</dtr:dnode>";
		Tree<Double> sum02 = new Tree<Double>(getRoot(xmlString),
				DataLoadingPolicy.doubleLoadingPolicy);

		xmlString = "<dtr:dnode variable='4'>"
				+ "  <dtr:dnode value='0' variable='3'>"
				+ "    <dtr:leaf value='0' data='3'/>"
				+ "    <dtr:leaf value='1' data='5'/>" + "  </dtr:dnode>"
				+ "  <dtr:dnode value='1' variable='3'>"
				+ "    <dtr:leaf value='0' data='5'/>"
				+ "    <dtr:leaf value='1' data='6'/>" + "  </dtr:dnode>"
				+ "</dtr:dnode>";
		Tree<Double> expected = new Tree<Double>(getRoot(xmlString),
				DataLoadingPolicy.doubleLoadingPolicy);

		sum01.append(sum02, AppendPolicy.sumPolicy, null);
		sum01.append(sum02, AppendPolicy.sumPolicy, null);
		assertEquals(expected, sum01);
	}

	/**
	 * 
	 * Test the minus method on trees
	 * 
	 */
	public void testMinus() {
		String xmlString = "<dtr:dnode variable='4'>"
				+ "  <dtr:leaf value='0' data='2'/>"
				+ "  <dtr:leaf value='1' data='3'/>" + "</dtr:dnode>";
		Tree<Double> sum01 = new Tree<Double>(getRoot(xmlString),
				DataLoadingPolicy.doubleLoadingPolicy);

		xmlString = "<dtr:dnode variable='2'>" + "  <dtr:leaf value='0' data='1'/>"
				+ "  <dtr:leaf value='1' data='2'/>" + "</dtr:dnode>";
		Tree<Double> sum02 = new Tree<Double>(getRoot(xmlString),
				DataLoadingPolicy.doubleLoadingPolicy);

		xmlString = "<dtr:dnode variable='4'>"
				+ "  <dtr:dnode value='0' variable='2'>"
				+ "    <dtr:leaf value='0' data='1'/>"
				+ "    <dtr:leaf value='1' data='0'/>" + "  </dtr:dnode>"
				+ "  <dtr:dnode value='1' variable='2'>"
				+ "    <dtr:leaf value='0' data='2'/>"
				+ "    <dtr:leaf value='1' data='1'/>" + "  </dtr:dnode>"
				+ "</dtr:dnode>";
		Tree<Double> expected = new Tree<Double>(getRoot(xmlString),
				DataLoadingPolicy.doubleLoadingPolicy);

		sum01.append(sum02, AppendPolicy.minusPolicy, null);
		assertEquals(expected, sum01);
	}

	/**
	 * 
	 * Test the simplify method with tests list on trees
	 * 
	 */
	public void testSimplifyWithConstraints() {
		String xmlRef = "<dtr:dnode variable='4'>"
				+ "  <dtr:leaf value='0' data='Hello'/>"
				+ "  <dtr:dnode value='1' variable='3'>"
				+ "    <dtr:leaf value='0' data='world'/>"
				+ "    <dtr:leaf value='1' data='and'/>"
				+ "    <dtr:leaf value='2' data='everybody'/>" + "  </dtr:dnode>"
				+ "</dtr:dnode>";

		Context testsList = new Context();
		testsList.put(Variable.getVariable("4"), "1");
		Tree<String> simplifiedTree = new Tree<String>(getRoot(xmlRef),
				DataLoadingPolicy.stringLoadingPolicy);
		simplifiedTree.cut(testsList);
		String xmlString = "<dtr:dnode value='1' variable='3'>"
				+ "  <dtr:leaf value='0' data='world'/>"
				+ "  <dtr:leaf value='1' data='and'/>"
				+ "  <dtr:leaf value='2' data='everybody'/>" + "</dtr:dnode>";
		assertEquals(getTree(xmlString).toString(), simplifiedTree.toString());

		testsList.clear();
		testsList.put(Variable.getVariable("3"), "0");
		simplifiedTree = new Tree<String>(getRoot(xmlRef),
				DataLoadingPolicy.stringLoadingPolicy);
		simplifiedTree.cut(testsList);
		xmlString = "<dtr:dnode variable='4'>"
				+ "  <dtr:leaf value='0' data='Hello'/>"
				+ "  <dtr:leaf value='1' data='world'/>" + "</dtr:dnode>";
		assertEquals(getTree(xmlString).toString(), simplifiedTree.toString());

		xmlRef = "<dtr:dnode variable='4'>"
				+ "  <dtr:leaf value='0' data='Hello'/>"
				+ "  <dtr:dnode value='1' variable='3'>"
				+ "    <dtr:leaf value='0' data='world'/>"
				+ "    <dtr:dnode value ='1' variable='2'>"
				+ "      <dtr:leaf value='1' data='and'/>" + "    </dtr:dnode>"
				+ "    <dtr:leaf value='2' data='everybody'/>" + "  </dtr:dnode>"
				+ "</dtr:dnode>";
		testsList.clear();
		testsList.put(Variable.getVariable("3"), "1");
		simplifiedTree = new Tree<String>(getRoot(xmlRef),
				DataLoadingPolicy.stringLoadingPolicy);
		simplifiedTree.cut(testsList);
		xmlString = "<dtr:dnode variable='4'>"
				+ "  <dtr:leaf value='0' data='Hello'/>"
				+ "  <dtr:leaf value='1' data='and'/>" + "</dtr:dnode>";
		assertEquals(getTree(xmlString).toString(), simplifiedTree.toString());
	}

	/**
	 * 
	 * Test the convert method on trees
	 * 
	 */
	public void testConvert() {
		String xmlString = "<dtr:dnode variable='4'>"
				+ "  <dtr:leaf value='0' data='3.4'/>"
				+ "  <dtr:dnode value='1' variable='3'>"
				+ "    <dtr:leaf value='0' data='2.1'/>"
				+ "    <dtr:dnode value ='1' variable='2'>"
				+ "      <dtr:leaf value='1' data='2.5'/>" + "    </dtr:dnode>"
				+ "    <dtr:leaf value='2' data='-0.5'/>" + "  </dtr:dnode>"
				+ "</dtr:dnode>";
		Tree<String> refTree = getTree(xmlString);
		Tree<Double> convertedTree = new Tree<Double>();
		new StringToDoublePolicy().convert(refTree, convertedTree);
		assertEquals(refTree.toString(), convertedTree.toString());

		Context testsList = new Context();
		testsList.put(Variable.getVariable("3"), "1");

		new StringToDoublePolicy().convert(refTree, convertedTree, testsList);
		xmlString = "<dtr:dnode variable='4'>"
				+ "  <dtr:leaf value='0' data='3.4'/>"
				+ "  <dtr:leaf value='1' data='2.5'/>" + "</dtr:dnode>";
		assertEquals(getTree(xmlString).toString(), convertedTree.toString());

		xmlString = "<dtr:dnode variable='4'>"
				+ "  <dtr:dnode value='1' variable='3'>"
				+ "    <dtr:dnode value ='0' variable='2'>"
				+ "      <dtr:leaf value='1' data='10'/>"
				+ "      <dtr:leaf value='2' data='20'/>"
				+ "      <dtr:leaf value='3' data='30'/>" + "    </dtr:dnode>"
				+ "    <dtr:dnode value ='1' variable='2'>"
				+ "      <dtr:leaf value='1' data='11'/>"
				+ "      <dtr:leaf value='2' data='21'/>"
				+ "      <dtr:leaf value='3' data='31'/>" + "    </dtr:dnode>"
				+ "  </dtr:dnode>" + "</dtr:dnode>";
		refTree = getTree(xmlString);
		testsList.clear();
		testsList.put(Variable.getVariable("2"), "1");
		new StringToDoublePolicy().convert(refTree, convertedTree, testsList);
		xmlString = "<dtr:dnode variable='3'>"
				+ "  <dtr:leaf value='0' data='10.0'/>"
				+ "  <dtr:leaf value='1' data='11.0'/>" + "</dtr:dnode>";
		assertEquals(getTree(xmlString).toString(), convertedTree.toString());

		String stringValuesSet = "<dtr:variablevalueset>"
				+ "  <dtr:valueset variable='0'>" + "    <dtr:value value='0'/>"
				+ "    <dtr:value value='1'/>" + "  </dtr:valueset>"
				+ "  <dtr:valueset variable='1'>" + "    <dtr:value value='0'/>"
				+ "    <dtr:value value='1'/>" + "  </dtr:valueset>"
				+ "</dtr:variablevalueset>";
		Map<Variable, Set<String>> valuesSet = new LinkedHashMap<Variable, Set<String>>();
		Context.loadValueSet(valuesSet, XmlUtils.getRoot(stringValuesSet));
		xmlString = "<dtr:dnode variable='0'>"
				+ "  <dtr:dnode value='1' variable='1'>"
				+ "    <dtr:leaf value='0' data='10'/>" + "  </dtr:dnode>"
				+ "</dtr:dnode>";
		refTree = getTree(xmlString);
		new StringToDoublePolicy().convert(valuesSet, refTree, convertedTree);
		xmlString = "<dtr:dnode variable='0'>"
				+ "  <dtr:dnode value='1' variable='1'>"
				+ "    <dtr:leaf value='0' data='10.0'/>"
				+ "    <dtr:leaf value='1' data='-55.0'/>" + "  </dtr:dnode>"
				+ "  <dtr:leaf value='0' data='-55.0'/>" + "</dtr:dnode>";
		assertEquals(getTree(xmlString).toString(), convertedTree.toString());
	}
}
