/*
 * 
 * Created on 21 sept. 2005
 *
 */
package animatlab.kodamat.core.utils;

import java.util.LinkedHashMap;
import java.util.Map;

/**
 * 
 * Class to measure times
 * 
 * @author Thomas Degris
 * 
 */
public class Chrono {
	/**
	 * 
	 * Set of time points
	 * 
	 */
	private Map<String, Double> timePoints = new LinkedHashMap<String, Double>();

	/**
	 * 
	 * Date of creation
	 * 
	 */
	private long creationTime;

	/**
	 * 
	 * Constructor
	 * 
	 */
	public Chrono() {
		start();
	}

	/**
	 * 
	 * Start the chrono
	 * 
	 */
	public void start() {
		creationTime = System.currentTimeMillis();
	}

	/**
	 * 
	 * @return get the current chrono time (in second)
	 * 
	 */
	final public double getCurrentMillis() {
		return System.currentTimeMillis() - creationTime;
	}

	/**
	 * 
	 * @return get the current chrono time (in second)
	 * 
	 */
	public double getCurrentChrono() {
		return (getCurrentMillis() / 1000.0);
	}

	/**
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		return toStringMillis(getCurrentMillis());
	}

	/**
	 * 
	 * Print the value a of chrono into a string
	 * 
	 * @param millisTime
	 *          time to print out to a string (ms)
	 * 
	 * @return the string representing the chrono
	 * 
	 */
	public static String toStringMillis(double millisTime) {
		if (millisTime < 1000) {
			return String.valueOf(millisTime) + "ms";
		}
		int seconds = (int) (millisTime / 1000.0);
		int hours = seconds / 3600;
		seconds = seconds - hours * 3600;
		int minutes = seconds / 60;
		seconds = seconds - minutes * 60;
		StringBuilder result = new StringBuilder();
		if (hours != 0) {
			result.append(hours + "h");
		}
		if (minutes != 0) {
			result.append(minutes + "m");
		}
		result.append(seconds + "s");
		return result.toString();
	}

	/**
	 * 
	 * Add a time point to this
	 * 
	 * @param name
	 *          name of the time point
	 * 
	 */
	public void addTimePoint(String name) {
		timePoints.put(name, getCurrentChrono());
	}

	/**
	 * 
	 * Get the time associated to a time
	 * 
	 * @param name
	 *          name of the time point
	 * 
	 * @return the time associated to the time point with the name 'name'
	 * 
	 */
	public Double getTimePoint(String name) {
		return timePoints.get(name);
	}

	/**
	 * 
	 * Clear all the time point
	 * 
	 */
	public void clearTimePoints() {
		timePoints.clear();
	}
}
