package animatlab.kodamat.experiments.main;

import java.util.List;

import animatlab.kodamat.experiments.protocol.AgentOverChi2Exp;
import animatlab.kodamat.experiments.protocol.AgentOverSizeExp;
import animatlab.kodamat.experiments.protocol.AgentOverTimeExp;
import animatlab.kodamat.experiments.protocol.Experiment;
import animatlab.kodamat.experiments.protocol.PoolRunner;
import animatlab.kodamat.reinforcementlearning.core.stats.Measure;
import animatlab.kodamat.reinforcementlearning.planning.guestrin.LearningGuestrinPlanning;
import animatlab.kodamat.reinforcementlearning.planning.spisvi.LearningSpiSvi;
import animatlab.kodamat.reinforcementlearning.planning.spudd.LearningSpudd;
import animatlab.kodamat.reinforcementlearning.planning.stats.Chi2ModelMeasure;
import animatlab.kodamat.reinforcementlearning.planning.stats.FunctionSizeMeasure;
import animatlab.kodamat.reinforcementlearning.planning.stats.RelativeErrorMeasure;

/**
 * 
 * Run the experiments for UAI 2006
 * 
 * @author Thomas Degris
 * 
 */
public class These extends PoolRunner {
	/**
	 * 
	 * Constructor
	 * 
	 */
	public These() {
		super("these");
	}

	/**
	 * 
	 * Main method
	 * 
	 * @param args
	 *          list of arguments passed to the program
	 * 
	 */
	public static void main(String[] args) {
		new These().runPools(args);
	}

	/**
	 * @see animatlab.kodamat.experiments.protocol.PoolRunner#buildExperimentList(List)
	 */
	@Override
	protected void buildExperimentList(List<Experiment> experiments) {
		chapter4(experiments);
		chapter5(experiments);
		chapter6(experiments);
	}

	/**
	 * 
	 * Add experiments of the chapter 6
	 * 
	 * @param experiments
	 *          the list of experiments to fill
	 * 
	 */
	private void chapter6(List<Experiment> experiments) {
		experiments.add(AgentOverSizeExp.addIncLearningExp("bestcase", 3, 10,
				20000, new Chi2ModelMeasure(), new RelativeErrorMeasure()));
		experiments.add(AgentOverSizeExp.addIncLearningExp("worstcase", 3, 10,
				20000, new Chi2ModelMeasure(), new RelativeErrorMeasure()));
		// experiments.add(AgentOverSizeExp.addIncLearningExpParamExplo(
		// "bestcase", 3, 10, 20000,
		// new Chi2ModelMeasure(), new RelativeErrorMeasure()));
		// experiments.add(AgentOverSizeExp.addIncLearningExpParamExplo(
		// "worstcase", 3, 10, 20000,
		// new Chi2ModelMeasure(), new RelativeErrorMeasure()));
		//		
		// experiments.add(AgentOverTimeExp.getFIEOverTime("coffeerobot", 4000, 1,
		// 0.01, 0.01));
		// experiments.add(AgentOverTimeExp.getFIEOverTime("coffeerobot", 4000, 1,
		// 0.0, 0.0));
		// experiments.add(AgentOverTimeExp.getFIEOverTime("coffeerobot", 4000, 1,
		// 1.0, 1.0));
		//		
		// experiments.add(AgentOverTimeExp.getFIEOverTime("factory", 20000, 1,
		// 0.01, 0.01));
		// experiments.add(AgentOverTimeExp.getFIEOverTime("factory", 20000, 1, 0.0,
		// 0.0));
		// experiments.add(AgentOverTimeExp.getFIEOverTime("factory", 20000, 100,
		// 0.0, 0.0));
	}

	/**
	 * 
	 * Add experiments of the chapter 5
	 * 
	 * @param experiments
	 *          the list of experiments to fill
	 * 
	 */
	private void chapter5(List<Experiment> experiments) {
		experiments.add(AgentOverTimeExp.getEGreedyOverTime("coffeerobot",
				new LearningSpiSvi(), 10000));
		experiments.add(AgentOverTimeExp.getEGreedyOverTime("factory",
				new LearningSpiSvi(), 20000));
		experiments.add(AgentOverTimeExp.getEGreedyOverTime("factory4",
				new LearningSpudd(), 10000));
		experiments.add(AgentOverTimeExp.getEGreedyOverTime("unidirectionalring40",
				new LearningGuestrinPlanning(), 20000));
	}

	/**
	 * 
	 * Add experiments of the chapter 4
	 * 
	 * @param experiments
	 *          the list of experiments to fill
	 * 
	 */
	private void chapter4(List<Experiment> experiments) {
		experiments.add(AgentOverChi2Exp.addBatchChi2Exp(
				"coffeerobot, coffeerobot.noisy", 500, Measure
						.measurePoints(0, 20, 0.5), new FunctionSizeMeasure(),
				new Chi2ModelMeasure(), new RelativeErrorMeasure()));
		experiments.add(AgentOverChi2Exp.addBatchChi2Exp(
				"coffeerobot, coffeerobot.noisy", 4000, Measure.measurePoints(0, 20,
						0.5), new FunctionSizeMeasure(), new Chi2ModelMeasure(),
				new RelativeErrorMeasure()));

		experiments.add(AgentOverSizeExp.addBatchLearningExp(
				"bestcase, bestcasenoise20", 3, 20, 20000, new FunctionSizeMeasure(),
				new Chi2ModelMeasure(), new RelativeErrorMeasure()));
		experiments.add(AgentOverSizeExp.addBatchLearningExp(
				"worstcase, worstcasenoise20", 3, 20, 20000, new FunctionSizeMeasure(),
				new Chi2ModelMeasure(), new RelativeErrorMeasure()));

		experiments.add(AgentOverTimeExp.addBatchPlanningExp("factory", 10000));
		experiments.add(AgentOverTimeExp.addBatchPlanningExp("factory4", 20000));
		experiments.add(AgentOverTimeExp.addBatchPlanningExp(
				"unidirectionalring40", 20000));
	}
}
