package animatlab.kodamat.experiments.main;

import java.util.Random;

import animatlab.kodamat.core.io.Exec;
import animatlab.kodamat.core.io.Log;
import animatlab.kodamat.core.utils.Chrono;
import animatlab.kodamat.core.utils.Parameters;
import animatlab.kodamat.experiments.protocol.Experiment;
import animatlab.kodamat.problems.ProblemRunner;
import animatlab.kodamat.reinforcementlearning.agent.ModelBasedAgentFactory;
import animatlab.kodamat.reinforcementlearning.core.acting.RandomActing;
import animatlab.kodamat.reinforcementlearning.core.stats.Measure;
import animatlab.kodamat.reinforcementlearning.core.stats.MeasureWriter;
import animatlab.kodamat.reinforcementlearning.model.forest.StaticForest;
import animatlab.kodamat.reinforcementlearning.model.forest.VariableLearningForest;
import animatlab.kodamat.reinforcementlearning.planning.spisvi.SpiSvi;
import animatlab.kodamat.reinforcementlearning.planning.stats.FunctionSizeMeasure;
import animatlab.kodamat.reinforcementlearning.planning.stats.OverChi2Measure;
import animatlab.kodamat.supervisedlearning.entropy.Chi2Measure;
import animatlab.kodamat.supervisedlearning.learningtree.ID3Tree;
import animatlab.kodamat.supervisedlearning.regression.LeastSquareMeasure;

/**
 * 
 * Thesis experiments without stats
 * 
 * @author Thomas Degris
 * 
 */
public class TheseNoPool {

	/**
	 * 
	 * Main method
	 * 
	 * @param args
	 *          list of arguments passed to the program
	 * 
	 */
	public static void main(String[] args) {
		// System.out.println("coffeerobot :");
		// testReorganization(Experiment.loadForest("coffeerobot"));
		// StaticForest factoryForest = Experiment.loadForest("factory");
		// factoryForest.getParameters().setKeyValue(Parameters.PARAM_KEY_THRESHOLD_CONVERGENCE,
		// "1.0");
		// System.out.println("factory :");
		// testReorganization(factoryForest);
		runAgentOverTimeOverChi2();
	}

	/**
	 * 
	 * Test the reorganization algorithm
	 * 
	 * @param staticForest
	 *          the problem to use for the test
	 * 
	 */
	protected static void testReorganization(StaticForest staticForest) {
		final int numberOfTry = 10;
		final int ignoreValue = 2;

		double sumNoOrderingTime = 0;
		double sumOrderingTime = 0;
		int sumNoOrderingValFunSize = 0;
		int sumOrderingValFunSize = 0;

		for (int tryIndex = 0; tryIndex < numberOfTry + ignoreValue; tryIndex++) {
			System.out.print(tryIndex + ": ");
			System.out.flush();
			staticForest.getParameters().setKeyValue(Parameters.PARAM_KEY_ORDERING,
					"false");
			SpiSvi spisvi = new SpiSvi(staticForest);
			spisvi.setLog(new Log("/dev/null"));
			Chrono chrono = new Chrono();
			spisvi.getSVI();

			double noOrderingTime = chrono.getCurrentMillis();
			int noOrderingValFunSize = spisvi.getOptimalValueFunction()
					.numberOfNodes();
			printResults("noordering: ", noOrderingTime, noOrderingValFunSize);

			staticForest.getParameters().setKeyValue(Parameters.PARAM_KEY_ORDERING,
					"true");
			spisvi = new SpiSvi(staticForest);
			spisvi.setLog(new Log("/dev/null"));
			chrono = new Chrono();
			spisvi.getSVI();

			double orderingTime = chrono.getCurrentMillis();
			int orderingValFunSize = spisvi.getOptimalValueFunction().numberOfNodes();
			printResults("ordering: ", orderingTime, orderingValFunSize);
			if (tryIndex > (ignoreValue - 1)) {
				sumNoOrderingTime += noOrderingTime;
				sumNoOrderingValFunSize += noOrderingValFunSize;
				sumOrderingTime += orderingTime;
				sumOrderingValFunSize += orderingValFunSize;
			} else {
				System.out.print("ignored ");
			}
			System.out.println();
		}
		System.out.print("ave: ");
		printResults("noordering: ", sumNoOrderingTime / numberOfTry,
				sumNoOrderingValFunSize / numberOfTry);
		printResults("ordering: ", sumOrderingTime / numberOfTry,
				sumOrderingValFunSize / numberOfTry);
		System.out.println();
	}

	/**
	 * 
	 * Print data
	 * 
	 * @param prefix
	 *          prefix to add to the print out
	 * @param time
	 *          computation time
	 * @param size
	 *          optimal value function size
	 * 
	 */
	private static void printResults(String prefix, double time, int size) {
		System.out.print(prefix + ": " + Chrono.toStringMillis(time) + " " + size
				+ " nodes ");
		System.out.flush();
	}

	/**
	 * 
	 * Run the agent over time and over chi2 (for 3D data)
	 * 
	 * 
	 */
	protected static void runAgentOverTimeOverChi2() {
		final double[] timeMeasurePoints = Measure.measurePoints(0, 4000, 100);
		double[] chi2MeasurePoints = Measure.measurePoints(0, 1000, 25);
		chi2MeasurePoints = Measure.addMesurePoints(chi2MeasurePoints, 1.0, 2.0,
				3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0, 10.0);

		Random random = new Random(0);
		StaticForest staticForest = Experiment.loadForest("coffeerobot");

		OverChi2Measure measure = new OverChi2Measure();
		measure.setMeasureData(staticForest, null, null, "./results");
		measure.setChi2MeasurePoints(chi2MeasurePoints);
		measure.addMeasure(new FunctionSizeMeasure());
		ID3Tree.Prob<String> inductionTree = new ID3Tree.Prob<String>(
				new Chi2Measure(), staticForest.getVariableOrdering());
		ID3Tree.Plain<Double> regressionTree = new ID3Tree.Plain<Double>(
				new LeastSquareMeasure(null), staticForest.getVariableOrdering());
		measure.addAgentFactory(new ModelBasedAgentFactory(null,
				new VariableLearningForest(null, null, inductionTree, regressionTree),
				null));
		MeasureWriter measureWriter = new MeasureWriter(measure);
		measureWriter.setMeasurePoints(timeMeasurePoints);
		measureWriter.reset();
		ProblemRunner problemRunner = new ProblemRunner(random, staticForest,
				new RandomActing(random, staticForest.getActionList()), Log.sysout);
		problemRunner.registerListener(measure);
		problemRunner.registerListener(measureWriter);
		problemRunner
				.runProblem((int) timeMeasurePoints[timeMeasurePoints.length - 1]);

		Exec.exec("gnuplot src/gnuplot/chi2ontimeonsize.gnuplot", Log.sysout);
	}

}
