/*
 * 
 * Created on 20 sept. 2005
 *
 */
package animatlab.kodamat.experiments.protocol;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;

import animatlab.kodamat.core.io.Log;
import animatlab.kodamat.core.utils.Chrono;
import animatlab.kodamat.problems.generated.ForestGeneration;
import animatlab.kodamat.problems.generated.syntheticmdps.LinearProblem;
import animatlab.kodamat.reinforcementlearning.planning.spisvi.SpiSvi;

/**
 * 
 * Measure the performance of the SPI algorithm for time and space on different
 * worst case and best case generated problem
 * 
 * @author Thomas Degris
 * 
 */
public class SpiPerformance {

	/**
	 * 
	 * Data structure of the different measures for one problem size
	 * 
	 */
	static class PerfomanceMeasured {
		/**
		 * 
		 * Time needed to compute the optimal policy
		 * 
		 */
		double computationTime;

		/**
		 * 
		 * Number of leaf of the value function
		 * 
		 */
		double numberOfLeaf;
	}

	/**
	 * 
	 * First problem size tested
	 * 
	 */
	static final int FIRST_PROBLEM_SIZE = 1;

	/**
	 * 
	 * Constructor
	 * 
	 */
	public SpiPerformance() {
		super();
	}

	/**
	 * 
	 * Main method
	 * 
	 * @param args
	 *          list of arguments passed to the program
	 * 
	 */
	public static void main(String[] args) {
		getSpiPerformance(new LinearProblem());
	}

	/**
	 * 
	 * Compute the performance of SPI on a generated problem of different size
	 * 
	 * @param problemGeneration
	 *          instance used to generated the problem
	 * 
	 */
	private static void getSpiPerformance(ForestGeneration problemGeneration) {
		PerfomanceMeasured[] result = new PerfomanceMeasured[20];

		String folderName = "./results";
		new File(folderName).mkdirs();
		Log log = new Log(folderName + "/spiperformance.log");
		String fileName = folderName + "/" + problemGeneration.getProblemName()
				+ ".perf";
		log.out().println("Logging to " + fileName + "...");
		FileWriter output = null;

		try {
			output = new FileWriter(fileName);
			for (int indexSize = 0; indexSize < result.length; indexSize++) {
				result[indexSize] = getPerformanceTime(problemGeneration, indexSize
						+ FIRST_PROBLEM_SIZE);
				output.write(String.valueOf(indexSize + FIRST_PROBLEM_SIZE) + " "
						+ result[indexSize].computationTime + " "
						+ result[indexSize].numberOfLeaf + "\n");
				output.flush();
			}
		} catch (IOException e) {
			e.printStackTrace();
			System.exit(1);
		}
	}

	/**
	 * 
	 * Get the performance of SPI for a generated problem and for a particular
	 * size
	 * 
	 * @param problemGeneration
	 *          the problem to generate
	 * @param problemSize
	 *          size of the problem to measure the performance with
	 * 
	 * @return the performance of SPI for the generated problem and for the
	 *         specific size
	 * 
	 */
	private static PerfomanceMeasured getPerformanceTime(
			ForestGeneration problemGeneration, int problemSize) {
		SpiSvi spisvi = new SpiSvi(problemGeneration.toForest());
		PerfomanceMeasured measures = new PerfomanceMeasured();
		Chrono chrono = new Chrono();
		spisvi.getSVI();
		measures.computationTime = chrono.getCurrentChrono();
		measures.numberOfLeaf = spisvi.getOptimalValueFunction().numberOfLeaves();
		return measures;
	}

}
