/*
 * 
 * Created on 3 nov. 2005
 *
 */
package animatlab.kodamat.experiments.test;

import java.io.File;
import java.util.List;

import junit.framework.TestCase;
import animatlab.kodamat.core.io.Log;
import animatlab.kodamat.core.utils.Parameters;
import animatlab.kodamat.experiments.protocol.AgentOverChi2Exp;
import animatlab.kodamat.experiments.protocol.AgentOverSizeExp;
import animatlab.kodamat.experiments.protocol.AgentOverTimeExp;
import animatlab.kodamat.experiments.protocol.Experiment;
import animatlab.kodamat.experiments.protocol.PoolRunner;
import animatlab.kodamat.reinforcementlearning.core.stats.Measure;
import animatlab.kodamat.reinforcementlearning.planning.guestrin.StaticGuestrin;
import animatlab.kodamat.reinforcementlearning.planning.spisvi.LearningSpiSvi;
import animatlab.kodamat.reinforcementlearning.planning.spisvi.StaticSpiSvi;
import animatlab.kodamat.reinforcementlearning.planning.stats.Chi2ModelMeasure;
import animatlab.kodamat.reinforcementlearning.planning.stats.FunctionSizeMeasure;
import animatlab.kodamat.reinforcementlearning.planning.stats.RelativeErrorMeasure;

/**
 * 
 * Test the SpiExperiment class
 * 
 * @author Thomas Degris
 * 
 */
public class TestExperiment extends TestCase {

	/**
	 * 
	 * Class used to test the Pool runner
	 * 
	 */
	private class TestPoolRunner extends PoolRunner {
		/**
		 * 
		 * Name of the study used ofr JUnit
		 * 
		 */
		static final String STUDY_NAME = "junit";

		/**
		 * 
		 * Constructor
		 * 
		 */
		public TestPoolRunner() {
			super("junit");
			numberOfExperiment = 2;
		}

		/**
		 * @see animatlab.kodamat.experiments.protocol.PoolRunner#buildExperimentList(List)
		 */
		@Override
		protected void buildExperimentList(List<Experiment> experiments) {

			experiments.add(AgentOverTimeExp.addBatchPlanningExp(
					"unidirectionalring02", 50));
			experiments.add(AgentOverSizeExp.addBatchLearningExp(
					"bestcase, bestcase20", 0, 4, 20000, new FunctionSizeMeasure(),
					new Chi2ModelMeasure(), new RelativeErrorMeasure()));
			experiments.add(AgentOverChi2Exp.addBatchChi2Exp("coffeerobot", 1000,
					Measure.measurePoints(10, 50, 10), new FunctionSizeMeasure(),
					new Chi2ModelMeasure(), new RelativeErrorMeasure()));
			experiments.add(AgentOverTimeExp.getEGreedyOverTime("coffeerobot",
					new LearningSpiSvi(), 500));
		}
	}

	/**
	 * 
	 * Test the forest to agent conversion with Guestrin
	 * 
	 */
	public void testConvertForestToAgent() {
		String agentFolder = "./.reports";
		checkProblem(agentFolder, "coffeerobot", StaticSpiSvi.AGENT_EXTENSION);
		checkProblem(agentFolder, "coffeerobot.noisy", StaticSpiSvi.AGENT_EXTENSION);
		checkProblem(agentFolder, "unidirectionalring06",
				StaticGuestrin.AGENT_EXTENSION);
	}

	/**
	 * 
	 * Check the conversion for a problem
	 * 
	 * @param agentFolder
	 *          the folder where the agent is saved
	 * @param problemName
	 *          name of the problem to test
	 * @param planningExtension
	 *          extension of the preferred planning method
	 * 
	 */
	private void checkProblem(String agentFolder, String problemName,
			String planningExtension) {
		Experiment.forestToAgent(problemName, agentFolder);
		assertTrue(new File(agentFolder + "/" + problemName + planningExtension)
				.canRead());
	}

	/**
	 * 
	 * Test all the availables experiment
	 * 
	 */
	public void testExperiment() {
		Log.removeRecursively(Parameters.OUTPUT_JUNIT_FOLDER + "/network");
		String[] args = { Parameters.OUTPUT_JUNIT_FOLDER + "/network",
				Parameters.OUTPUT_JUNIT_FOLDER + "/local" };
		new TestPoolRunner().runPools(args);
		new TestPoolRunner().runPools(args);
	}
}
