package animatlab.kodamat.problems.generated;

import static animatlab.kodamat.problems.generated.syntheticmdps.ExponProblem.WORSTCASE_PROBLEM_NAME;
import static animatlab.kodamat.problems.generated.syntheticmdps.LinearProblem.BESTCASE_PROBLEM_NAME;
import static animatlab.kodamat.problems.generated.syntheticmdps.NoisyLinearProblem.NOISYCASE_PROBLEM_NAME;
import static animatlab.kodamat.problems.generated.sysadmin.FullyConnected.FULLY_CONNECTED_PROBLEM_NAME;
import static animatlab.kodamat.problems.generated.sysadmin.UnidirectionalRing.UNIDIRECTIONAL_RING_PROBLEM_NAME;

import java.util.ArrayList;
import java.util.List;

import animatlab.kodamat.core.functions.FunctionsSum;
import animatlab.kodamat.core.proba.IProbabilities;
import animatlab.kodamat.core.trees.Leaf;
import animatlab.kodamat.core.trees.Tree;
import animatlab.kodamat.core.utils.Variable;
import animatlab.kodamat.core.utils.VariableOrdering;
import animatlab.kodamat.problems.generated.syntheticmdps.ExponProblem;
import animatlab.kodamat.problems.generated.syntheticmdps.LinearProblem;
import animatlab.kodamat.problems.generated.syntheticmdps.NoisyLinearProblem;
import animatlab.kodamat.problems.generated.sysadmin.FullyConnected;
import animatlab.kodamat.problems.generated.sysadmin.UnidirectionalRing;
import animatlab.kodamat.reinforcementlearning.model.StaticCPD;
import animatlab.kodamat.reinforcementlearning.model.forest.StaticForest;

/**
 * 
 * Helper class to easily generate problems definition
 * 
 * @author Thomas Degris
 * 
 */
public abstract class ForestGeneration {
	/**
	 * 
	 * Size of the problem to generate
	 * 
	 */
	protected int problemSize;

	/**
	 * 
	 * Name of the problem generated
	 * 
	 */
	private final String problemName;

	/**
	 * 
	 * Noise in the problem to pass from one state to another
	 * 
	 */
	protected double noise = 0;

	/**
	 * 
	 * String to indicate the noise when not null
	 * 
	 */
	public static final String NOISE_STRING = "noise";

	/**
	 * 
	 * Constructor
	 * 
	 * @param problemName
	 *          name of the problem to generate
	 * 
	 */
	protected ForestGeneration(String problemName) {
		this.problemName = problemName;
	}

	/**
	 * 
	 * @return the name of the problem generated in the xml file
	 * 
	 */
	public String getProblemName() {
		return problemName;
	}

	/**
	 * 
	 * Set all the problem data to be generated
	 * 
	 * @param problemSize
	 *          size of the problem
	 * 
	 */
	public void setProblemData(int problemSize) {
		this.problemSize = problemSize;
	}

	/**
	 * 
	 * Set all the problem data to be generated
	 * 
	 * @param noise
	 *          level of noise in the problem
	 * 
	 */
	public void setNoise(double noise) {
		this.noise = noise;
	}

	/**
	 * 
	 * @return the legend of the forest
	 * 
	 */
	private VariableOrdering getVariableOrdering() {
		return VariableOrdering.buildNumericVariableOrdering(problemSize);
	}

	/**
	 * 
	 * @return the action set of the forest
	 * 
	 */
	protected List<String> getActionList() {
		List<String> result = new ArrayList<String>();
		for (int actionIndex = 0; actionIndex < problemSize; actionIndex++) {
			result.add(extractActionName(actionIndex));
		}
		return result;
	}

	/**
	 * 
	 * Build a set of basis function for the forest
	 * 
	 * @param forest
	 *          the forest to generate
	 * 
	 * @return the set of created basis function
	 * 
	 */
	protected FunctionsSum getBasisFunction(StaticForest forest) {
		return null;
	}

	/**
	 * 
	 * Build the set of terminal conditions
	 * 
	 * @param forest
	 *          the forest to generate
	 * 
	 */
	protected void buildTerminalCondition(StaticForest forest) {
		// No terminal condition by default
	}

	/**
	 * 
	 * Build the set of initial conditions
	 * 
	 * @param forest
	 *          the forest to generate
	 * 
	 */
	protected void buildInitialCondition(StaticForest forest) {
		// No initial condition by default
	}

	/**
	 * 
	 * Build the set of reward functions
	 * 
	 * @param forest
	 *          the forest to generate
	 * 
	 */
	abstract protected void buildRewardFunctions(StaticForest forest);

	/**
	 * Build a set of transitions functions
	 * 
	 * @param forest
	 *          the forest to generate
	 * 
	 */
	protected void buildTransitionFunction(StaticForest forest) {
		for (String action : forest.getActionList()) {
			buildActionVariableTransition(forest, action);
		}
	}

	/**
	 * 
	 * The name of action from its index
	 * 
	 * @param actionIndex
	 *          the index of the action
	 * 
	 * @return the name of the action
	 * 
	 */
	protected String extractActionName(int actionIndex) {
		return String.valueOf(actionIndex);
	}

	/**
	 * 
	 * The name of action from its index
	 * 
	 * @param actionIndex
	 *          the index of the action
	 * 
	 * @return the name of the action
	 * 
	 */
	public final String extractActionName(String actionIndex) {
		return extractActionName(Integer.parseInt(actionIndex));
	}

	/**
	 * 
	 * Build a CPT tree for an action and one variable
	 * 
	 * @param forest
	 *          the forest to generate
	 * @param action
	 *          the action to build the tree for
	 * @param variable
	 *          the index of the variable to build the tree for
	 * @return the tree representing the conditional distribution probability
	 * 
	 */
	abstract protected Tree<IProbabilities<String>> getVariableActionCptNode(
			StaticForest forest, String action, Variable variable);

	/**
	 * 
	 * Get the transitions for all variable for a given action
	 * 
	 * @param forest
	 *          the forest to generate
	 * @param action
	 *          the action to build the tree
	 * 
	 */
	protected void buildActionVariableTransition(StaticForest forest,
			String action) {
		for (int variableIndex = 0; variableIndex < problemSize; variableIndex++) {
			Variable variable = getVariable(variableIndex);
			forest.setTransitionModel(action, variable, getVariableActionCptNode(
					forest, action, variable));
		}
	}

	/**
	 * 
	 * Extract the problem name from a problem description
	 * 
	 * @param problemDescription
	 *          the problem description
	 * 
	 * @return the name of the problem
	 * 
	 */
	public static String getProblemName(String problemDescription) {
		StringBuffer name = new StringBuffer();
		int charIndex = 0;
		int noiseIndex = problemDescription.indexOf(NOISE_STRING);
		if (noiseIndex == -1) {
			noiseIndex = problemDescription.length();
		}
		while ((charIndex < noiseIndex)
				&& Character.isLetter(problemDescription.charAt(charIndex))) {
			name.append(problemDescription.charAt(charIndex));
			charIndex++;
		}
		return name.toString();
	}

	/**
	 * 
	 * Create a node in the tree specifying the variable stayed at the same value
	 * if a probability of 1
	 * 
	 * @param leaf
	 *          the leaf in the tree containing the deterministic probability
	 * @param value
	 *          the value of the variable that stays the same
	 * 
	 */
	protected void createUniqueLeafProbability(Leaf<IProbabilities<String>> leaf,
			String value) {
		StaticCPD proba = new StaticCPD();
		proba.setProba(value, 1);
		leaf.set(proba);
	}

	/**
	 * 
	 * Create a tree containing one leaf with a probability of one for value
	 * 
	 * @param forest
	 *          the forest to build
	 * @param value
	 *          the value contained at the leaf
	 * 
	 * @return a new tree
	 * 
	 */
	protected Tree<IProbabilities<String>> createLeafTree(StaticForest forest,
			String value) {
		Tree<IProbabilities<String>> tree = new Tree<IProbabilities<String>>();
		createUniqueLeafProbability((Leaf<IProbabilities<String>>) tree.getRoot(),
				value);
		return tree;
	}

	/**
	 * 
	 * @return file name of the problem
	 * 
	 */
	public String getProblemFileName() {
		return getProblemDescription() + StaticForest.PROBLEM_EXTENSION;
	}

	/**
	 * 
	 * @return the description of this problem
	 * 
	 */
	public String getProblemDescription() {
		StringBuffer result = new StringBuffer(getProblemName());
		result.append(String.format("%02d", problemSize));
		if (noise > 0) {
			result.append(String.format("noise%02d", Math.round(noise * 100)));
		}
		return result.toString();
	}

	/**
	 * 
	 * Extract the size of a problem from a problem description
	 * 
	 * @param problemDescription
	 *          the problem description
	 * @return the size of the problem
	 * 
	 */
	public static int getProblemSize(String problemDescription) {
		int startIndex = ForestGeneration.getProblemName(problemDescription)
				.length();
		int endStringIndex = startIndex;
		while (endStringIndex < problemDescription.length()
				&& Character.isDigit(problemDescription.charAt(endStringIndex))) {
			endStringIndex++;
		}
		String sizeString = problemDescription
				.substring(startIndex, endStringIndex);
		if (sizeString.length() == 0) {
			return 0;
		}
		return Integer.parseInt(sizeString);
	}

	/**
	 * 
	 * Get the variable associated to an index
	 * 
	 * @param index
	 *          the index of the variable to get
	 * @return the variable at a time t
	 * 
	 */
	static public Variable getVariable(int index) {
		return Variable.getVariable(String.valueOf(index));
	}

	/**
	 * 
	 * Get the variable associated to an index
	 * 
	 * @param variable
	 *          the variable to get the index
	 * @return the index of the variable
	 * 
	 */
	protected int getIndex(Variable variable) {
		return Integer.parseInt(variable.variable);
	}

	/**
	 * 
	 * Build the file name associated to a problem
	 * 
	 * @param problemDescription
	 *          the description of the problem
	 * @return the file name associated to a problem
	 * 
	 */
	public static String getFileName(String problemDescription) {
		ForestGeneration problem = createProblemGeneration(problemDescription);
		if (problem == null) {
			return problemDescription + StaticForest.PROBLEM_EXTENSION;
		}
		return problem.getProblemFileName();
	}

	/**
	 * 
	 * @return build a forest corresponding to the problem
	 * 
	 */
	public StaticForest toForest() {
		StaticForest forest = new StaticForest(getVariableOrdering(),
				getActionList());
		buildParameters(forest);
		buildRewardFunctions(forest);
		buildInitialCondition(forest);
		buildTerminalCondition(forest);
		buildTransitionFunction(forest);
		forest.setBasisFunctions(getBasisFunction(forest));
		return forest;
	}

	/**
	 * 
	 * Set the parameters of a function
	 * 
	 * @param forest
	 *          the forest to build
	 * 
	 */
	protected void buildParameters(StaticForest forest) {
		// No parameters by default
	}

	/**
	 * 
	 * The generated forest corresponding to the problem description
	 * 
	 * @param problemDescription
	 *          the description of the problem to generate
	 * @return a static forest corresponding to the problem description
	 * 
	 */
	public static StaticForest toForest(String problemDescription) {
		ForestGeneration forestGeneration = createProblemGeneration(problemDescription);
		if (forestGeneration == null) {
			return null;
		}
		return forestGeneration.toForest();
	}

	/**
	 * 
	 * Create an instance of a problem generation from a problem name
	 * 
	 * @param problemDescription
	 *          the description of the problem to create
	 * @return the instance to generate file with the definition of the problem
	 * 
	 */
	public static ForestGeneration createProblemGeneration(
			String problemDescription) {
		ForestGeneration result = null;
		String problemName = ForestGeneration.getProblemName(problemDescription);
		if (problemName.equals(BESTCASE_PROBLEM_NAME)) {
			result = new LinearProblem();
		} else if (problemName.equals(NOISYCASE_PROBLEM_NAME)) {
			result = new NoisyLinearProblem();
		} else if (problemName.equals(WORSTCASE_PROBLEM_NAME)) {
			result = new ExponProblem();
		} else if (problemName.equals(UNIDIRECTIONAL_RING_PROBLEM_NAME)) {
			result = new UnidirectionalRing();
		} else if (problemName.equals(FULLY_CONNECTED_PROBLEM_NAME)) {
			result = new FullyConnected();
		}

		if (result == null) {
			return null;
		}
		result.setProblemData(ForestGeneration.getProblemSize(problemDescription));
		double noise = 0.0;
		int noiseStringIndex = problemDescription
				.indexOf(ForestGeneration.NOISE_STRING);
		if (noiseStringIndex != -1) {
			noise = Double.parseDouble(problemDescription.substring(noiseStringIndex
					+ ForestGeneration.NOISE_STRING.length()));
		}
		if (noise > 1) {
			noise = noise / 100.0;
		}
		result.setNoise(noise);
		result.setOptions(problemDescription);
		return result;
	}

	/**
	 * 
	 * Set different options in the problem
	 * 
	 * @param problemDescription
	 *          description of the problem
	 * 
	 */
	protected void setOptions(String problemDescription) {
		// Nothing to customize by default
	}

	/**
	 * 
	 * @return returns the noise of the problem
	 * 
	 */
	public double getNoise() {
		return noise;
	}

	/**
	 * 
	 * @return Returns the size of problem
	 * 
	 */
	public int getProblemSize() {
		return problemSize;
	}

}
