/*
 * 
 * Created on 8 nov. 2005
 *
 */
package animatlab.kodamat.problems.generated.syntheticmdps;

import animatlab.kodamat.core.conditions.AndCondition;
import animatlab.kodamat.core.conditions.OperatorCondition;
import animatlab.kodamat.core.conditions.VariableCondition;
import animatlab.kodamat.core.proba.IProbabilities;
import animatlab.kodamat.core.trees.DNode;
import animatlab.kodamat.core.trees.Leaf;
import animatlab.kodamat.core.trees.Tree;
import animatlab.kodamat.core.utils.Parameters;
import animatlab.kodamat.core.utils.Variable;
import animatlab.kodamat.reinforcementlearning.model.forest.StaticForest;

/**
 * 
 * Class for generation a best case problem as a xml file read by Forest
 * 
 * @author Thomas Degris
 * 
 */
public class ExponProblem extends SyntheticMDPsProblem {

	/**
	 * 
	 * Name of the bestcase problem
	 * 
	 */
	public static final String WORSTCASE_PROBLEM_NAME = "worstcase";

	/**
	 * 
	 * Constructor
	 * 
	 */
	public ExponProblem() {
		super(WORSTCASE_PROBLEM_NAME);
	}

	/**
	 * @see animatlab.kodamat.problems.generated.ForestGeneration#buildParameters(animatlab.kodamat.reinforcementlearning.model.forest.StaticForest)
	 */
	@Override
	protected void buildParameters(StaticForest forest) {
		super.buildParameters(forest);
		if (problemSize < 8) {
			forest.getParameters().setKeyValue(
					Parameters.PARAM_KEY_THRESHOLD_CONVERGENCE, "0.1");
		} else if (problemSize == 8) {
			forest.getParameters().setKeyValue(
					Parameters.PARAM_KEY_THRESHOLD_CONVERGENCE, "0.01");
		} else if (problemSize == 9) {
			forest.getParameters().setKeyValue(
					Parameters.PARAM_KEY_THRESHOLD_CONVERGENCE, "0.0001");
		} else if (problemSize == 10) {
			forest.getParameters().setKeyValue(
					Parameters.PARAM_KEY_THRESHOLD_CONVERGENCE, "0.00001");
		} else {
			forest.getParameters().setKeyValue(Parameters.PARAM_KEY_AUTO_AGENT,
					"false");
		}
	}

	/**
	 * @see animatlab.kodamat.problems.generated.ForestGeneration#getVariableActionCptNode(animatlab.kodamat.reinforcementlearning.model.forest.StaticForest,
	 *      java.lang.String, animatlab.kodamat.core.utils.Variable)
	 */
	@Override
	protected Tree<IProbabilities<String>> getVariableActionCptNode(
			StaticForest forest, String action, Variable variable) {
		Tree<IProbabilities<String>> result = new Tree<IProbabilities<String>>();
		int actionIndex = Integer.parseInt(action);
		if (getIndex(variable) > actionIndex) {
			deterministicSameValue((Leaf<IProbabilities<String>>) result.getRoot(),
					variable);
			return result;
		}
		if (getIndex(variable) < actionIndex) {
			createUniqueLeafProbability((Leaf<IProbabilities<String>>) result
					.getRoot(), "false");
			return result;
		}
		Leaf<IProbabilities<String>> currentLeaf = (Leaf<IProbabilities<String>>) result
				.getRoot();
		if (noise > 0) {
			createBinaryProbaLeaf(currentLeaf, "true");
		} else {
			createUniqueLeafProbability(currentLeaf, "true");
		}
		for (int variableIndex = 0; variableIndex < actionIndex; variableIndex++) {
			Variable currentVariable = getVariable(variableIndex);
			DNode<IProbabilities<String>> dnode = currentLeaf
					.replace(currentVariable);
			createUniqueLeafProbability(dnode.addLeaf("false"), "false");
			currentLeaf = dnode.addLeaf("true");
			if (noise > 0) {
				createBinaryProbaLeaf(currentLeaf, "true");
			} else {
				createUniqueLeafProbability(currentLeaf, "true");
			}
		}
		return result;
	}

	/**
	 * @see animatlab.kodamat.problems.generated.ForestGeneration#buildInitialCondition(animatlab.kodamat.reinforcementlearning.model.forest.StaticForest)
	 */
	@Override
	protected void buildInitialCondition(StaticForest forest) {
		OperatorCondition condition = new AndCondition();
		for (int variableIndex = 0; variableIndex < problemSize; variableIndex++) {
			condition.addArg(new VariableCondition(getVariable(variableIndex),
					"false"));
		}
		forest.setInitialCondition(condition);
	}
}
