/*
 * 
 * Created on 14 sept. 2005
 *
 */
package animatlab.kodamat.problems.generated.syntheticmdps;

import animatlab.kodamat.core.conditions.AndCondition;
import animatlab.kodamat.core.conditions.OperatorCondition;
import animatlab.kodamat.core.conditions.VariableCondition;
import animatlab.kodamat.core.proba.IProbabilities;
import animatlab.kodamat.core.trees.DNode;
import animatlab.kodamat.core.trees.Leaf;
import animatlab.kodamat.core.trees.Tree;
import animatlab.kodamat.core.utils.Variable;
import animatlab.kodamat.reinforcementlearning.model.forest.StaticForest;

/**
 * 
 * Class for generation a best case problem as a xml file read by Forest
 * 
 * @author Thomas Degris
 * 
 */
public class LinearProblem extends SyntheticMDPsProblem {
	/**
	 * 
	 * Name of the bestcase problem
	 * 
	 */
	public static final String BESTCASE_PROBLEM_NAME = "bestcase";

	/**
	 * 
	 * Constructor
	 * 
	 */
	public LinearProblem() {
		super(BESTCASE_PROBLEM_NAME);
	}

	/**
	 * @see animatlab.kodamat.problems.generated.ForestGeneration#getVariableActionCptNode(animatlab.kodamat.reinforcementlearning.model.forest.StaticForest,
	 *      java.lang.String, animatlab.kodamat.core.utils.Variable)
	 */
	@Override
	protected Tree<IProbabilities<String>> getVariableActionCptNode(
			StaticForest forest, String action, Variable variable) {
		Tree<IProbabilities<String>> result = new Tree<IProbabilities<String>>();
		int actionIndex = Integer.parseInt(action);
		if (getIndex(variable) < actionIndex) {
			if ((getIndex(variable) + 1 == actionIndex) && (noise > 0)) {
				DNode<IProbabilities<String>> dnode = result.getRoot()
						.replace(variable);
				createUniqueLeafProbability(dnode.addLeaf("false"), "false");
				createBinaryProbaLeaf(dnode.addLeaf("true"), "true");
				return result;
			}
			deterministicSameValue((Leaf<IProbabilities<String>>) result.getRoot(),
					variable);
			return result;
		}
		if (getIndex(variable) > actionIndex) {
			createUniqueLeafProbability((Leaf<IProbabilities<String>>) result
					.getRoot(), "false");
			return result;
		}

		Leaf<IProbabilities<String>> currentLeaf = (Leaf<IProbabilities<String>>) result
				.getRoot();
		createUniqueLeafProbability(currentLeaf, "true");
		for (int variableIndex = 0; variableIndex < actionIndex; variableIndex++) {
			Variable currentVariable = getVariable(variableIndex);
			DNode<IProbabilities<String>> dnode = currentLeaf
					.replace(currentVariable);
			createUniqueLeafProbability(dnode.addLeaf("false"), "false");
			currentLeaf = dnode.addLeaf("true");
			createUniqueLeafProbability(currentLeaf, "true");
		}
		return result;
	}

	// /**
	// * @see
	// animatlab.kodamat.problems.generated.ForestGeneration#buildParameters(animatlab.kodamat.reinforcementlearning.model.forest.StaticForest)
	// **/
	// @Override
	// protected void buildParameters(StaticForest forest) {
	// forest.getParameters().setKeyValue(Parameters.PARAM_KEY_EPISODE_TIME,
	// String.valueOf(problemSize * 2));
	// super.buildParameters(forest);
	// }
	//	
	/**
	 * @see animatlab.kodamat.problems.generated.ForestGeneration#buildTerminalCondition(animatlab.kodamat.reinforcementlearning.model.forest.StaticForest)
	 */
	@Override
	protected void buildTerminalCondition(StaticForest forest) {
		OperatorCondition condition = new AndCondition();
		for (int variableIndex = 0; variableIndex < problemSize; variableIndex++) {
			condition
					.addArg(new VariableCondition(getVariable(variableIndex), "true"));
		}
		forest.setTerminalCondition(condition);
	}

}
