/*
 * 
 * Created on 20 sept. 2005
 *
 */
package animatlab.kodamat.problems.generated.syntheticmdps;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import animatlab.kodamat.core.proba.IProbabilities;
import animatlab.kodamat.core.trees.DNode;
import animatlab.kodamat.core.trees.Leaf;
import animatlab.kodamat.core.trees.Tree;
import animatlab.kodamat.core.trees.Vertex;
import animatlab.kodamat.core.utils.Parameters;
import animatlab.kodamat.core.utils.Variable;
import animatlab.kodamat.problems.generated.ForestGeneration;
import animatlab.kodamat.reinforcementlearning.model.StaticCPD;
import animatlab.kodamat.reinforcementlearning.model.forest.StaticForest;

/**
 * 
 * Class for generating a problem
 * 
 * @author Thomas Degris
 * 
 */
public abstract class SyntheticMDPsProblem extends ForestGeneration {
	/**
	 * 
	 * Value of the reward
	 * 
	 */
	public static final double REWARD_VALUE = 1.0;

	/**
	 * 
	 * Constructor
	 * 
	 * @param problemName
	 *          name of the generated problem
	 * 
	 */
	protected SyntheticMDPsProblem(String problemName) {
		super(problemName);
	}

	/**
	 * @see animatlab.kodamat.problems.generated.ForestGeneration#buildParameters(animatlab.kodamat.reinforcementlearning.model.forest.StaticForest)
	 */
	@Override
	protected void buildParameters(StaticForest forest) {
		forest.getParameters().setKeyValue(Parameters.PARAM_KEY_ORDERING_THRESHOLD,
				"null");
		forest.getParameters().setKeyValue(Parameters.PARAM_KEY_ORDERING, "false");
		forest.getParameters().setKeyValue(Parameters.PARAM_KEY_RESET_REWARD,
				"false");
	}

	/**
	 * 
	 * Create a new leaf in a tree
	 * 
	 * @param document
	 *          the document owning the node
	 * @param value
	 *          the value to reach the new leaf
	 * 
	 * @return the xml node representing the leaf
	 * 
	 */
	protected Element createLeaf(Document document, String value) {
		Element leaf = document.createElement(Leaf.NODE_NAME);
		if (value != null) {
			leaf.setAttribute(Vertex.ATTRIBUTE_VALUE, value);
		}
		return leaf;
	}

	/**
	 * @see animatlab.kodamat.problems.generated.ForestGeneration#buildRewardFunctions(animatlab.kodamat.reinforcementlearning.model.forest.StaticForest)
	 */
	@Override
	protected void buildRewardFunctions(StaticForest forest) {
		Tree<Double> rewardTree = new Tree<Double>();
		Leaf<Double> currentLeaf = (Leaf<Double>) rewardTree.getRoot();
		for (int variableIndex = 0; variableIndex < problemSize; variableIndex++) {
			Variable variable = getVariable(variableIndex);
			DNode<Double> dnode = currentLeaf.replace(variable);
			dnode.addLeaf("false").set(0.0);
			currentLeaf = dnode.addLeaf("true");
		}
		currentLeaf.set(REWARD_VALUE);
		forest.addReward(rewardTree);
	}

	/**
	 * 
	 * Create a (sub)tree returning the same value than a variable
	 * 
	 * @param leaf
	 *          the leaf where to create the deterministic value
	 * @param variable
	 *          the index of the variable to test
	 * 
	 */
	protected void deterministicSameValue(Leaf<IProbabilities<String>> leaf,
			Variable variable) {
		DNode<IProbabilities<String>> dnode = leaf.replace(variable);
		createUniqueLeafProbability(dnode.addLeaf("false"), "false");
		createUniqueLeafProbability(dnode.addLeaf("true"), "true");
	}

	/**
	 * 
	 * Create a probability node for binary value
	 * 
	 * @param leaf
	 *          the leaf where to create the deterministic value
	 * @param resultValue
	 *          value with best probability
	 * 
	 */
	protected void createBinaryProbaLeaf(Leaf<IProbabilities<String>> leaf,
			String resultValue) {
		StaticCPD staticCPD = new StaticCPD();
		if (resultValue.equals("true")) {
			staticCPD.setProba("false", noise);
			staticCPD.setProba("true", 1 - noise);
		} else {
			staticCPD.setProba("false", 1 - noise);
			staticCPD.setProba("true", noise);
		}
		leaf.set(staticCPD);
	}
}
