package animatlab.kodamat.problems.generated.sysadmin;

import animatlab.kodamat.core.functions.FunctionsSum;
import animatlab.kodamat.core.proba.IProbabilities;
import animatlab.kodamat.core.trees.DNode;
import animatlab.kodamat.core.trees.Leaf;
import animatlab.kodamat.core.trees.Tree;
import animatlab.kodamat.core.utils.Variable;
import animatlab.kodamat.reinforcementlearning.model.StaticCPD;
import animatlab.kodamat.reinforcementlearning.model.forest.StaticForest;

/**
 * 
 * Sys admin problem where all the machines are connected together
 * 
 * @author Thomas Degris
 * 
 */
public class FullyConnected extends SysAdminProblem {
	/**
	 * 
	 * Name of the bestcase problem
	 * 
	 */
	public static final String FULLY_CONNECTED_PROBLEM_NAME = "fullyconnected";

	/**
	 * 
	 * Constructor
	 * 
	 */
	public FullyConnected() {
		super(FULLY_CONNECTED_PROBLEM_NAME);
	}

	/**
	 * @see animatlab.kodamat.problems.generated.ForestGeneration#getBasisFunction(animatlab.kodamat.reinforcementlearning.model.forest.StaticForest)
	 */
	@Override
	protected FunctionsSum getBasisFunction(StaticForest forest) {
		FunctionsSum treeBasisFunction = computeTreeBasisFunction(forest);
		return forest.decomposeBasisFunction(treeBasisFunction);
	}

	/**
	 * 
	 * Create a probability node for binary value
	 * 
	 * @param leaf
	 *          the leaf where to create the deterministic value
	 * @param workingProba
	 *          probibility that the machine will work
	 * 
	 */
	protected void createProbaLeaf(Leaf<IProbabilities<String>> leaf,
			double workingProba) {
		StaticCPD staticCPD = new StaticCPD();
		staticCPD.setProba("W", workingProba);
		staticCPD.setProba("D", 1 - workingProba);
		leaf.set(staticCPD);
	}

	/**
	 * @see animatlab.kodamat.problems.generated.ForestGeneration#getVariableActionCptNode(animatlab.kodamat.reinforcementlearning.model.forest.StaticForest,
	 *      java.lang.String, animatlab.kodamat.core.utils.Variable)
	 */
	@Override
	protected Tree<IProbabilities<String>> getVariableActionCptNode(
			StaticForest forest, String action, Variable variable) {
		if (action.equals(extractActionName(variable.variable))) {
			return createLeafTree(forest, "W");
		}
		Tree<IProbabilities<String>> result = new Tree<IProbabilities<String>>();
		Leaf<IProbabilities<String>> workingLeaf = (Leaf<IProbabilities<String>>) result
				.getRoot();
		for (int variableIndex = 0; variableIndex < problemSize; variableIndex++) {
			Variable currentVariable = getVariable(variableIndex);
			DNode<IProbabilities<String>> dnode = workingLeaf
					.replace(currentVariable);
			Leaf<IProbabilities<String>> deadLeaf = dnode.addLeaf("D");
			createProbaLeaf(deadLeaf, 0.1);
			workingLeaf = dnode.addLeaf("W");
		}
		createProbaLeaf(workingLeaf, 0.9);
		return result;
	}

}
