package animatlab.kodamat.problems.generated.sysadmin;

import animatlab.kodamat.core.io.XmlUtils;
import animatlab.kodamat.core.proba.IProbabilities;
import animatlab.kodamat.core.trees.Tree;
import animatlab.kodamat.core.utils.Variable;
import animatlab.kodamat.reinforcementlearning.model.forest.StaticForest;

/**
 * 
 * Generate a sysadmin problem with a unidirectional ring topology
 * 
 * @author Thomas Degris
 * 
 */
public class UnidirectionalRing extends SysAdminProblem {
	/**
	 * 
	 * Name of the bestcase problem
	 * 
	 */
	public static final String UNIDIRECTIONAL_RING_PROBLEM_NAME = "unidirectionalring";

	/**
	 * 
	 * Constructor
	 * 
	 */
	public UnidirectionalRing() {
		super(UNIDIRECTIONAL_RING_PROBLEM_NAME);
	}

	/**
	 * @see animatlab.kodamat.problems.generated.ForestGeneration#getVariableActionCptNode(animatlab.kodamat.reinforcementlearning.model.forest.StaticForest,
	 *      java.lang.String, animatlab.kodamat.core.utils.Variable)
	 */
	@Override
	protected Tree<IProbabilities<String>> getVariableActionCptNode(
			StaticForest forest, String action, Variable variable) {
		if (action.equals(extractActionName(variable.variable))) {
			return createLeafTree(forest, "W");
		}
		Variable precMachineState = getVariable((getIndex(variable) + problemSize - 1)
				% problemSize);
		String treeTemplate = "<dtr:dnode variable='" + precMachineState.variable
				+ "'>" + "  <dtr:dnode value='D' variable='" + variable.variable + "'>"
				+ "    <dtr:leaf value='D'/>" + "    <dtr:leaf value='W'/>"
				+ "  </dtr:dnode>" + "  <dtr:dnode value='W' variable='"
				+ variable.variable + "'>" + "    <dtr:leaf value='D'/>"
				+ "    <dtr:leaf value='W'/>" + "  </dtr:dnode>" + "</dtr:dnode>";
		Tree<IProbabilities<String>> tree = new Tree<IProbabilities<String>>(
				XmlUtils.getRoot(treeTemplate), null);
		setProba(tree, "D,D", 0.05, precMachineState, variable);
		setProba(tree, "D,W", 0.5, precMachineState, variable);
		setProba(tree, "W,D", 0.09, precMachineState, variable);
		setProba(tree, "W,W", 0.9, precMachineState, variable);
		return tree;
	}
}
