/*
 * 
 * Created on Jan 10, 2006
 *
 */
package animatlab.kodamat.problems.mazespace;

import java.awt.Color;
import java.awt.Graphics;
import java.awt.Point;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import animatlab.kodamat.core.functions.IFunction;
import animatlab.kodamat.core.io.IToXml;
import animatlab.kodamat.core.io.Log;
import animatlab.kodamat.core.utils.Parameters;
import animatlab.kodamat.core.utils.Context;
import animatlab.kodamat.core.utils.Variable;
import animatlab.kodamat.problems.mazespace.MazeView.DrawingPolicy;
import animatlab.kodamat.reinforcementlearning.agent.IAgent;
import animatlab.kodamat.reinforcementlearning.agent.ModelBasedAgent;
import animatlab.kodamat.reinforcementlearning.model.AbstractFactoredModel;
import animatlab.kodamat.reinforcementlearning.planning.IIncrementalPlanning;

/**
 * 
 * IncrementalAgent running in the maze
 * 
 * @author Thomas Degris
 * 
 */
public class MazeAgent implements IAgent, IToXml<Object> {
	/**
	 * 
	 * Policy to draw the maze with nothing inside
	 * 
	 */
	private class ColoredValueTreeDrawingPolicy implements DrawingPolicy {
		/**
		 * 
		 * Minimum value to display
		 * 
		 */
		private double minimumValue;

		/**
		 * 
		 * Maximum value to display
		 * 
		 */
		private double maximumValue;

		/**
		 * 
		 * Constructor
		 * 
		 */
		ColoredValueTreeDrawingPolicy() {
			super();
		}

		/**
		 * @see animatlab.kodamat.problems.mazespace.MazeView.DrawingPolicy#draw(java.awt.Point,
		 *      java.awt.Graphics)
		 */
		public void draw(Point position, Graphics graphics) {
			Context state = new Context();
			if ((actionValueFunctions == null)
					|| (actionValueFunctions.size() != mazeModel.getActionList().size())) {
				graphics.setColor(Color.WHITE);
			} else {
				getMinMaxValue();
				mazeModel.setState(position, state);
				Double value = getValue(state);
				if ((value != null) && (maximumValue > minimumValue)) {
					float factor = (float) ((maximumValue - value) / (maximumValue - minimumValue));
					assert (factor <= 1 && factor >= 0);
					graphics.setColor(new Color(1.0f, 1.0f, factor));
				} else {
					graphics.setColor(Color.WHITE);
				}
			}
			graphics.fillRect(position.x * MazeView.CASE_SIZE, position.y
					* MazeView.CASE_SIZE, (position.x + 1) * MazeView.CASE_SIZE,
					(position.y + 1) * MazeView.CASE_SIZE);
		}

		/**
		 * 
		 * Tests all the states of the maze to determine the minimum and the maximum
		 * value to display
		 * 
		 */
		private void getMinMaxValue() {
			minimumValue = Double.MAX_VALUE;
			maximumValue = -Double.MAX_VALUE;
			Point position = new Point();
			Context state = new Context();
			for (int x = 0; x < mazeModel.getDimension().width; x++) {
				for (int y = 0; y < mazeModel.getDimension().height; y++) {
					position.x = x;
					position.y = y;
					if (mazeModel.getValueAt(position) != MazeModel.WALL_VALUE) {
						mazeModel.setState(position, state);
						Double value = getValue(state);
						if (value != null) {
							if (value > maximumValue) {
								maximumValue = value;
							}
							if (value < minimumValue) {
								minimumValue = value;
							}
						}
					}
				}
			}
			if (!(maximumValue > minimumValue)) {
				maximumValue += 1;
			}
		}

	}

	/**
	 * 
	 * Policy to draw the maze with nothing inside
	 * 
	 */
	private class WrittenValueTreeDrawingPolicy extends
			ColoredValueTreeDrawingPolicy {
		/**
		 * 
		 * Constructor
		 * 
		 */
		WrittenValueTreeDrawingPolicy() {
			super();
		}

		/**
		 * @see animatlab.kodamat.problems.mazespace.MazeAgent.ColoredValueTreeDrawingPolicy#draw(java.awt.Point,
		 *      java.awt.Graphics)
		 */
		@Override
		public void draw(Point position, Graphics graphics) {
			if (actionValueFunctions == null) {
				return;
			}
			Context state = new Context();
			mazeModel.setState(position, state);
			String valueText = "??";
			Double value = getValue(state);
			if (value != null) {
				valueText = String.valueOf(value);
			}
			graphics.setColor(Color.BLACK);
			graphics.drawString(valueText, position.x * MazeView.CASE_SIZE + 3,
					position.y * MazeView.CASE_SIZE + MazeView.CASE_SIZE / 2);
		}

	}

	/**
	 * 
	 * Policy to draw the maze with nothing inside
	 * 
	 */
	protected class PolicyTreeDrawingPolicy implements DrawingPolicy {
		/**
		 * Draw an action in a case of the maze
		 * 
		 * @param position
		 *          in the maze to draw
		 * @param graphics
		 *          the graphics where to draw the action
		 * @param actionToDraw
		 *          the action to draw
		 */
		protected void drawAction(Point position, Graphics graphics,
				String actionToDraw) {
			int action = Integer.parseInt(actionToDraw);
			Point target = new Point(MazeView.CASE_SIZE / 2
					+ MazeModel.ACTIONS[action][0] * MazeView.CASE_SIZE / 2,
					MazeView.CASE_SIZE / 2 + MazeModel.ACTIONS[action][1]
							* MazeView.CASE_SIZE / 2);
			Point center = new Point(MazeView.CASE_SIZE / 2, MazeView.CASE_SIZE / 2);
			target.x += position.x * MazeView.CASE_SIZE;
			target.y += position.y * MazeView.CASE_SIZE;
			center.x += position.x * MazeView.CASE_SIZE;
			center.y += position.y * MazeView.CASE_SIZE;
			graphics.drawLine(center.x, center.y, target.x, target.y);
			graphics.drawLine(center.x + MazeModel.ACTIONS[action][1]
					* MazeView.CASE_SIZE / 8 + MazeModel.ACTIONS[action][0]
					* MazeView.CASE_SIZE / 4, center.y + MazeModel.ACTIONS[action][0]
					* MazeView.CASE_SIZE / 8 + MazeModel.ACTIONS[action][1]
					* MazeView.CASE_SIZE / 4, target.x, target.y);
			graphics.drawLine(center.x - MazeModel.ACTIONS[action][1]
					* MazeView.CASE_SIZE / 8 + MazeModel.ACTIONS[action][0]
					* MazeView.CASE_SIZE / 4, center.y - MazeModel.ACTIONS[action][0]
					* MazeView.CASE_SIZE / 8 + MazeModel.ACTIONS[action][1]
					* MazeView.CASE_SIZE / 4, target.x, target.y);
		}

		/**
		 * @see animatlab.kodamat.problems.mazespace.MazeView.DrawingPolicy#draw(java.awt.Point,
		 *      java.awt.Graphics)
		 */
		public void draw(Point position, Graphics graphics) {
			Context currentState = new Context();
			mazeModel.setState(position, currentState);
			Context currentStateAction = new Context(currentState);
			for (int index = 0; index < MazeModel.actions.length; index++) {
				if (agent.getModel().getActionTestedSet().contains(
						MazeModel.actions[index])) {
					currentStateAction.put(Variable.ACTION, MazeModel.actions[index]);
					final float brightness = 0.5f;
					Color reference;
					if (!testedStateAction.contains(currentStateAction)) {
						reference = new Color(brightness, 0.0f, 0.0f);
					} else {
						reference = new Color(0.0f, brightness, 0.0f);
					}
					Color actionColor = reference;
					if ((actionValueFunctions != null)
							&& (actionValueFunctions.size() == mazeModel.getActionList()
									.size())) {
						Set<String> actions = getBestAction(currentState);
						if ((actions != null) && actions.contains(MazeModel.actions[index])) {
							actionColor = new Color(reference.getRed(), reference.getGreen(),
									255);
						}
					}
					graphics.setColor(actionColor);
					drawAction(position, graphics, MazeModel.actions[index]);
				}
			}
		}
	}

	/**
	 * 
	 * Policy to draw the agent in the maze
	 * 
	 */
	class AgentDrawingPolicy implements DrawingPolicy {
		/**
		 * 
		 * Size of the agent in the maze
		 * 
		 */
		private final int AGENT_SIZE = 10;

		/**
		 * @see animatlab.kodamat.problems.mazespace.MazeView.DrawingPolicy#draw(java.awt.Point,
		 *      java.awt.Graphics)
		 */
		public void draw(Point position, Graphics graphics) {
			if (agentPosition == null) {
				return;
			}
			if (!agentPosition.equals(position)) {
				return;
			}
			graphics.setColor(Color.BLUE);
			graphics.fillOval(MazeView.CASE_SIZE * position.x + MazeView.CASE_SIZE
					/ 2 - (AGENT_SIZE / 2), MazeView.CASE_SIZE * position.y
					+ MazeView.CASE_SIZE / 2 - (AGENT_SIZE / 2), AGENT_SIZE, AGENT_SIZE);
		}
	}

	/**
	 * 
	 * Model of the maze
	 * 
	 */
	final protected MazeModel mazeModel;

	/**
	 * 
	 * The value tree to draw
	 * 
	 */
	Map<String, IFunction<Double>> actionValueFunctions = null;

	/**
	 * 
	 * Reward function of the agent
	 * 
	 */
	IFunction<Double> rewardFunction = null;

	/**
	 * 
	 * The set of tested state/action
	 * 
	 */
	Set<Context> testedStateAction = new LinkedHashSet<Context>();

	/**
	 * 
	 * IncrementalAgent to run in the environment
	 * 
	 */
	final ModelBasedAgent agent;

	/**
	 * 
	 * IncrementalAgent position in the maze
	 * 
	 */
	Point agentPosition;

	/**
	 * 
	 * Last state of the agent
	 * 
	 */
	Context lastState;

	/**
	 * 
	 * Last action executed by the agent
	 * 
	 */
	String lastAction;

	/**
	 * 
	 * Constructor
	 * 
	 * @param mazeView
	 *          view of the maze where to run this agent
	 * @param agent
	 *          agent to run in the environment
	 * 
	 * 
	 */
	public MazeAgent(MazeView mazeView, ModelBasedAgent agent) {
		mazeModel = mazeView.mazeModel;
		this.agent = agent;
		registerOnViewer(mazeView);
	}

	/**
	 * 
	 * Compute the value for a state
	 * 
	 * @param state
	 *          the state to compute the value for
	 * @return the value of the state
	 * 
	 */
	double getValue(Context state) {
		return getValueFromValueFunction(state);
	}

	/**
	 * 
	 * Compute the value function displayed from the action-value function
	 * 
	 * @param state
	 *          the state to display
	 * @return the value
	 * 
	 */
	protected double getValueFromReward(Context state) {
		double reward = 0.0;
		if (rewardFunction == null) {
			return reward;
		}
		for (String action : mazeModel.getActionList()) {
			Context stateWithAction = new Context(state);
			stateWithAction.put(Variable.ACTION, action);
			Double actionReward = rewardFunction.classify(stateWithAction);
			if (actionReward != null) {
				reward = reward + actionReward;
			}
		}
		return reward;
	}

	/**
	 * 
	 * Compute the value function displayed from the action-value function
	 * 
	 * @param state
	 *          the state to display
	 * @return the value
	 * 
	 */
	protected double getValueFromValueFunction(Context state) {
		double max = -Double.MAX_VALUE;
		for (String action : mazeModel.getActionList()) {
			IFunction<Double> actionValueFunction = actionValueFunctions.get(action);
			if ((actionValueFunction != null)
					&& (actionValueFunction.classify(state) != null)) {
				max = Math.max(max, actionValueFunction.classify(state));
			}
		}
		return max;
	}

	/**
	 * 
	 * Compute the best action for a state
	 * 
	 * @param state
	 *          the state to compute the value for
	 * @return the best action
	 * 
	 */
	Set<String> getBestAction(Context state) {
		double max = -Double.MAX_VALUE;
		Set<String> bestActions = new LinkedHashSet<String>();
		for (String action : mazeModel.getActionList()) {
			if (actionValueFunctions.get(action).classify(state) == null) {
				return null;
			}
			double value = actionValueFunctions.get(action).classify(state);
			if (Math.abs(value - max) < Parameters.EPSILON) {
				bestActions.add(action);
			} else if (value > max) {
				bestActions.clear();
				bestActions.add(action);
				max = value;
			}
		}
		return bestActions;
	}

	/**
	 * 
	 * Register this maze agent to be displayed on the maze view
	 * 
	 * @param mazeView
	 *          the view to register the drawing policy to
	 * 
	 */
	protected void registerOnViewer(MazeView mazeView) {
		mazeView.removePolicies();
		mazeView.registerPolicy(new ColoredValueTreeDrawingPolicy());
		mazeView.registerPolicy(new PolicyTreeDrawingPolicy());
		mazeView.registerPolicy(new WrittenValueTreeDrawingPolicy());
		mazeView.registerPolicy(new AgentDrawingPolicy());
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.agent.IAgent#getAction(int,
	 *      Context, List)
	 */
	public String getAction(int time, Context state, List<Double> reward) {
		if (lastState != null) {
			Context stateAction = new Context(lastState);
			stateAction.put(Variable.ACTION, lastAction);
			testedStateAction.add(stateAction);
		}
		agentPosition = mazeModel.stateToPosition.get(state);
		lastAction = agent.getAction(time, state, reward);
		lastState = state;
		IIncrementalPlanning incrementalPlanning = (IIncrementalPlanning) agent
				.getPlanning();
		actionValueFunctions = incrementalPlanning.getActionValueFunctions();
		if (((AbstractFactoredModel) incrementalPlanning.getModel())
				.getRewardFunction(null).size() > 0) {
			rewardFunction = ((AbstractFactoredModel) incrementalPlanning.getModel())
					.getRewardFunction(null).functions().get(0);
		}
		return lastAction;

	}

	/**
	 * @see animatlab.kodamat.core.io.IToXml#toXml(org.w3c.dom.Document,
	 *      animatlab.kodamat.core.io.IToXml.IToXmlPolicy)
	 */
	@SuppressWarnings("unchecked")
	public Element toXml(Document document, IToXmlPolicy policy) {
		if (!(agent.getPlanning() instanceof IToXml)) {
			return null;
		}
		return ((IToXml) agent.getPlanning()).toXml(document, null);
	}

	/**
	 * @see animatlab.kodamat.core.io.ILog#setLog(animatlab.kodamat.core.io.Log)
	 */
	public void setLog(Log log) {
		// Nothing to log
	}
}
