package animatlab.kodamat.problems.tests;

import java.util.List;
import java.util.Random;

import junit.framework.TestCase;
import animatlab.kodamat.core.io.HtmlPage;
import animatlab.kodamat.core.io.Log;
import animatlab.kodamat.core.io.XmlUtils;
import animatlab.kodamat.core.utils.Context;
import animatlab.kodamat.problems.ProblemRunner;
import animatlab.kodamat.reinforcementlearning.agent.IAgent;
import animatlab.kodamat.reinforcementlearning.core.IStepListener;
import animatlab.kodamat.reinforcementlearning.model.forest.StaticForest;

/**
 * 
 * Test the factory problem
 * 
 * @author Thomas Degris
 * 
 */
public class TestFactory extends TestCase {
	/**
	 * 
	 * Test the problem factory
	 * 
	 */
	public void testFactory() {
		validateProblem("factory");
	}

	/**
	 * 
	 * Test the problem factory
	 * 
	 */
	public void testFactory04() {
		validateProblem("factory4");
	}

	/**
	 * 
	 * Validate if a factory problem can be terminated
	 * 
	 * @param problem
	 *          the problem to test
	 * 
	 */
	private void validateProblem(final String problem) {
		final StaticForest forest = new StaticForest(XmlUtils
				.getRootFromFile("./data/forests/" + problem
						+ StaticForest.PROBLEM_EXTENSION));
		final Random random = new Random(0);
		final List<String> actionList = forest.getActionList();
		IAgent agent = new IAgent() {
			int actionIndex = 0;

			public String getAction(int time, Context state, List<Double> reward) {
				if (state == null) {
					return null;
				}
				String result = actionList.get(actionIndex % actionList.size());
				actionIndex++;
				if (result.contains("shape")) {
					return getAction(time, null, null);
				}
				return result;
			}

			public void setLog(Log log) {
				// Nothing to do with log
			}
		};
		ProblemRunner problemRunner = new ProblemRunner(random, forest, agent,
				Log.sysout);
		problemRunner.registerListener(new IStepListener() {
			int lastTerminalStateTime = 0;

			int maxNonTerminalTime = forest.getActionList().size() * 5;

			public void listenStep(int time, Context currentState,
					String actionChoosen, Context nextState, List<Double> lastReward) {
				if (currentState == null) {
					lastTerminalStateTime = time;
				}
				if ((time - lastTerminalStateTime) >= maxNonTerminalTime) {
					HtmlPage htmlDotPage = new HtmlPage("./.reports/" + problem,
							"index.html");
					Log.sysout.out().println(
							"\nDumping in " + htmlDotPage.getAbsolutePath());
					htmlDotPage.setHtmlPageSetTitle(problem);
					htmlDotPage.setTitle("Index");
					forest.dump(htmlDotPage);
					htmlDotPage.write();
				}
				assert ((time - lastTerminalStateTime) < maxNonTerminalTime);
			}

			public void terminate() {
				// Nothing to do
			}
		});
		problemRunner.runProblem(forest.getActionList().size() * 100);
	}
}
