package animatlab.kodamat.reinforcementlearning.agent;

import java.util.Random;

import animatlab.kodamat.core.IProblemDefinition;
import animatlab.kodamat.core.io.ILog;
import animatlab.kodamat.reinforcementlearning.planning.IPlanning;

/**
 * 
 * Interface used to create agents to run on a problem
 * 
 * @author Thomas Degris
 * 
 */
public interface IAgentFactory extends ILog {
	/**
	 * 
	 * Create an new agent
	 * 
	 * @param random
	 *          random instance that will be used by the agent
	 * @param problem
	 *          the problem where the agent will be
	 * @param solution
	 *          the solution of the problem (may be null)
	 * @return the new agent
	 * 
	 */
	IAgent createAgent(Random random, IProblemDefinition problem,
			IPlanning solution);

	/**
	 * 
	 * @return the name of this agent
	 * 
	 */
	String getAgentName();

	/**
	 * 
	 * @return flag to indicate if this measure require the solution or not
	 * 
	 */
	boolean requireSolution();
}
