/*
 * 
 * Created on Jan 11, 2006
 *
 */
package animatlab.kodamat.reinforcementlearning.core.acting;

import java.util.List;
import java.util.Map;
import java.util.Random;

import animatlab.kodamat.core.functions.IFunction;
import animatlab.kodamat.core.utils.Context;
import animatlab.kodamat.reinforcementlearning.planning.IIncrementalPlanning;

/**
 * 
 * Epsilon greedy acting policy
 * 
 * @author Thomas Degris
 * 
 */
public class EpsilonGreedy extends Greedy {
	/**
	 * 
	 * EPSILON constant for the e-greedy policy
	 * 
	 */
	private double epsilon = 0.1;

	/**
	 * 
	 * Constructor
	 * 
	 * 
	 */
	public EpsilonGreedy() {
		super();
	}

	/**
	 * 
	 * Constructor
	 * 
	 * @param random
	 *          random instance
	 * @param actionList
	 *          set of available actions
	 * 
	 */
	public EpsilonGreedy(Random random, List<String> actionList) {
		super(random, actionList);
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.core.acting.IActing#getAction(java.util.Map,
	 *      Context)
	 */
	@Override
	public String getAction(Map<String, IFunction<Double>> qualityFunctions,
			Context state) {
		if (random.nextDouble() < epsilon) {
			return randomAction();
		}
		return super.getAction(qualityFunctions, state);
	}

	/**
	 * 
	 * Set epsilon
	 * 
	 * @param epsilon
	 *          new value for epsilon
	 * 
	 */
	public void setEpsilon(double epsilon) {
		this.epsilon = epsilon;
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.core.acting.IActing#newInstance(java.util.Random,
	 *      java.util.List, IIncrementalPlanning)
	 */
	@Override
	public IActing newInstance(Random random, List<String> actionList,
			IIncrementalPlanning learningPlanning) {
		return new EpsilonGreedy(random, actionList);
	}

}
