/*
 * 
 * Created on 8 janv. 2006
 *
 */
package animatlab.kodamat.reinforcementlearning.core.acting;

import java.util.List;
import java.util.Map;
import java.util.Random;

import animatlab.kodamat.core.functions.IFunction;
import animatlab.kodamat.core.io.ILog;
import animatlab.kodamat.core.utils.Context;
import animatlab.kodamat.reinforcementlearning.planning.IIncrementalPlanning;

/**
 * 
 * Acting strategy of an agent.
 * 
 * @author Thomas Degris
 * 
 */
public interface IActing extends ILog {
	/**
	 * 
	 * Get the action from the set of quality functions builded by a planning
	 * algorithm
	 * 
	 * @param qualityFunctions
	 *          the map of quality functions
	 * @param state
	 *          current state of the agent
	 * 
	 * @return the action decided by the acting strategy
	 * 
	 */
	public String getAction(Map<String, IFunction<Double>> qualityFunctions,
			Context state);

	/**
	 * 
	 * @return the name of the acting policy
	 * 
	 */
	public String getName();

	/**
	 * 
	 * New instance of this acting policy
	 * 
	 * @param random
	 *          random instance to use to select actions
	 * @param actionList
	 *          list of available actions
	 * @param learningPlanning
	 *          model used by the agent
	 * @return a new instance of this
	 * 
	 */
	public IActing newInstance(Random random, List<String> actionList,
			IIncrementalPlanning learningPlanning);
}
