/*
 * 
 * Created on Jan 11, 2006
 *
 */
package animatlab.kodamat.reinforcementlearning.core.acting;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Random;

import animatlab.kodamat.core.IProblemDefinition;
import animatlab.kodamat.core.functions.IFunction;
import animatlab.kodamat.core.io.Log;
import animatlab.kodamat.core.utils.Context;
import animatlab.kodamat.reinforcementlearning.agent.IAgent;
import animatlab.kodamat.reinforcementlearning.agent.IAgentFactory;
import animatlab.kodamat.reinforcementlearning.planning.IIncrementalPlanning;
import animatlab.kodamat.reinforcementlearning.planning.IPlanning;

/**
 * 
 * Random acting policy
 * 
 * @author Thomas Degris
 * 
 */
public class RandomActing implements IActing, IAgent, IAgentFactory {
	/**
	 * 
	 * Random instance
	 * 
	 */
	protected final Random random;

	/**
	 * 
	 * List of possible actions
	 * 
	 */
	protected final List<String> actionList;

	/**
	 * 
	 * Log to use to log things
	 * 
	 */
	protected Log log = Log.sysout;

	/**
	 * 
	 * Constructor
	 * 
	 * @param random
	 *          random instance
	 * @param actionSet
	 *          set of available actions
	 * 
	 */
	public RandomActing(Random random, List<String> actionSet) {
		this.random = random;
		actionList = new ArrayList<String>(actionSet);
	}

	/**
	 * 
	 * Constructor when this as factory
	 * 
	 */
	public RandomActing() {
		super();
		random = null;
		actionList = null;
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.core.acting.IActing#getAction(java.util.Map,
	 *      animatlab.kodamat.core.utils.Context)
	 */
	public String getAction(Map<String, IFunction<Double>> qualityFunctions,
			Context state) {
		return getAction(-1, state, null);
	}

	/**
	 * 
	 * @return a random action
	 * 
	 */
	protected String randomAction() {
		return actionList.get(random.nextInt(actionList.size()));
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.core.acting.IActing#getName()
	 */
	public String getName() {
		return getClass().getSimpleName().toLowerCase();
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.agent.IAgent#getAction(int,
	 *      Context, java.util.List)
	 */
	public String getAction(int time, Context state, List<Double> reward) {
		if (state == null) {
			return null;
		}
		return randomAction();
	}

	/**
	 * @see animatlab.kodamat.core.io.ILog#setLog(animatlab.kodamat.core.io.Log)
	 */
	public void setLog(Log log) {
		this.log = log;
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.agent.IAgentFactory#createAgent(Random,
	 *      IProblemDefinition,
	 *      animatlab.kodamat.reinforcementlearning.planning.IPlanning)
	 */
	public IAgent createAgent(Random random, IProblemDefinition problem,
			IPlanning solution) {
		return new RandomActing(random, problem.getActionList());
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.core.acting.IActing#newInstance(java.util.Random,
	 *      java.util.List, IIncrementalPlanning)
	 */
	public IActing newInstance(Random random, List<String> actionList,
			IIncrementalPlanning learningPlanning) {
		return new RandomActing(random, actionList);
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.agent.IAgentFactory#getAgentName()
	 */
	public String getAgentName() {
		return getName();
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.agent.IAgentFactory#requireSolution()
	 */
	public boolean requireSolution() {
		return false;
	}
}
