/*
 * 
 * Created on 15 nov. 2005
 *
 */
package animatlab.kodamat.reinforcementlearning.core.stats;

import java.util.List;
import java.util.Map;

import animatlab.kodamat.core.IProblemDefinition;
import animatlab.kodamat.core.utils.Parameters;
import animatlab.kodamat.core.utils.Context;
import animatlab.kodamat.reinforcementlearning.agent.IAgent;
import animatlab.kodamat.reinforcementlearning.planning.IPlanning;

/**
 * 
 * Measure logging the reward obtained by the agent during the experiment
 * 
 * @author Thomas Degris
 * 
 */
public class RewardMeasure extends Measure {
	/**
	 * 
	 * Gamma value to use to compute the discount reward
	 * 
	 */
	private double gamma;

	/**
	 * 
	 * Discounted reward sum
	 * 
	 */
	private double discountedReward;

	/**
	 * 
	 * Sum of the reward obtained
	 * 
	 */
	private double rewardSum;

	/**
	 * 
	 * Current time
	 * 
	 */
	private int time;

	/**
	 * 
	 * Constructor
	 * 
	 */
	public RewardMeasure() {
		super();
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.core.IStepListener#listenStep(int,
	 *      Context, java.lang.String, Context, java.util.List)
	 */
	@Override
	public void listenStep(int time, Context currentState, String actionChoosen,
			Context nextState, List<Double> lastReward) {
		this.time = time;
		if (lastReward != null) {
			double reward = Context.getSum(lastReward);
			discountedReward = discountedReward * gamma + reward;
			rewardSum += reward;
		}
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.core.stats.Measure#setMeasureData(animatlab.kodamat.core.IProblemDefinition,
	 *      animatlab.kodamat.reinforcementlearning.planning.IPlanning,
	 *      animatlab.kodamat.reinforcementlearning.agent.IAgent,
	 *      java.lang.String)
	 */
	@Override
	public void setMeasureData(IProblemDefinition problem, IPlanning solution,
			IAgent agent, String resultsFolder) {
		super.setMeasureData(problem, solution, agent, resultsFolder);
		reset();
		if (problem.getParameters().hasKey(Parameters.PARAM_KEY_GAMMA)) {
			gamma = problem.getParameters().getKeyD(Parameters.PARAM_KEY_GAMMA);
		} else {
			gamma = Parameters.GAMMA;
		}
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.core.stats.Measure#computeMeasures(Map)
	 */
	@Override
	public void computeMeasures(Map<String, Double> data) {
		if (time == 0) {
			return;
		}
		data.put("disrw", discountedReward);
		data.put("averw", rewardSum / time);
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.core.stats.Measure#reset()
	 */
	@Override
	public void reset() {
		discountedReward = 0;
		time = 0;
		rewardSum = 0;
	}

}
