/*
 * 
 * Created on 15 nov. 2005
 *
 */
package animatlab.kodamat.reinforcementlearning.core.stats;

import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import animatlab.kodamat.core.IProblemDefinition;
import animatlab.kodamat.core.utils.Context;
import animatlab.kodamat.core.utils.Variable;
import animatlab.kodamat.reinforcementlearning.agent.IAgent;
import animatlab.kodamat.reinforcementlearning.planning.IPlanning;

/**
 * 
 * Measure logging the number of couple state/action tested by the agent
 * 
 * @author Thomas Degris
 * 
 */
public class StateActionMeasure extends Measure {
	/**
	 * 
	 * Name for the state/action value
	 * 
	 */
	public static final String NAME_STATEACTION = "staact";

	/**
	 * 
	 * Name for the state/action value
	 * 
	 */
	public static final String NAME_STATEACTION_RATIO = "staactp";

	/**
	 * 
	 * Set of couple state/action visited by the agent
	 * 
	 */
	private Set<Context> stateActionSet;

	/**
	 * 
	 * Constructor
	 * 
	 */
	public StateActionMeasure() {
		super();
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.core.IStepListener#listenStep(int,
	 *      Context, java.lang.String, Context, java.util.List)
	 */
	@Override
	public void listenStep(int time, Context currentState, String actionChoosen,
			Context nextState, List<Double> lastReward) {
		if (currentState == null) {
			return;
		}
		Context currentStateAction = new Context(currentState);
		currentStateAction.put(Variable.ACTION, actionChoosen);
		stateActionSet.add(currentStateAction);
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.core.stats.Measure#computeMeasures(Map)
	 */
	@Override
	public void computeMeasures(Map<String, Double> data) {
		data.put(NAME_STATEACTION, (double) stateActionSet.size());
		double ratio = stateActionSet.size()
				/ (double) (problem.getNumberOfStates() * problem.getActionList()
						.size());
		data.put(NAME_STATEACTION_RATIO, ratio);
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.core.stats.Measure#setMeasureData(animatlab.kodamat.core.IProblemDefinition,
	 *      animatlab.kodamat.reinforcementlearning.planning.IPlanning,
	 *      animatlab.kodamat.reinforcementlearning.agent.IAgent,
	 *      java.lang.String)
	 */
	@Override
	public void setMeasureData(IProblemDefinition problem, IPlanning solution,
			IAgent agent, String resultsFolder) {
		super.setMeasureData(problem, solution, agent, resultsFolder);
		reset();
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.core.stats.Measure#reset()
	 */
	@Override
	public void reset() {
		stateActionSet = new LinkedHashSet<Context>();
	}
}
