/*
 * 
 * Created on Jan 17, 2006
 *
 */
package animatlab.kodamat.reinforcementlearning.model;

import java.util.List;
import java.util.Set;

import animatlab.kodamat.core.IModel;
import animatlab.kodamat.core.IProblemDefinition;
import animatlab.kodamat.core.utils.Context;
import animatlab.kodamat.core.utils.VariableOrdering;

/**
 * 
 * Model able to learn from experiment
 * 
 * @author Thomas Degris
 * 
 */
public interface ILearningModel extends IModel {
	/**
	 * 
	 * Set the legend for this FMDP
	 * 
	 * @param variableOrdering
	 *          the legend to set
	 * 
	 */
	public void setVariableOrdering(VariableOrdering variableOrdering);

	/**
	 * 
	 * @return the variable ordering used for learning
	 * 
	 */
	public VariableOrdering getVariableOrdering();

	/**
	 * Learn an experience in the world
	 * 
	 * @param lastTimeState
	 *          last state the agent was
	 * @param lastTimeAction
	 *          last action the agent did
	 * @param currentState
	 *          current state of the agent
	 * @param reward
	 *          reward received by the agent
	 */
	public void learn(Context lastTimeState, String lastTimeAction,
			Context currentState, List<Double> reward);

	/**
	 * 
	 * @return the set of actions already tried by the agent
	 * 
	 */
	public Set<String> getActionTestedSet();

	/**
	 * 
	 * Create a new instance of this (with the same parameters)
	 * 
	 * @param problem
	 *          list of available actions
	 * @return a new instance of this ready to use
	 * 
	 */
	public ILearningModel newInstance(IProblemDefinition problem);
}
