/*
 * 
 * Created on 8 janv. 2006
 *
 */
package animatlab.kodamat.reinforcementlearning.model.forest;

import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import animatlab.kodamat.core.io.HtmlPage;
import animatlab.kodamat.core.proba.IProbabilities;
import animatlab.kodamat.core.trees.Tree;
import animatlab.kodamat.core.utils.Context;
import animatlab.kodamat.core.utils.Variable;
import animatlab.kodamat.reinforcementlearning.model.AbstractFactoredModel;

/**
 * 
 * Abstract forest for common function between different forests
 * 
 * @author Thomas Degris
 * 
 */
abstract public class AbstractForest extends AbstractFactoredModel {
	/**
	 * 
	 * Name of the node of the reward function
	 * 
	 */
	static public final String NODE_NAME_FOREST = AbstractFactoredModel.NAMESPACE
			+ ":" + "forest";

	/**
	 * 
	 * Forest of CPD tree
	 * 
	 */
	protected Map<Variable, Map<String, Tree<IProbabilities<String>>>> forest = null;

	/**
	 * 
	 * Set of possible value for each variables
	 * 
	 */
	protected final Map<Variable, Set<String>> variableValueSet = new LinkedHashMap<Variable, Set<String>>();

	/**
	 * 
	 * Flag to avoid the computation of the multi-valued data
	 * 
	 */
	protected boolean multiValuedFlag;

	/**
	 * 
	 * Constraint used to extract the tree for an action
	 * 
	 */
	final protected Context actionConstraint = new Context();

	/**
	 * 
	 * Constructor
	 * 
	 * @param actionList
	 *          action set used in this forest
	 * 
	 */
	public AbstractForest(List<String> actionList) {
		super(actionList);
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.model.AbstractFactoredModel#getTransitionFunction(animatlab.kodamat.core.utils.Variable,
	 *      java.lang.String)
	 */
	@Override
	abstract public Tree<IProbabilities<String>> getTransitionFunction(
			Variable variable, String action);

	/**
	 * @see animatlab.kodamat.core.IModel#getValueSets()
	 */
	public Map<Variable, Set<String>> getValueSets() {
		return variableValueSet;
	}

	/**
	 * @see animatlab.kodamat.core.IModel#getValueSet(animatlab.kodamat.core.utils.Variable)
	 */
	public Set<String> getValueSet(Variable variable) {
		return variableValueSet.get(variable);
	}

	/**
	 * 
	 * Dump this forest into a html file
	 * 
	 * @param htmlPage
	 *          html page containing the links to the figures
	 * 
	 */
	abstract public void dump(HtmlPage htmlPage);

	/**
	 * 
	 * @return true if the problem is multi-valued
	 * 
	 */
	public boolean checkIfMultiValued() {
		for (Map.Entry<Variable, Set<String>> entry : getValueSets().entrySet()) {
			if (entry.getValue().size() > 2) {
				return true;
			}
		}
		return false;
	}

	/**
	 * 
	 * @return true if the problem is multi-valued
	 * 
	 */
	final public boolean isMultiValued() {
		return multiValuedFlag;
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.model.AbstractFactoredModel#getXmlRootNodeName()
	 */
	public @Override
	String getXmlRootNodeName() {
		return NODE_NAME_FOREST;
	}

	/**
	 * 
	 * Extract a reward function given a particular action
	 * 
	 * @param globalReward
	 *          the reward from which the action reward has to be extracted
	 * @param action
	 *          the action to extract the reward for
	 * @return the action reward tree
	 * 
	 */
	protected Tree<Double> extractRewardFunction(Tree<Double> globalReward,
			String action) {
		if (action != null) {
			actionConstraint.put(Variable.ACTION, action);
			return globalReward.clone(null, actionConstraint);
		}
		return globalReward.clone();
	}

	/**
	 * 
	 * Get the reward function as a single tree
	 * 
	 * @param action
	 *          the action to get the reward function from
	 * @return the tree representing the reward function
	 * 
	 */
	abstract public Tree<Double> getRewardTree(String action);
}
