/*
 * 
 * Created on Dec 29, 2005
 *
 */
package animatlab.kodamat.reinforcementlearning.model.forest.test;

import static animatlab.kodamat.core.io.XmlUtils.getRoot;
import static animatlab.kodamat.core.io.XmlUtils.getRootFromFile;

import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Random;

import junit.framework.TestCase;
import animatlab.kodamat.core.functions.IPartition;
import animatlab.kodamat.core.io.DataLoadingPolicy;
import animatlab.kodamat.core.proba.IProbabilities;
import animatlab.kodamat.core.trees.Tree;
import animatlab.kodamat.core.utils.Context;
import animatlab.kodamat.core.utils.Variable;
import animatlab.kodamat.core.utils.VariableOrdering;
import animatlab.kodamat.reinforcementlearning.model.StaticCPD;
import animatlab.kodamat.reinforcementlearning.model.forest.StaticForest;

/**
 * 
 * Test loading a forest from a xml file
 * 
 * @author Thomas Degris
 * 
 */
public class TestForestLoading extends TestCase {
	/**
	 * 
	 * Test loading the coffeerobot problem
	 * 
	 */
	public void testCoffeeRobotLoading() {
		StaticForest forest = new StaticForest(
				getRootFromFile("data/forests/coffeerobot.forest"));

		Map<Variable, Integer> expectedLegend = new LinkedHashMap<Variable, Integer>();
		expectedLegend.put(Variable.getVariable("huc"), 0);
		expectedLegend.put(Variable.getVariable("l"), 5);
		checkLegend(expectedLegend, forest.getVariableOrdering());

		String xmlString = "<dtr:dnode variable='huc'>"
				+ "  <dtr:dnode value='1' variable='w'>"
				+ "    <dtr:leaf data='0.9' value='1'/>"
				+ "    <dtr:leaf data='1.0' value='0'/>" + "  </dtr:dnode>"
				+ "  <dtr:dnode value='0' variable='w'>"
				+ "    <dtr:leaf data='0.0' value='1'/>"
				+ "    <dtr:leaf data='0.1' value='0'/>" + "  </dtr:dnode>"
				+ "</dtr:dnode>";
		Tree<Double> expectedRewardTree = new Tree<Double>(getRoot(xmlString),
				DataLoadingPolicy.doubleLoadingPolicy);
		assertEquals(expectedRewardTree.toString(), forest.getRewardTree(null)
				.toString());
		assertEquals(1.0, forest.getRewardMax());

		xmlString = "<dtr:dnode variable='huc'>" + "  <dtr:leaf value='1'>"
				+ "    <spi:probability probability='1.0' value='1'/>"
				+ "  </dtr:leaf>" + "  <dtr:dnode value='0' variable='hrc'>"
				+ "    <dtr:dnode value='1' variable='l'>"
				+ "      <dtr:leaf value='1'>"
				+ "        <spi:probability probability='0.8' value='1'/>"
				+ "        <spi:probability probability='0.2' value='0'/>"
				+ "      </dtr:leaf>" + "      <dtr:leaf value='0'>"
				+ "        <spi:probability probability='1.0' value='0'/>"
				+ "      </dtr:leaf>" + "    </dtr:dnode>" + "    <dtr:leaf value='0'>"
				+ "      <spi:probability probability='1.0' value='0'/>"
				+ "    </dtr:leaf>" + "  </dtr:dnode>" + "</dtr:dnode>";
		Tree<IProbabilities<String>> expectedCpdTree = new Tree<IProbabilities<String>>(
				getRoot(xmlString), StaticCPD.LOADING_POLICY);
		assertEquals(expectedCpdTree.toString(), forest.getTransitionFunction(
				Variable.getVariable("huc"), "delc").toString());

		// checkTerminalInitialLoading(forest);
	}

	/**
	 * 
	 * Check if terminal and initial conditions are correctly loaded
	 * 
	 * @param forest
	 *          the loaded forest to test
	 * 
	 */
	protected void checkTerminalInitialLoading(StaticForest forest) {
		for (int testIndex = 0; testIndex < 100; testIndex++) {
			Context testedState = forest.getInitialState(new Random(0));
			assertEquals(false, forest.isTerminal(testedState));
			testedState.put(Variable.getVariable("huc"), "1");
			assertEquals(true, forest.isTerminal(testedState));
		}
	}

	/**
	 * 
	 * Check the legend of a forest against a faked legend
	 * 
	 * @param expectedLegend
	 *          the fake legend
	 * @param variableOrdering
	 *          the variable ordering to check against
	 * 
	 */
	private void checkLegend(Map<Variable, Integer> expectedLegend,
			VariableOrdering variableOrdering) {
		for (Map.Entry<Variable, Integer> entry : expectedLegend.entrySet()) {
			assertEquals((int) entry.getValue(), variableOrdering.indexOf(entry
					.getKey()));
		}
	}

	/**
	 * 
	 * Test loading the coffeerobot problem
	 * 
	 */
	public void testToNoisyMethod() {
		StaticForest forest = new StaticForest(
				getRootFromFile("data/forests/coffeerobot.forest"));
		forest.toNoisy(0.1);
		for (String action : forest.getActionList()) {
			for (Variable variable : forest.getVariableSet()) {
				for (IPartition<IProbabilities<String>> cpd : forest
						.getTransitionFunction(variable, action).partitions()) {
					assertTrue(cpd.get().getValueSet().size() > 1);
				}
			}
		}
	}
}
