/*
 * 
 * Created on 4 janv. 2006
 *
 */
package animatlab.kodamat.reinforcementlearning.model.forest.test;

import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Random;

import junit.framework.TestCase;
import animatlab.kodamat.core.io.Log;
import animatlab.kodamat.core.io.XmlUtils;
import animatlab.kodamat.problems.ProblemRunner;
import animatlab.kodamat.reinforcementlearning.agent.ModelBasedAgent;
import animatlab.kodamat.reinforcementlearning.core.acting.RandomActing;
import animatlab.kodamat.reinforcementlearning.core.trajectory.Trajectory;
import animatlab.kodamat.reinforcementlearning.model.ILearningModel;
import animatlab.kodamat.reinforcementlearning.model.forest.VariableActionLearningForest;
import animatlab.kodamat.reinforcementlearning.model.forest.VariableLearningForest;
import animatlab.kodamat.reinforcementlearning.planning.spisvi.StaticSpiSvi;
import animatlab.kodamat.reinforcementlearning.planning.stats.RelativeErrorMeasure;
import animatlab.kodamat.supervisedlearning.entropy.Chi2Measure;
import animatlab.kodamat.supervisedlearning.learningtree.SID4Tree;

/**
 * 
 * Test the learning method of learning forests in deterministic problem
 * 
 * @author Thomas Degris
 * 
 */
public class TestLearningForest extends TestCase {
	/**
	 * 
	 * World model used in the experiment
	 * 
	 */
	protected StaticSpiSvi coffeeRobotProblem;

	/**
	 * @see junit.framework.TestCase#setUp()
	 */
	@Override
	public void setUp() {
		coffeeRobotProblem = new StaticSpiSvi(XmlUtils
				.getRootFromFile("data/forests/coffeerobot.spisvi.agent"));
	}

	/**
	 * 
	 * Test the learning forest on CoffeeRobot
	 * 
	 * @param forestTested
	 *          the forest to test
	 * 
	 */
	private void testLearningForest(ILearningModel forestTested) {
		final Random random = new Random(0);
		ModelBasedAgent agent = new ModelBasedAgent(forestTested, new RandomActing(
				random, coffeeRobotProblem.getModel().getActionList()));
		ProblemRunner problemRunner = new ProblemRunner(random, coffeeRobotProblem
				.getModel(), agent, Log.sysout);
		Trajectory trajectoryLearning = new Trajectory(forestTested
				.getVariableOrdering(), forestTested.getActionList());
		problemRunner.registerListener(trajectoryLearning);
		problemRunner.runProblem(4000);
		RelativeErrorMeasure valueFunctionMeasure = new RelativeErrorMeasure();
		Map<String, Double> values = new LinkedHashMap<String, Double>();
		valueFunctionMeasure.setMeasureData(coffeeRobotProblem.getModel(),
				coffeeRobotProblem, agent, null);
		valueFunctionMeasure.computeMeasures(values);
		assert (values.get(RelativeErrorMeasure.VALFUNAVE_LABEL) < 0.2);
	}

	/**
	 * 
	 * Test the learning forest on Mazespace
	 * 
	 */
	public void testVariableLearningForestWithCoffeeRobot() {
		ILearningModel forestTested = new VariableLearningForest(coffeeRobotProblem
				.getModel().getActionList(), coffeeRobotProblem.getModel()
				.getVariableOrdering(), new SID4Tree.Prob<String>(new Chi2Measure(1.0),
				coffeeRobotProblem.getModel().getVariableOrdering()), null);
		testLearningForest(forestTested);
	}

	/**
	 * 
	 * Test the learning forest on Mazespace
	 * 
	 */
	public void testVariableActionLearningForestWithCoffeeRobot() {
		ILearningModel forestTested = new VariableActionLearningForest(
				coffeeRobotProblem.getModel().getActionList(), coffeeRobotProblem
						.getModel().getVariableOrdering(), new SID4Tree.Prob<String>(
						new Chi2Measure(1.0), coffeeRobotProblem.getModel()
								.getVariableOrdering()), null);
		testLearningForest(forestTested);
	}

}
