package animatlab.kodamat.reinforcementlearning.planning;

import java.util.List;
import java.util.Map;
import java.util.Set;

import animatlab.kodamat.core.IModel;
import animatlab.kodamat.core.functions.IFunction;
import animatlab.kodamat.core.io.Log;
import animatlab.kodamat.core.utils.Context;
import animatlab.kodamat.reinforcementlearning.model.ILearningModel;

/**
 * 
 * Dummy class for the IIncrementalPlanning interface returning null for the set
 * of action-value function
 * 
 * @author Thomas Degris
 * 
 */
public class DummyPlanning implements IIncrementalPlanning {
	/**
	 * 
	 * Model (not) used by this planning instance
	 * 
	 */
	final private IModel model;

	/**
	 * 
	 * Constructor
	 * 
	 */
	public DummyPlanning() {
		model = null;
	}

	/**
	 * 
	 * Constructor
	 * 
	 * @param model
	 *          model (not) used by this planning instance
	 * 
	 */
	private DummyPlanning(IModel model) {
		this.model = model;
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.planning.IIncrementalPlanning#getActionValueFunctions()
	 */
	public Map<String, IFunction<Double>> getActionValueFunctions() {
		return null;
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.planning.IIncrementalPlanning#getIncrementalPlanning(Context,
	 *      java.lang.String, Context, java.util.List)
	 */
	public Map<String, IFunction<Double>> getIncrementalPlanning(
			Context lastState, String action, Context currentState,
			List<Double> reward) {
		return null;
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.planning.IIncrementalPlanning#newInstance(animatlab.kodamat.reinforcementlearning.model.ILearningModel)
	 */
	public IIncrementalPlanning newInstance(ILearningModel learningModel) {
		return new DummyPlanning(learningModel);
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.planning.IPlanning#getModel()
	 */
	public IModel getModel() {
		return model;
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.planning.IPlanning#getName()
	 */
	public String getName() {
		return "dummy";
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.planning.IPlanning#getOptimalPolicy()
	 */
	public IFunction<Set<String>> getOptimalPolicy() {
		return null;
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.planning.IPlanning#getOptimalValueFunction()
	 */
	public IFunction<Double> getOptimalValueFunction() {
		return null;
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.planning.IPlanning#solve()
	 */
	public void solve() {
		// Nothing to solve
	}

	/**
	 * @see animatlab.kodamat.core.io.ILog#setLog(animatlab.kodamat.core.io.Log)
	 */
	public void setLog(Log log) {
		// Nothing to log
	}

}
