/*
 * 
 * Created on 8 janv. 2006
 *
 */
package animatlab.kodamat.reinforcementlearning.planning;

import java.util.List;
import java.util.Map;

import animatlab.kodamat.core.functions.IFunction;
import animatlab.kodamat.core.utils.Context;
import animatlab.kodamat.reinforcementlearning.model.ILearningModel;

/**
 * 
 * Incremental planning that can be used by an implementation of Acting
 * 
 * @author Thomas Degris
 * 
 */
public interface IIncrementalPlanning extends IPlanning {
	/**
	 * 
	 * Run incremental planning and return the computed quality function tables
	 * 
	 * @param lastState
	 *          last state the agent was
	 * @param action
	 *          last action the agent did
	 * @param currentState
	 *          current state of the agent
	 * @param reward
	 *          reward received by the agent
	 * @return a map of quality functions
	 * 
	 */
	Map<String, IFunction<Double>> getIncrementalPlanning(Context lastState,
			String action, Context currentState, List<Double> reward);

	/**
	 * 
	 * Create a new instance of this
	 * 
	 * @param learningModel
	 *          the model to use
	 * @return a new instance of this
	 * 
	 */
	IIncrementalPlanning newInstance(ILearningModel learningModel);

	/**
	 * 
	 * Get the set of action-value functions
	 * 
	 * @return the set of action-value functions
	 * 
	 */
	Map<String, IFunction<Double>> getActionValueFunctions();
}
