package animatlab.kodamat.reinforcementlearning.planning;

import animatlab.kodamat.core.utils.Parameters;
import animatlab.kodamat.reinforcementlearning.model.forest.AbstractForest;
import animatlab.kodamat.reinforcementlearning.planning.guestrin.StaticGuestrin;
import animatlab.kodamat.reinforcementlearning.planning.spisvi.SpiSvi;
import animatlab.kodamat.reinforcementlearning.planning.spudd.StaticSpudd;

/**
 * 
 * Some usefull functions to use planning algorithms
 * 
 * @author Thomas Degris
 * 
 */
public class PlanningUtils {

	/**
	 * 
	 * Get the default planning method for a given problem
	 * 
	 * @param model
	 *          the model to get the planning method for
	 * @return the default planning method for
	 * 
	 */
	static public IPlanning getDefaultPlanning(AbstractForest model) {
		String planningMethod = model.getParameters().getKeyS(
				Parameters.PARAM_KEY_PLANNING_METHOD);
		return getPlanningMethod(model, planningMethod);
	}

	/**
	 * 
	 * Create a new planning method, given its name and a model
	 * 
	 * @param model
	 *          the model the planning method will use
	 * @param planningMethod
	 *          the name of the planning method
	 * @return the new planning method
	 * 
	 */
	public static IPlanning getPlanningMethod(AbstractForest model,
			String planningMethod) {
		if (planningMethod == "none") {
			return null;
		} else if (StaticGuestrin.PLANNING_METHOD_NAME.equals(planningMethod)) {
			return new StaticGuestrin(model);
		} else if (StaticSpudd.PLANNING_METHOD_NAME.equals(planningMethod)) {
			return new StaticSpudd(model);
		}
		return new SpiSvi(model);
	}

}
