package animatlab.kodamat.reinforcementlearning.planning.guestrin.linearprogramming;

import java.io.File;
import java.io.IOException;

import animatlab.kodamat.core.io.Exec;

/**
 * 
 * An incremental version of the LP definition
 * 
 * @author Thomas Degris
 * 
 */
public class IncrementalLPProblem extends LPProblem {
	/**
	 * 
	 * Last computed solution in MPS format
	 * 
	 */
	File solFile = null;

	/**
	 * 
	 * Constructor
	 * 
	 */
	public IncrementalLPProblem() {
		super();
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.planning.guestrin.linearprogramming.LPProblem#runSolver(java.io.File,
	 *      java.io.File)
	 */
	@Override
	protected void runSolver(File inputFile, File outputFile) {
		File mpsFile = null;
		try {
			mpsFile = File.createTempFile("spiti", ".mps");
		} catch (IOException e) {
			e.printStackTrace();
			System.exit(1);
			return;
		}
		StringBuilder command = new StringBuilder(GLPSOL
				+ " --presol --orig --cpxlp --check " + "--wmps "
				+ mpsFile.getAbsolutePath() + " " + inputFile.getAbsolutePath());
		int exitCode = Exec.exec(command.toString(), log);
		File oldSolFile = solFile;
		exitCode = runIncrementalSolver(outputFile, mpsFile);
		if ((exitCode != 0) || (!solFile.canRead())) {
			resetSolution();
			exitCode = runIncrementalSolver(outputFile, mpsFile);
		}
		assert (exitCode == 0);
		mpsFile.delete();
		if (oldSolFile != null) {
			oldSolFile.delete();
		}
	}

	/**
	 * 
	 * Run the solver starting with the solution
	 * 
	 * @param outputFile
	 *          output file where the solution is read
	 * @param mpsFile
	 *          file declaring the problem
	 * @return the exit code of the solver
	 * 
	 */
	private int runIncrementalSolver(File outputFile, File mpsFile) {
		StringBuilder command;
		int exitCode;
		String timeoutOption = getTimeoutOption();
		command = new StringBuilder(GLPSOL + " --nopresol --nosteep --mps "
				+ timeoutOption + mpsFile.getAbsolutePath() + " -o "
				+ outputFile.getAbsolutePath());
		if (solFile != null) {
			command.append(" --bas " + solFile.getAbsolutePath());
		}
		try {
			solFile = File.createTempFile("spiti", ".sol.mps");
		} catch (IOException e) {
			e.printStackTrace();
			System.exit(1);
		}
		command.append(" --wbas " + solFile.getAbsolutePath());
		exitCode = Exec.exec(command.toString(), log);
		return exitCode;
	}

	/**
	 * 
	 * Reset the last computed solution
	 * 
	 */
	public void resetSolution() {
		if (solFile != null) {
			solFile.delete();
		}
		solFile = null;
	}

	/**
	 * @see java.lang.Object#finalize()
	 */
	@Override
	public void finalize() {
		resetSolution();
	}
}
