package animatlab.kodamat.reinforcementlearning.planning.guestrin.linearprogramming;

import java.util.Set;

/**
 * 
 * Constraint in the LP definition
 * 
 * @author Thomas Degris
 * 
 */
public class LPConstraint extends LPConstraintAbstract {
	/**
	 * 
	 * the variable defined by this constraint
	 * 
	 */
	final private LPNamedVariable definedVariable;

	/**
	 * 
	 * Constructor
	 * 
	 * @param definedVariable
	 *          the variable defined by this constraint
	 * @param operator
	 *          operator used in the constraint
	 * @param rightArg
	 *          the value of the right arguments
	 * @param namedVariables
	 *          the other variables used in the constraint
	 * 
	 */
	public LPConstraint(LPNamedVariable definedVariable, String operator,
			double rightArg, Set<LPNamedVariable> namedVariables) {
		super(operator, rightArg, namedVariables);
		this.definedVariable = definedVariable;
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.planning.guestrin.linearprogramming.LPConstraintAbstract#definedVariableToLPString()
	 */
	@Override
	protected String definedVariableToLPString() {
		return " - " + definedVariable.toLPString();
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.planning.guestrin.linearprogramming.LPConstraintAbstract#definedVariableToString()
	 */
	@Override
	protected String definedVariableToString() {
		return " - " + definedVariable;
	}
}
