package animatlab.kodamat.reinforcementlearning.planning.guestrin.linearprogramming;

import java.util.Set;

/**
 * 
 * Abstract LP constraint
 * 
 * @author Thomas Degris
 * 
 */
public abstract class LPConstraintAbstract extends LPConstraintLine {
	/**
	 * 
	 * operator used in the constraint
	 * 
	 */
	final private String operator;

	/**
	 * 
	 * the value of the right arguments
	 * 
	 */
	final private double rightArg;

	/**
	 * 
	 * the other variables used in the constraint
	 * 
	 */
	final private Set<LPNamedVariable> namedVariables;

	/**
	 * 
	 * Constructor
	 * 
	 * @param operator
	 *          operator used in the constraint
	 * @param rightArg
	 *          the value of the right arguments
	 * @param namedVariables
	 *          the other variables used in the constraint
	 * 
	 */
	public LPConstraintAbstract(String operator, double rightArg,
			Set<LPNamedVariable> namedVariables) {
		super(true);
		this.operator = operator;
		this.rightArg = rightArg;
		this.namedVariables = namedVariables;
	}

	/**
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		StringBuilder result = new StringBuilder();
		String separator = " + ";
		if (namedVariables != null) {
			for (LPNamedVariable variable : namedVariables) {
				result.append(variable);
				result.append(separator);
			}
			result.delete(result.length() - separator.length(), result.length());
		}
		result.append(definedVariableToString());
		result.append(" " + operator + " ");
		result.append(rightArg);
		return result.toString();
	}

	/**
	 * 
	 * @return the LP representation of this constraint
	 * 
	 */
	abstract protected String definedVariableToLPString();

	/**
	 * 
	 * @return the LP representation of this constraint
	 * 
	 */
	abstract protected String definedVariableToString();

	/**
	 * @see animatlab.kodamat.reinforcementlearning.planning.guestrin.linearprogramming.LPConstraintLine#toLPString()
	 */
	@Override
	public String toLPString() {
		StringBuilder result = new StringBuilder();
		String separator = " + ";
		if ((namedVariables != null) && !namedVariables.isEmpty()) {
			for (LPNamedVariable variable : namedVariables) {
				String variableLPString = variable.toLPString();
				if (variableLPString != null) {
					result.append(variable.toLPString());
					result.append(separator);
				}
			}
			if (result.length() > 0) {
				result.delete(result.length() - separator.length(), result.length());
			}
		}
		result.append(definedVariableToLPString());
		result.append(" " + operator + " ");
		result.append(rightArg);
		return result.toString();
	}

}
