package animatlab.kodamat.reinforcementlearning.planning.guestrin.linearprogramming;

import java.util.Set;

/**
 * 
 * Constraint in the LP definition with multiple defined variables definition
 * (meaning with multiple minus sign)
 * 
 * @author Thomas Degris
 * 
 */
public class LPConstraintDefinedVariablesList extends LPConstraintAbstract {
	/**
	 * 
	 * the variable defined by this constraint
	 * 
	 */
	final private Set<LPNamedVariable> definedVariables;

	/**
	 * 
	 * Constructor
	 * 
	 * @param definedVariables
	 *          set of defined variables
	 * @param operator
	 *          operator used in the constraint
	 * @param rightArg
	 *          the value of the right arguments
	 * @param namedVariables
	 *          the other variables used in the constraint
	 * 
	 */
	public LPConstraintDefinedVariablesList(
			Set<LPNamedVariable> definedVariables, String operator, double rightArg,
			Set<LPNamedVariable> namedVariables) {
		super(operator, rightArg, namedVariables);
		this.definedVariables = definedVariables;
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.planning.guestrin.linearprogramming.LPConstraintAbstract#definedVariableToLPString()
	 */
	@Override
	protected String definedVariableToLPString() {
		StringBuilder result = new StringBuilder();
		for (LPNamedVariable variable : definedVariables) {
			String variableLPString = variable.toLPString();
			if (variableLPString != null) {
				result.append(" - ");
				result.append(variableLPString);
			}
		}
		return result.toString();
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.planning.guestrin.linearprogramming.LPConstraintAbstract#definedVariableToString()
	 */
	@Override
	protected String definedVariableToString() {
		StringBuilder result = new StringBuilder();
		for (LPNamedVariable variable : definedVariables) {
			result.append(" - ");
			result.append(variable.toString());
		}
		return result.toString();
	}

}
