package animatlab.kodamat.reinforcementlearning.planning.guestrin.linearprogramming;

/**
 * 
 * Symbolic name in the LP problem
 * 
 * @author Thomas Degris
 * 
 */
public class LPNamedVariable extends LPVariable {

	/**
	 * 
	 * Bound of this variable
	 * 
	 */
	public enum VariableBound {
		/**
		 * 
		 * No bound
		 * 
		 */
		FREE,

		/**
		 * 
		 * The value is solved
		 * 
		 */
		EQUAL,

		/**
		 * 
		 * The value is over or equal to a value
		 * 
		 */
		OVER
	}

	/**
	 * 
	 * Min bound of the variable
	 * 
	 */
	private double value = Double.NaN;

	/**
	 * 
	 * The type of bound of this variable
	 * 
	 */
	private VariableBound boundType = VariableBound.FREE;

	/**
	 * 
	 * Name of the variable
	 * 
	 */
	final protected String name;

	/**
	 * 
	 * Constructor
	 * 
	 * @param name
	 *          name of the variable
	 * 
	 */
	public LPNamedVariable(String name) {
		this.name = name;
	}

	/**
	 * @see java.lang.Object#hashCode()
	 */
	@Override
	public int hashCode() {
		return name.hashCode();
	}

	/**
	 * @see java.lang.Object#equals(java.lang.Object)
	 */
	@Override
	public boolean equals(Object object) {
		if (object == null) {
			return false;
		}
		if (object == this) {
			return true;
		}
		if (getClass() != object.getClass()) {
			return false;
		}
		return name.equals(((LPNamedVariable) object).name);
	}

	/**
	 * 
	 * Set the bound of this variable
	 * 
	 * @param infBound
	 *          the value of the bound
	 * 
	 */
	public void setInfBound(double infBound) {
		boundType = VariableBound.OVER;
		if (Double.isNaN(value)) {
			value = infBound;
		} else {
			value = Math.max(value, infBound);
		}
	}

	/**
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		if (!Double.isNaN(value)) {
			return name + "(=" + value + ")";
		}
		if (!Double.isNaN(solution)) {
			return name + "(=" + solution + ")";
		}
		return name + "(=?)";
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.planning.guestrin.linearprogramming.LPVariable#toLPString()
	 */
	@Override
	public String toLPString() {
		return name;
	}

	/**
	 * 
	 * @return a string representing the bound of this variable
	 * 
	 */
	public String toBoundString() {
		StringBuilder result = new StringBuilder(name);
		result.append(" ");
		switch (boundType) {
		case FREE:
			result.append("free");
			break;
		case OVER:
			result.append(">= ");
			result.append(value);
			break;
		case EQUAL:
			result.append("= ");
			result.append(value);
			break;
		}
		return result.toString();
	}

	/**
	 * 
	 * Set a value to this variable
	 * 
	 * @param value
	 *          the value used as solution
	 * 
	 */
	public void solve(double value) {
		boundType = VariableBound.EQUAL;
		this.value = value;
	}
}
