package animatlab.kodamat.reinforcementlearning.planning.guestrin.tests;

import static animatlab.kodamat.core.io.XmlUtils.getRoot;

import java.util.List;
import java.util.Map;
import java.util.Set;

import junit.framework.TestCase;
import animatlab.kodamat.core.functions.FunctionsSum;
import animatlab.kodamat.core.functions.IFunction;
import animatlab.kodamat.core.functions.IPartitionFunction;
import animatlab.kodamat.core.functions.UnaryOperatorPolicy;
import animatlab.kodamat.core.io.DataLoadingPolicy;
import animatlab.kodamat.core.io.XmlUtils;
import animatlab.kodamat.core.proba.IProbabilities;
import animatlab.kodamat.core.rules.PlainRules;
import animatlab.kodamat.core.rules.Rule;
import animatlab.kodamat.core.trees.AppendPolicy;
import animatlab.kodamat.core.trees.Tree;
import animatlab.kodamat.core.utils.Parameters;
import animatlab.kodamat.core.utils.Context;
import animatlab.kodamat.core.utils.Variable;
import animatlab.kodamat.core.utils.VariableOrdering;
import animatlab.kodamat.problems.generated.sysadmin.Disconnected;
import animatlab.kodamat.problems.generated.sysadmin.SysAdminProblem;
import animatlab.kodamat.reinforcementlearning.core.acting.Greedy;
import animatlab.kodamat.reinforcementlearning.model.AbstractFactoredModel;
import animatlab.kodamat.reinforcementlearning.model.forest.StaticForest;
import animatlab.kodamat.reinforcementlearning.model.rules.StaticRulesModel;
import animatlab.kodamat.reinforcementlearning.planning.guestrin.DecisionListPolicy;
import animatlab.kodamat.reinforcementlearning.planning.guestrin.GuestrinPlanning;
import animatlab.kodamat.reinforcementlearning.planning.guestrin.RuleBasedBackproj;
import animatlab.kodamat.reinforcementlearning.planning.guestrin.TreeBasedBackproj;
import animatlab.kodamat.reinforcementlearning.planning.guestrin.linearprogramming.LPNamedVariable;
import animatlab.kodamat.reinforcementlearning.planning.guestrin.linearprogramming.LPProblem;
import animatlab.kodamat.reinforcementlearning.planning.guestrin.linearprogramming.LPVariable;
import animatlab.kodamat.reinforcementlearning.planning.spisvi.SpiSvi;
import animatlab.kodamat.reinforcementlearning.planning.spisvi.StaticSpiSvi;
import animatlab.kodamat.reinforcementlearning.planning.spisvi.tests.TestSpiSvi;
import animatlab.kodamat.supervisedlearning.core.Example;
import animatlab.kodamat.supervisedlearning.entropy.KSMeasure;
import animatlab.kodamat.supervisedlearning.learningtree.ID3Tree;

/**
 * 
 * Test case to test Guestrin planning algorithm
 * 
 * @author Thomas Degris
 * 
 */
public class TestGuestrin extends TestCase {
	/**
	 * 
	 * Forest used testing purpose
	 * 
	 */
	static private final StaticRulesModel dummyForest = new StaticRulesModel(
			XmlUtils.getRoot("<fst:rules>" + "  <dtr:legend>"
					+ "    <spi:variable legend='a' index='0'/>"
					+ "    <spi:variable legend='b' index='1'/>"
					+ "    <spi:variable legend='c' index='2'/>" + "  </dtr:legend>"
					+ "  <dtr:variablevalueset>" + "    <dtr:valueset variable='a'>"
					+ "      <dtr:value value='0'/>" + "      <dtr:value value='1'/>"
					+ "    </dtr:valueset>" + "    <dtr:valueset variable='b'>"
					+ "      <dtr:value value='0'/>" + "      <dtr:value value='1'/>"
					+ "    </dtr:valueset>" + "    <dtr:valueset variable='c'>"
					+ "      <dtr:value value='0'/>" + "      <dtr:value value='1'/>"
					+ "    </dtr:valueset>" + "    <dtr:valueset variable='d'>"
					+ "      <dtr:value value='0'/>" + "      <dtr:value value='1'/>"
					+ "    </dtr:valueset>" + "  </dtr:variablevalueset>"
					+ "</fst:rules>"), null);

	/**
	 * 
	 * Forest used testing purpose
	 * 
	 */
	static private final StaticRulesModel plumbingForest = new StaticRulesModel(
			XmlUtils.getRoot("<fst:rules>" + "  <dtr:legend>"
					+ "    <spi:variable legend='Painting' index='0'/>"
					+ "    <spi:variable legend='Plumbing' index='1'/>"
					+ "    <spi:variable legend='Electrical' index='2'/>"
					+ "  </dtr:legend>" + "  <fst:transitions>"
					+ "   <fst:action name='paint'>"
					+ "     <fst:variable variable='Painting'>" + "       <rdr:rules>"
					+ "         <rdr:rule>"
					+ "           <rdr:cond variable='Electrical' value='Not done'/>"
					+ "           <spi:probability probability='1.0' value='Not done'/>"
					+ "         </rdr:rule>" + "         <rdr:rule>"
					+ "           <rdr:cond variable='Electrical' value='Done'/>"
					+ "           <rdr:cond variable='Plumbing' value='Not done'/>"
					+ "           <spi:probability probability='1.0' value='Not done'/>"
					+ "         </rdr:rule>" + "         <rdr:rule>"
					+ "           <rdr:cond variable='Electrical' value='Done'/>"
					+ "           <rdr:cond variable='Plumbing' value='Done'/>"
					+ "           <spi:probability probability='0.05' value='Not done'/>"
					+ "           <spi:probability probability='0.95' value='Done'/>"
					+ "         </rdr:rule>" + "       </rdr:rules>"
					+ "     </fst:variable>" + "   </fst:action>"
					+ "   <fst:action name='notpaint'>"
					+ "     <fst:variable variable='Painting'>" + "       <rdr:rules>"
					+ "         <rdr:rule>"
					+ "           <rdr:cond variable='Electrical' value='Not done'/>"
					+ "           <spi:probability probability='1.0' value='Not done'/>"
					+ "         </rdr:rule>" + "         <rdr:rule>"
					+ "           <rdr:cond variable='Electrical' value='Done'/>"
					+ "           <rdr:cond variable='Plumbing' value='Not done'/>"
					+ "           <spi:probability probability='1.0' value='Not done'/>"
					+ "         </rdr:rule>" + "         <rdr:rule>"
					+ "           <rdr:cond variable='Electrical' value='Done'/>"
					+ "           <rdr:cond variable='Plumbing' value='Done'/>"
					+ "           <rdr:cond variable='Painting' value='Not done'/>"
					+ "           <spi:probability probability='1.0' value='Not done'/>"
					+ "         </rdr:rule>" + "         <rdr:rule>"
					+ "           <rdr:cond variable='Electrical' value='Done'/>"
					+ "           <rdr:cond variable='Plumbing' value='Done'/>"
					+ "           <rdr:cond variable='Painting' value='Done'/>"
					+ "           <spi:probability probability='0.1' value='Not done'/>"
					+ "           <spi:probability probability='0.9' value='Done'/>"
					+ "         </rdr:rule>" + "       </rdr:rules>"
					+ "     </fst:variable>" + "   </fst:action>"
					+ "  </fst:transitions>" + "</fst:rules>"), null);

	/**
	 * 
	 * Build a rules set from a xml definition
	 * 
	 * @param xmlString
	 *          the xml definition of the set of rules
	 * @param variableOrdering
	 *          the legend to use to load the rule
	 * 
	 * @return the new set of rules
	 * 
	 */
	private PlainRules<Double> getRules(String xmlString,
			VariableOrdering variableOrdering) {
		return new PlainRules<Double>(getRoot(xmlString),
				DataLoadingPolicy.doubleLoadingPolicy);
	}

	/**
	 * 
	 * Test the back projection operator on set of rules
	 * 
	 */
	public void testTreeBackProj() {
		GuestrinPlanning lpPlanning = new GuestrinPlanning(TestSpiSvi.xyzForest);
		lpPlanning.gamma = 0.9;
		lpPlanning.setPolicy(new RuleBasedBackproj(TestSpiSvi.xyzForest));
		PlainRules<Double> expectedRules = getRules("<rdr:rules>"
				+ "  <rdr:rule data='19'>"
				+ "    <rdr:cond variable='z' value='true'/>" + "  </rdr:rule>"
				+ "  <rdr:rule data='8.1'>"
				+ "    <rdr:cond variable='z' value='false'/>"
				+ "    <rdr:cond variable='y' value='true'/>" + "  </rdr:rule>"
				+ "  <rdr:rule data='0'>"
				+ "    <rdr:cond variable='z' value='false'/>"
				+ "    <rdr:cond variable='y' value='false'/>" + "  </rdr:rule>"
				+ "</rdr:rules>", TestSpiSvi.xyzForest.getVariableOrdering());
		Tree<Double> reward = (Tree<Double>) TestSpiSvi.xyzForest
				.getRewardFunction("simpleaction").functions().get(0);
		PlainRules<Double> backProjection = lpPlanning.rulesBackProj(
				"simpleaction", reward);
		discountAndAddReward(reward, backProjection, TestSpiSvi.xyzForest
				.getValueSets(), 0.9);

		assertEquals(expectedRules, backProjection);

		TreeBasedBackproj treeBasedPlanning = new TreeBasedBackproj(
				TestSpiSvi.xyzForest, 0.9);
		Tree<Double> treeBackProjection = treeBasedPlanning.treeBackProj(
				"simpleaction", reward);
		discountAndAddReward(lpPlanning, treeBackProjection, "simpleaction", 0.9);

		backProjection = lpPlanning.rulesBackProj("simpleaction",
				treeBackProjection);
		expectedRules = getRules("<rdr:rules>" + "  <rdr:rule data='19'>"
				+ "    <rdr:cond variable='z' value='true'/>" + "  </rdr:rule>"
				+ "  <rdr:rule data='17.91'>"
				+ "    <rdr:cond variable='z' value='false'/>"
				+ "    <rdr:cond variable='y' value='true'/>" + "  </rdr:rule>"
				+ "  <rdr:rule data='7.29'>"
				+ "    <rdr:cond variable='z' value='false'/>"
				+ "    <rdr:cond variable='y' value='false'/>"
				+ "    <rdr:cond variable='x' value='true'/>" + "  </rdr:rule>"
				+ "</rdr:rules>", TestSpiSvi.xyzForest.getVariableOrdering());

		assertEquals(expectedRules, backProjection);
	}

	/**
	 * 
	 * Discount a function and then add a reward function
	 * 
	 * @param lpPlanning
	 *          algorithm to use to compute the regresssion
	 * @param backProjection
	 *          the function to manipulate
	 * @param action
	 *          the action to get the reward
	 * @param gamma
	 *          gamma to use to discount
	 * 
	 */
	public void discountAndAddReward(GuestrinPlanning lpPlanning,
			Tree<Double> backProjection, String action, double gamma) {
		// Applying discounted factor
		new UnaryOperatorPolicy.MultiplyOperator(gamma).applyPolicy(backProjection);
		AbstractFactoredModel model = lpPlanning.getModel();
		// Adding reward function
		for (IFunction<Double> reward : model.getRewardFunction(action).functions()) {
			backProjection
					.append((Tree<Double>) reward, AppendPolicy.sumPolicy, null);
		}
	}

	/**
	 * 
	 * Test the back projection operator on set of rules
	 * 
	 */
	public void testRulesBackProj() {
		GuestrinPlanning ruleBasedFactoredLP = new GuestrinPlanning(
				TestSpiSvi.xyzForest);
		ruleBasedFactoredLP.setPolicy(new RuleBasedBackproj(TestSpiSvi.xyzForest));
		PlainRules<Double> expectedRules = getRules("<rdr:rules>"
				+ "  <rdr:rule data='19'>"
				+ "    <rdr:cond variable='z' value='true'/>" + "  </rdr:rule>"
				+ "  <rdr:rule data='8.1'>"
				+ "    <rdr:cond variable='z' value='false'/>"
				+ "    <rdr:cond variable='y' value='true'/>" + "  </rdr:rule>"
				+ "  <rdr:rule data='0'>"
				+ "    <rdr:cond variable='z' value='false'/>"
				+ "    <rdr:cond variable='y' value='false'/>" + "  </rdr:rule>"
				+ "</rdr:rules>", TestSpiSvi.xyzForest.getVariableOrdering());
		IPartitionFunction<Double> reward = (IPartitionFunction<Double>) ruleBasedFactoredLP
				.getModel().getRewardFunction("simpleaction").functions().get(0);
		PlainRules<Double> backProjection = ruleBasedFactoredLP.rulesBackProj(
				"simpleaction", reward);
		discountAndAddReward(reward, backProjection, TestSpiSvi.xyzForest
				.getValueSets(), 0.9);

		assertEquals(expectedRules, backProjection);

		PlainRules<Double> ruleBackProjection = ruleBasedFactoredLP.rulesBackProj(
				"simpleaction", reward);
		discountAndAddReward(ruleBasedFactoredLP, ruleBackProjection,
				"simpleaction", 0.9);
		backProjection = ruleBasedFactoredLP.rulesBackProj("simpleaction",
				ruleBackProjection);
		expectedRules = getRules("<rdr:rules>" + "  <rdr:rule data='19'>"
				+ "    <rdr:cond variable='z' value='true'/>" + "  </rdr:rule>"
				+ "  <rdr:rule data='17.91'>"
				+ "    <rdr:cond variable='z' value='false'/>"
				+ "    <rdr:cond variable='y' value='true'/>" + "  </rdr:rule>"
				+ "  <rdr:rule data='7.29'>"
				+ "    <rdr:cond variable='z' value='false'/>"
				+ "    <rdr:cond variable='y' value='false'/>"
				+ "    <rdr:cond variable='x' value='true'/>" + "  </rdr:rule>"
				+ "</rdr:rules>", TestSpiSvi.xyzForest.getVariableOrdering());

		assertEquals(expectedRules, backProjection);
	}

	/**
	 * 
	 * Discount a function and then add a reward function
	 * 
	 * @param ruleBasedFactoredLP
	 *          planning algorithm to use to compute the regression
	 * @param backProjection
	 *          the function to manipulate
	 * @param action
	 *          the action to get the reward
	 * @param gamma
	 *          gamma to use to discount
	 * 
	 */
	private void discountAndAddReward(GuestrinPlanning ruleBasedFactoredLP,
			PlainRules<Double> backProjection, String action, double gamma) {
		// Applying discounted factor
		new UnaryOperatorPolicy.MultiplyOperator(gamma).applyPolicy(backProjection);
		// Adding reward functions
		FunctionsSum reward = ruleBasedFactoredLP.getModel().getRewardFunction(
				action);
		for (IFunction<Double> rj : reward.functions()) {
			backProjection.combine((IPartitionFunction<Double>) rj,
					PlainRules.doubleSumPolicy, ruleBasedFactoredLP.getModel()
							.getValueSets());
		}
	}

	/**
	 * 
	 * Discount a function and then add a reward function
	 * 
	 * @param reward
	 *          the reward function to add
	 * @param backProjection
	 *          the function to manipulate
	 * @param valuesSet
	 *          set of values for the variables
	 * @param gamma
	 *          gamma to use to discount
	 * @return the result of the computation
	 * 
	 */
	private PlainRules<Double> discountAndAddReward(IFunction<Double> reward,
			PlainRules<Double> backProjection, Map<Variable, Set<String>> valuesSet,
			double gamma) {
		// Applying discounted factor
		new UnaryOperatorPolicy.MultiplyOperator(gamma).applyPolicy(backProjection);
		// Adding reward function
		backProjection.combine((IPartitionFunction<Double>) reward,
				PlainRules.doubleSumPolicy, valuesSet);
		return backProjection;
	}

	/**
	 * 
	 * Test the RuleMaxOut algorithm
	 * 
	 */
	public void testRuleMaxOut() {
		GuestrinPlanning ruleBasedFactoredLP = new GuestrinPlanning(dummyForest);
		LPProblem problem = new LPProblem();
		PlainRules<LPVariable> rules = new PlainRules<LPVariable>(
				getRoot("<rdr:rules>" + "  <rdr:rule data='u1'>"
						+ "    <rdr:cond variable='a' value='1'/>"
						+ "    <rdr:cond variable='b' value='1'/>" + "  </rdr:rule>"
						+ "  <rdr:rule data='u2'>"
						+ "    <rdr:cond variable='a' value='1'/>"
						+ "    <rdr:cond variable='b' value='1'/>"
						+ "    <rdr:cond variable='c' value='1'/>" + "  </rdr:rule>"
						+ "  <rdr:rule data='u3'>"
						+ "    <rdr:cond variable='a' value='1'/>"
						+ "    <rdr:cond variable='b' value='0'/>" + "  </rdr:rule>"
						+ "  <rdr:rule data='u4'>"
						+ "    <rdr:cond variable='a' value='1'/>"
						+ "    <rdr:cond variable='b' value='1'/>"
						+ "    <rdr:cond variable='c' value='0'/>" + "  </rdr:rule>"
						+ "</rdr:rules>"), problem.loadingPolicy);
		PlainRules<LPVariable> result = ruleBasedFactoredLP.ruleMaxOut(problem,
				rules, Variable.getVariable("b"));
		assertEquals(11, problem.getNumberOfConstraint());
		assert (rules.isEmpty());
		for (Rule<LPVariable> rule : result) {
			assertEquals(null, rule.getContext().get(1));
		}
	}

	/**
	 * 
	 * Test the projection of basis function on the plumbing example
	 * 
	 */
	public void testRuleBackProj() {
		RuleBasedBackproj ruleBasedFactoredLP = new RuleBasedBackproj(
				plumbingForest);
		PlainRules<Double> hi = getRules("<rdr:rules>" + "  <rdr:rule data='100'>"
				+ "    <rdr:cond variable='Painting' value='Done'/>" + "  </rdr:rule>"
				+ "</rdr:rules>", plumbingForest.getVariableOrdering());
		PlainRules<Double> expectedRules = getRules("<rdr:rules>"
				+ "  <rdr:rule data='95'>"
				+ "    <rdr:cond variable='Plumbing' value='Done'/>"
				+ "    <rdr:cond variable='Electrical' value='Done'/>"
				+ "  </rdr:rule>" + "</rdr:rules>", plumbingForest
				.getVariableOrdering());
		PlainRules<Double> backProjection = ruleBasedFactoredLP.ruleBackProj(
				"paint", hi.iterator().next());
		assertEquals(expectedRules, backProjection);

		expectedRules = getRules("<rdr:rules>" + "  <rdr:rule data='90'>"
				+ "    <rdr:cond variable='Plumbing' value='Done'/>"
				+ "    <rdr:cond variable='Electrical' value='Done'/>"
				+ "    <rdr:cond variable='Painting' value='Done'/>" + "  </rdr:rule>"
				+ "</rdr:rules>", plumbingForest.getVariableOrdering());
		backProjection = ruleBasedFactoredLP.ruleBackProj("notpaint", hi.iterator()
				.next());
		assertEquals(expectedRules, backProjection);
	}

	/**
	 * 
	 * Test the backproj operator on the coffeerobot problem
	 * 
	 */
	public void testRulesBackProjOnCoffeeRobot() {
		StaticSpiSvi spisvi = new StaticSpiSvi(XmlUtils
				.getRootFromFile("data/forests/coffeerobot.guestrin.agent"));
		GuestrinPlanning guestrinBackprojPlanning = new GuestrinPlanning(spisvi
				.getModel());
		guestrinBackprojPlanning
				.setPolicy(new RuleBasedBackproj(spisvi.getModel()));
		for (String action : spisvi.getModel().getActionList()) {
			checkBackProj(spisvi, guestrinBackprojPlanning, spisvi.getModel()
					.getRewardTree(action), action);
			checkBackProj(spisvi, guestrinBackprojPlanning, spisvi
					.getOptimalValueFunction(), action);
		}
		GuestrinPlanning boutilierOnTreeBackProjPlanning = new GuestrinPlanning(
				spisvi.getModel());
		boutilierOnTreeBackProjPlanning.setPolicy(new TreeBasedBackproj(spisvi
				.getModel(), Parameters.GAMMA));
		for (String action : spisvi.getModel().getActionList()) {
			checkBackProj(spisvi, boutilierOnTreeBackProjPlanning, spisvi.getModel()
					.getRewardTree(action), action);
			checkBackProj(spisvi, boutilierOnTreeBackProjPlanning, spisvi
					.getOptimalValueFunction(), action);
		}
	}

	/**
	 * 
	 * Check the backprojection computed by SPI against the backprojection using
	 * Guestrin algorithm
	 * 
	 * @param spisvi
	 *          spisvi algo to check with
	 * @param ruleBasedFactoredLP
	 *          the planning algorithm to test
	 * @param valueTree
	 *          tree to use in the backprojection
	 * @param action
	 *          action for which we want to compute the back projection
	 * 
	 */
	private void checkBackProj(SpiSvi spisvi,
			GuestrinPlanning ruleBasedFactoredLP, Tree<Double> valueTree,
			String action) {
		StaticForest model = (StaticForest) spisvi.getModel();

		Tree<Map<Variable, IProbabilities<String>>> pregress = spisvi.getPRegress(
				valueTree.getRoot(), action, null);
		Tree<Double> expectedResult = spisvi.constructFVTree(action, pregress,
				valueTree);

		PlainRules<Double> computedResult = ruleBasedFactoredLP.rulesBackProj(
				action, valueTree);

		for (Rule<Double> rule : computedResult) {
			Context state = new Context();
			Context context = rule.getContext();
			for (Variable variable : model.getVariableSet()) {
				String value = context.get(variable);
				if (value != null) {
					state.put(variable, value);
				} else {
					state.put(variable, "");
				}
			}

			double expectedValue = expectedResult.classify(state);
			double ruleValue = rule.get();
			if (Math.abs(expectedValue - ruleValue) > Parameters.EPSILON) {
				System.out.println(state.toString() + ": "
						+ String.valueOf(expectedValue) + " != "
						+ String.valueOf(ruleValue));
				assertEquals(expectedValue, ruleValue);
			}
		}
	}

	/**
	 * 
	 * Compute the value function of the simple problem using different variables
	 * in the value function and the reward function
	 * 
	 */
	public void testRulesPlanningLimitedRewardFunctionScope() {
		StaticForest forest = new StaticForest(XmlUtils
				.getRootFromFile("data/forests/simple.guestrin.forest"));
		GuestrinPlanning treeFactoredLP = new GuestrinPlanning(forest);
		assertTrue(testGuestrinPlanning(forest, treeFactoredLP) < 0.1);
	}

	/**
	 * 
	 * Compute the value function of the CoffeeRobot problem (compute the
	 * backprojection with Spi algorithm)
	 * 
	 */
	public void testGuestrinPlanningWithSpiBackProjOnCoffeeRobot() {
		StaticSpiSvi spisvi = new StaticSpiSvi(XmlUtils
				.getRootFromFile("data/forests/coffeerobot.spisvi.agent"));
		StaticForest staticForest = spisvi.getModel();
		staticForest.setBasisFunctions(staticForest
				.decomposeBasisFunction(staticForest.getBasisFunctions()));

		double error;
		GuestrinPlanning treeBasedPlanning = new GuestrinPlanning(spisvi.getModel());
		error = testGuestrinPlanning(staticForest, treeBasedPlanning);
		assertTrue(error < 10.0);
	}

	/**
	 * 
	 * Compute the value function of the CoffeeRobot problem (compute the
	 * backprojection with Spi algorithm)
	 * 
	 */
	public void testPolicyConstruction() {
		GuestrinPlanning treeBasedPlanning = new GuestrinPlanning(new StaticForest(
				XmlUtils.getRootFromFile("data/forests/coffeerobot.forest")));
		treeBasedPlanning.solve();
		DecisionListPolicy optimalPolicy = treeBasedPlanning
				.getOptimalDecisionListPolicy();
		optimalPolicy.removeUselessRules();
		TestSpiSvi.testOptimalPolicy(optimalPolicy);
	}

	/**
	 * 
	 * Apply Guestring planning method and compare it to Boutilier
	 * 
	 * @param forest
	 *          static forest used as reference
	 * @param ruleBasedFactoredLP
	 *          guestring planning to check
	 * 
	 * @return the error computed
	 * 
	 */
	private double testGuestrinPlanning(StaticForest forest,
			GuestrinPlanning ruleBasedFactoredLP) {
		ruleBasedFactoredLP.solve();
		Map<String, IFunction<Double>> actionValues = ruleBasedFactoredLP
				.getQualityFunctionWrapper();
		Greedy greedy = new Greedy(null, forest.getActionList());
		Context[] stateSpace = Context.createStateSet(forest);
		ID3Tree.Plain<String> policy = new ID3Tree.Plain<String>(new KSMeasure(),
				forest.getVariableOrdering());
		for (Context state : stateSpace) {
			policy.learn(new Example<String>(greedy.getAction(actionValues, state),
					forest.getVariableOrdering().stateToList(state)));
		}
		policy.flush();
		IFunction<Double> valueFunction = ruleBasedFactoredLP
				.getApproximatedValueFunction();
		return computeError(forest, valueFunction);
	}

	/**
	 * 
	 * Compute an error between an approximated value function and the exact value
	 * function
	 * 
	 * @param forest
	 *          the forest defining the problem
	 * @param rulesValueFunction
	 *          the approximated value function to estimate
	 * @return the error computed
	 * 
	 */
	private double computeError(StaticForest forest,
			IFunction<Double> rulesValueFunction) {
		StaticSpiSvi spisvi = new StaticSpiSvi(forest);
		spisvi.getSPI();
		IFunction<Double> forestValueFunction = spisvi.getOptimalValueFunction();
		Context[] stateSpace = Context.createStateSet(forest);
		double error = 0;
		for (Context state : stateSpace) {
			error = Math.max(error, Math.abs(forestValueFunction.classify(state)
					- rulesValueFunction.classify(state)));
		}
		return error;
	}

	/**
	 * 
	 * Test the back projection operator on set of rules
	 * 
	 */
	public void testLPResolutionOnDisconnected() {
		SysAdminProblem problem = new Disconnected();
		problem.setProblemData(1);
		StaticForest forest = problem.toForest();
		GuestrinPlanning ruleBasedFactoredLP = new GuestrinPlanning(problem
				.toForest());
		List<LPNamedVariable> omegas = ruleBasedFactoredLP.factoredLPA();
		assert (omegas.size() == forest.getBasisFunctions().size());
		assertEquals(198.0, omegas.get(0).getSolution());
	}

}
