package animatlab.kodamat.reinforcementlearning.planning.spisvi;

import java.util.ArrayList;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Random;
import java.util.Set;

import animatlab.kodamat.core.functions.IPartition;
import animatlab.kodamat.core.trees.Tree;
import animatlab.kodamat.core.utils.Context;
import animatlab.kodamat.core.utils.Variable;
import animatlab.kodamat.core.utils.VariableOrdering;
import animatlab.kodamat.supervisedlearning.core.Example;
import animatlab.kodamat.supervisedlearning.entropy.VariableMeasure;
import animatlab.kodamat.supervisedlearning.learningtree.ID4Tree;

/**
 * 
 * Reorganize a tree
 * 
 * @author Thomas Degris
 * @param <T>
 *          the type contained at the leaf of the tree
 * 
 */
public class ReorganizeTree<T> {
	/**
	 * 
	 * Constructor
	 * 
	 */
	public ReorganizeTree() {
		super();
	}

	/**
	 * 
	 * Reorganize a policy tree
	 * 
	 * @param tree
	 *          the tree to reorganize
	 * @param valueSets
	 *          sets of value
	 * @param measure
	 *          the measure to use to reorganize the tree
	 * @return the result of the reorganization
	 * 
	 */
	public Tree<T> reorganize(Tree<T> tree, Map<Variable, Set<String>> valueSets,
			VariableMeasure<T> measure) {
		VariableOrdering variableOrdering = VariableOrdering
				.buildOrdering(valueSets.keySet());
		ID4Tree.Plain<T> learningTree = new ID4Tree.Plain<T>(measure,
				variableOrdering);
		Set<Example<T>> examples = getExampleSet(variableOrdering, tree, valueSets);
		for (Example<T> example : examples) {
			learningTree.learn(example);
		}
		return learningTree;
	}

	/**
	 * 
	 * Build a example database from a tree
	 * 
	 * @param variableOrdering
	 *          variable ordering to build the examples
	 * @param tree
	 *          the tree to build the examples from
	 * @param valueSets
	 *          sets of value
	 * 
	 * @return a set of examples
	 * 
	 */
	private Set<Example<T>> getExampleSet(VariableOrdering variableOrdering,
			Tree<T> tree, Map<Variable, Set<String>> valueSets) {
		Set<Example<T>> examplesSet = new LinkedHashSet<Example<T>>();
		Random random = new Random(0);
		Map<Variable, List<String>> valueLists = new LinkedHashMap<Variable, List<String>>();
		for (Map.Entry<Variable, Set<String>> entry : valueSets.entrySet()) {
			valueLists.put(entry.getKey(), new ArrayList<String>(entry.getValue()));
		}
		for (IPartition<T> leaf : tree.partitions()) {
			Context context = leaf.getContext();
			int numberOfExamples = (valueSets.size() - context.size()) * 4;
			T value = leaf.get();
			for (int count = 0; count < numberOfExamples; count++) {
				List<String> state = new ArrayList<String>();
				for (int variable = 0; variable < valueSets.size(); variable++) {
					Variable var = variableOrdering.get(variable);
					String varValue = context.get(var);
					if (varValue == null) {
						List<String> valueSet = valueLists.get(var);
						varValue = valueSet.get(random.nextInt(valueSet.size()));
					}
					state.add(varValue);
				}
				examplesSet.add(new Example<T>(value, state));
			}
		}
		return examplesSet;
	}
}
