/*
 * 
 * Created on Oct 10, 2005
 *
 */
package animatlab.kodamat.reinforcementlearning.planning.spisvi.tests;

import java.util.Iterator;
import java.util.Map;
import java.util.Random;
import java.util.Set;

import junit.framework.TestCase;
import animatlab.kodamat.core.functions.IFunction;
import animatlab.kodamat.core.io.XmlUtils;
import animatlab.kodamat.core.trees.AppendPolicy;
import animatlab.kodamat.core.trees.EqualPolicy;
import animatlab.kodamat.core.trees.Leaf;
import animatlab.kodamat.core.trees.Tree;
import animatlab.kodamat.core.utils.Parameters;
import animatlab.kodamat.core.utils.Context;
import animatlab.kodamat.core.utils.Variable;
import animatlab.kodamat.reinforcementlearning.model.forest.StaticForest;
import animatlab.kodamat.reinforcementlearning.planning.spisvi.SpiSvi;
import animatlab.kodamat.reinforcementlearning.planning.spisvi.StaticSpiSvi;
import animatlab.kodamat.reinforcementlearning.planning.spisvi.TreeBasedPlanning;
import animatlab.kodamat.supervisedlearning.regression.LeastSquareMeasure;
import animatlab.kodamat.supervisedlearning.utils.PartitionsToTree;

/**
 * 
 * Test the SPI/SVI convergence on the Coffee Robot problem
 * 
 * @author Thomas Degris
 * 
 */
public class TestSpiSvi extends TestCase {
	/**
	 * 
	 * Forest used testing purpose taken from Boutilier : Stochastic Dynamic
	 * Programming with Factored Representations
	 * 
	 */
	static public final StaticForest xyzForest = new StaticForest(XmlUtils
			.getRoot("<fst:forest>" + "  <dtr:legend>"
					+ "    <spi:variable legend='x' index='0'/>"
					+ "    <spi:variable legend='y' index='1'/>"
					+ "    <spi:variable legend='z' index='2'/>" + "  </dtr:legend>"
					+ "  <fst:transitions>" + "   <fst:action name='simpleaction'>"
					+ "     <fst:variable variable='x'>" + "       <dtr:leaf>"
					+ "         <spi:probability probability='1.0' value='true'/>"
					+ "       </dtr:leaf>" + "     </fst:variable>"
					+ "     <fst:variable variable='y'>"
					+ "       <dtr:dnode variable='y'>"
					+ "         <dtr:leaf value='true'>"
					+ "           <spi:probability probability='1.0' value='true'/>"
					+ "         </dtr:leaf>"
					+ "         <dtr:dnode value='false' variable='x'>"
					+ "           <dtr:leaf value='true'>"
					+ "             <spi:probability probability='0.9' value='true'/>"
					+ "             <spi:probability probability='0.1' value='false'/>"
					+ "           </dtr:leaf>" + "           <dtr:leaf value='false'>"
					+ "             <spi:probability probability='1.0' value='false'/>"
					+ "           </dtr:leaf>" + "         </dtr:dnode>"
					+ "       </dtr:dnode>" + "     </fst:variable>"
					+ "     <fst:variable variable='z'>"
					+ "       <dtr:dnode variable='z'>"
					+ "         <dtr:leaf value='true'>"
					+ "           <spi:probability probability='1.0' value='true'/>"
					+ "         </dtr:leaf>"
					+ "         <dtr:dnode value='false' variable='y'>"
					+ "           <dtr:leaf value='true'>"
					+ "             <spi:probability probability='0.9' value='true'/>"
					+ "             <spi:probability probability='0.1' value='false'/>"
					+ "           </dtr:leaf>" + "           <dtr:leaf value='false'>"
					+ "             <spi:probability probability='1.0' value='false'/>"
					+ "           </dtr:leaf>" + "         </dtr:dnode>"
					+ "       </dtr:dnode>" + "     </fst:variable>" + "   </fst:action>"
					+ "  </fst:transitions>" + "  <fst:reward>"
					+ "    <dtr:dnode variable='z' weight='1'>"
					+ "      <dtr:leaf value='true' data='10'/>"
					+ "      <dtr:leaf value='false' data='0'/>" + "    </dtr:dnode>"
					+ "  </fst:reward>" + "</fst:forest>"));

	/**
	 * 
	 * Precision to compare double value
	 * 
	 */
	protected int precision = 1;

	/**
	 * 
	 * Test if the reward max is correctly computed from the tree
	 * 
	 */
	public void testRewardMax() {
		StaticSpiSvi spisvi = new StaticSpiSvi(XmlUtils
				.getRootFromFile("data/forests/coffeerobot.forest"));
		assertEquals(1.0, spisvi.getModel().getRewardMax());
	}

	/**
	 * 
	 * Test the append method on SpiSviTree
	 * 
	 * 
	 */
	public void testAppendMethod() {
		StaticSpiSvi spisvi = new StaticSpiSvi(XmlUtils
				.getRootFromFile("data/forests/coffeerobot.forest"));
		Tree<Double> rewardTree = spisvi.getModel().getRewardTree(null);
		Tree<Double> workingTree = new Tree<Double>();
		((Leaf<Double>) workingTree.getRoot()).set(10.0);

		workingTree.append(rewardTree, AppendPolicy.sumPolicy, null);
		workingTree.append(spisvi.getModel().getRewardTree(null),
				AppendPolicy.sumPolicy, null);

		String xmlString = "<dtr:dnode variable='huc'>"
				+ "  <dtr:dnode value='1' variable='w'>"
				+ "    <dtr:leaf data='11.8' value='1'/>"
				+ "    <dtr:leaf data='12.0' value='0'/>" + "  </dtr:dnode>"
				+ "  <dtr:dnode value='0' variable='w'>"
				+ "    <dtr:leaf data='10.0' value='1'/>"
				+ "    <dtr:leaf data='10.2' value='0'/>" + "  </dtr:dnode>"
				+ "</dtr:dnode>";
		Tree<Double> expectedRewardTree = spisvi.loadValueFunction(XmlUtils
				.getRoot(xmlString));
		assertEquals(true, expectedRewardTree.equals(workingTree,
				EqualPolicy.valueLeafEqualPolicy));
	}

	/**
	 * 
	 * Test that Spi converges
	 * 
	 */
	public void testFiftyIteration() {
		StaticSpiSvi spisvi = new StaticSpiSvi(XmlUtils
				.getRootFromFile("data/forests/coffeerobot.forest"));
		spisvi.gamma = 0.9;
		Tree<Set<String>> policyTree = spisvi.getNewSpiSviPolicy("delc");
		Tree<Double> valueFunction = spisvi.getModel().getRewardTree(null);
		for (int index = 1; index <= 50; index++) {
			valueFunction = spisvi.getRegress(valueFunction, policyTree);
		}
		assertEquals(8.9, getValueFromState(valueFunction, "huc=1,w=1"));
		assertEquals(9.9, getValueFromState(valueFunction, "huc=1,w=0"));
		assertEquals(6.5, getValueFromState(valueFunction, "huc=0,hrc=1,w=1,l=1"));
		assertEquals(7.5, getValueFromState(valueFunction, "huc=0,hrc=1,w=0,l=1"));
		assertEquals(0.0, getValueFromState(valueFunction, "huc=0,hrc=1,w=1,l=0"));
		assertEquals(0.9, getValueFromState(valueFunction, "huc=0,hrc=1,w=0,l=0"));
		assertEquals(0.0, getValueFromState(valueFunction, "huc=0,hrc=0,w=1"));
		assertEquals(0.9, getValueFromState(valueFunction, "huc=0,hrc=0,w=0"));
	}

	/**
	 * 
	 * Test the optimal policy
	 * 
	 * @param policy
	 *          the optimal policy tree to test
	 * 
	 */
	static public void testOptimalPolicy(IFunction<Set<String>> policy) {
		assertTrue(policy.classify(
				Context.getState("huc=0,hrc=1,w=0,r=1,u=1,l=1", null)).contains("delc"));
		assertTrue(policy.classify(
				Context.getState("huc=0,hrc=1,w=0,r=1,u=1,l=0", null)).contains("move"));
		assertTrue(policy.classify(
				Context.getState("huc=0,hrc=0,w=0,r=1,u=1,l=1", null)).contains("move"));
		assertTrue(policy.classify(
				Context.getState("huc=0,hrc=0,w=0,r=1,u=0,l=1", null)).contains("getu"));
	}

	/**
	 * 
	 * Test the SVI algorithm
	 * 
	 */
	public void testSVI() {
		StaticSpiSvi spisvi = new StaticSpiSvi(XmlUtils
				.getRootFromFile("data/forests/coffeerobot.forest"));
		testOptimalPolicy(spisvi.getSVI());
	}

	/**
	 * 
	 * Test that TBP converges
	 * 
	 */
	public void testTBP() {
		StaticForest staticForest = new StaticForest(XmlUtils
				.getRootFromFile("data/forests/coffeerobot.forest"));
		testTBPvsSpiSviRegress(staticForest);
		TreeBasedPlanning tbp = new TreeBasedPlanning(staticForest);
		testOptimalPolicy(tbp.getSVI());
	}

	/**
	 * 
	 * Compare the backprojection computed by SpiSvi with TBP
	 * 
	 * @param staticForest
	 *          the forest to use in the back projection
	 * 
	 */
	private void testTBPvsSpiSviRegress(StaticForest staticForest) {
		final int numberOfIteration = 10;
		SpiSvi spisvi = new SpiSvi(staticForest);
		TreeBasedPlanning tbp = new TreeBasedPlanning(staticForest);

		Tree<Double> spisviValFun = staticForest.getRewardTree(null);
		Tree<Double> tbpValFun = staticForest.getRewardTree(null);

		for (int index = 0; index < numberOfIteration; index++) {
			spisviValFun = regress(spisvi, spisviValFun);
			tbpValFun = regress(tbp, tbpValFun);
			Context[] stateSet = Context.createStateSet(staticForest);
			for (Context state : stateSet) {
				assertTrue(Math.abs(spisviValFun.classify(state)
						- tbpValFun.classify(state)) < Parameters.EPSILON);
			}
		}
	}

	/**
	 * 
	 * Regress one step on a tree
	 * 
	 * @param spisvi
	 *          planning to use
	 * @param treeVk
	 *          tree to regress
	 * 
	 * @return a new tree regressed
	 * 
	 */
	@SuppressWarnings("null")
	private Tree<Double> regress(SpiSvi spisvi, Tree<Double> treeVk) {
		Tree<Double> treeVkp1 = null;
		for (String action : spisvi.getModel().getActionList()) {
			// (a) Compute Tree(QaVk)=Regress(Tree(Vk),a) for each action a
			Tree<Double> treeQaVk = spisvi.treeBackproj(treeVk, action);
			// (b) Merge (via maximization) the trees Tree(QaVk) to obtain Tree(Vk+1)
			treeVkp1 = spisvi.maxPolicy.merge(treeQaVk, treeVkp1);
		}
		treeVkp1.simplify(EqualPolicy.valueLeafEqualPolicy);
		return treeVkp1;
	}

	/**
	 * 
	 * Test the SPI algorithm
	 * 
	 */
	public void testSPI() {
		StaticSpiSvi spisvi = new StaticSpiSvi(XmlUtils
				.getRootFromFile("data/forests/coffeerobot.forest"));
		testOptimalPolicy(spisvi.getSPI());
	}

	/**
	 * 
	 * Test SPI/SVI on a forest generated from cstrike
	 * 
	 */
	public void testSpiSviOnCStrike01() {
		StaticSpiSvi spisvi = new StaticSpiSvi(XmlUtils
				.getRootFromFile("data/forests/bot01.agent"));
		spisvi
				.setUnknownProbabilityPolicy(spisvi.new OptimisticUnknownProbabilityPolicy());
		Tree<Double> valueFunction = spisvi.getOptimalValueFunction();
		Tree<Double> regress = spisvi.getRegress(valueFunction, " the bomb site B");
		valueFunction = spisvi.maxPolicy.merge(regress, valueFunction);
	}

	/**
	 * 
	 * Test SPI/SVI on a forest generated from cstrike
	 * 
	 */
	public void testSpiSviOnCStrike02() {
		StaticSpiSvi spisvi = new StaticSpiSvi(XmlUtils
				.getRootFromFile("data/forests/bot02.agent"));
		spisvi.getSVI();
		Tree<Double> valueFunction = spisvi.getOptimalValueFunction();
		spisvi.getRegress(valueFunction, " the bomb site A");
	}

	/**
	 * 
	 * Get the value associated to a state from a tree
	 * 
	 * @param valueTree
	 *          the tree to look for the value
	 * @param state
	 *          the state to look for in the tree
	 * @return the value associated the the state
	 * 
	 */
	private double getValueFromState(IFunction<Double> valueTree, Context state) {
		return round(valueTree.classify(state), precision);
	}

	/**
	 * 
	 * Round the value to be comparable with static constant
	 * 
	 * @param value
	 *          the value to round
	 * @param precision
	 *          the precision wanted
	 * 
	 * @return the value rounded
	 * 
	 */
	public static double round(double value, int precision) {
		return Math.floor(value * Math.pow(10, precision))
				/ Math.pow(10, precision);
	}

	/**
	 * 
	 * Get the value associated to a state from a tree
	 * 
	 * @param valueTree
	 *          the tree to look for the value
	 * @param state
	 *          the state to look for in the tree
	 * @return the value associated the the state
	 * 
	 */
	private double getValueFromState(IFunction<Double> valueTree, String state) {
		return getValueFromState(valueTree, Context.getState(state, null));
	}

	/**
	 * 
	 * Test the value function reorganization
	 * 
	 * 
	 */
	public void testValueFunctionReorganization() {
		StaticSpiSvi spisvi = new StaticSpiSvi(XmlUtils
				.getRootFromFile("data/forests/factory.forest"));
		spisvi.setOrdering(null);
		spisvi.MAX_ITE = 2;
		spisvi.getSVI();
		PartitionsToTree<Double> partitionsToTree = new PartitionsToTree<Double>(
				new LeastSquareMeasure());
		Tree<Double> optimalValueFunction = spisvi.getOptimalValueFunction();
		Tree<Double> tree = partitionsToTree.build(optimalValueFunction);
		compareTree(optimalValueFunction, tree, spisvi.getModel().getValueSets());
	}

	/**
	 * 
	 * Compare two trees with random examples
	 * 
	 * @param expected
	 *          the tree used as reference
	 * @param tested
	 *          the tree to test
	 * @param valueSets
	 *          the set of value for each variable
	 * 
	 */
	private void compareTree(Tree<Double> expected, Tree<Double> tested,
			Map<Variable, Set<String>> valueSets) {
		Random random = new Random(0);
		assert (tested.numberOfNodes() < expected.numberOfNodes());
		for (int numberOfTest = 0; numberOfTest < 10000; numberOfTest++) {
			Context state = new Context();
			state.clear();
			for (Variable variable : valueSets.keySet()) {
				Set<String> valueSet = valueSets.get(variable);
				int valueIndex = random.nextInt(valueSet.size());
				Iterator<String> valueIterator = valueSet.iterator();
				String value;
				do {
					value = valueIterator.next();
					valueIndex--;
				} while (valueIndex >= 0);
				state.put(variable, value);
			}
			// assertEquals(expected.classify(state), tested.classify(state));
			assertTrue(Math.abs(expected.classify(state) - tested.classify(state)) < Parameters.EPSILON);
		}
	}
}
