/*
 * 
 * Created on Jan 12, 2006
 *
 */
package animatlab.kodamat.reinforcementlearning.planning.spudd;

import static animatlab.kodamat.core.io.XmlUtils.getFirstChild;
import static animatlab.kodamat.core.io.XmlUtils.getRootFromFile;
import static animatlab.kodamat.core.io.XmlUtils.save;
import static animatlab.kodamat.reinforcementlearning.model.forest.StaticForest.PROBLEM_EXTENSION;

import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import animatlab.kodamat.reinforcementlearning.model.forest.AbstractForest;
import animatlab.kodamat.reinforcementlearning.model.forest.StaticForest;
import animatlab.kodamat.reinforcementlearning.planning.IPlanning;

/**
 * 
 * SpiSvi instance loaded from a xml file to get a pre-computed optimal value
 * function and optimal policy
 * 
 * @author Thomas Degris
 * 
 */
public class StaticSpudd extends Spudd {
	/**
	 * 
	 * Name of this planning method
	 * 
	 */
	public static final String PLANNING_METHOD_NAME = "spudd";

	/**
	 * 
	 * File to load to run the test
	 * 
	 */
	static public String AGENT_EXTENSION = ".spudd.agent";

	/**
	 * 
	 * Constructor
	 * 
	 * @param element
	 *          xml tree to load the agent from
	 * 
	 */
	public StaticSpudd(Element element) {
		super(null);
		assert (element.getNodeName().equals(NODE_NAME_AGENT) || element
				.getNodeName().equals(AbstractForest.NODE_NAME_FOREST));
		if (element.getNodeName().equals(AbstractForest.NODE_NAME_FOREST)) {
			setForest(new StaticForest(element));
		}

		NodeList nodeList = element.getChildNodes();
		for (int index = 0; index < nodeList.getLength(); index++) {
			Node childNode = nodeList.item(index);
			if (childNode.getNodeType() == Node.ELEMENT_NODE) {
				if (childNode.getNodeName().equals(AbstractForest.NODE_NAME_FOREST)) {
					setForest(new StaticForest((Element) childNode));
				}
				if (childNode.getNodeName().equals(NODE_NAME_VALUE_FUNCTION)) {
					lastComputedValueFunction = loadValueFunction(getFirstChild(childNode));
				}
				if (childNode.getNodeName().equals(NODE_NAME_POLICY)) {
					lastComputedPolicy = loadPolicy(getFirstChild(childNode));
				}
			}
		}
	}

	/**
	 * 
	 * Constructor
	 * 
	 * @param forest
	 *          the model used by the algorithm
	 * 
	 */
	public StaticSpudd(AbstractForest forest) {
		super(forest);
	}

	/**
	 * 
	 * Convert a forest xml file into an agent xml file
	 * 
	 * @param forest
	 *          the forest to convert
	 * @param outputFile
	 *          file where to save the forest
	 * @return the planning to compute the policy of the agent
	 * 
	 */
	static public IPlanning convertProblemToAgent(AbstractForest forest,
			String outputFile) {
		Spudd spudd = new Spudd(forest);
		spudd.solve();
		save(spudd, null, outputFile + AGENT_EXTENSION);
		return spudd;
	}

	/**
	 * 
	 * Main method
	 * 
	 * @param args
	 *          list of arguments passed to the program
	 * 
	 */
	public static void main(String[] args) {
		StaticSpudd.convertProblemToAgent(new StaticForest(getRootFromFile(args[0]
				+ PROBLEM_EXTENSION)), args[0]);
	}
}
