package animatlab.kodamat.reinforcementlearning.planning.spudd.test;

import junit.framework.TestCase;
import animatlab.kodamat.core.io.XmlUtils;
import animatlab.kodamat.core.proba.IProbabilities;
import animatlab.kodamat.core.trees.Leaf;
import animatlab.kodamat.core.trees.Tree;
import animatlab.kodamat.core.utils.Variable;
import animatlab.kodamat.reinforcementlearning.model.StaticCPD;
import animatlab.kodamat.reinforcementlearning.model.forest.StaticForest;
import animatlab.kodamat.reinforcementlearning.planning.spisvi.tests.TestSpiSvi;
import animatlab.kodamat.reinforcementlearning.planning.spudd.Spudd;

/**
 * 
 * Test Spudd planning method
 * 
 * @author Thomas Degris
 * 
 */
public class TestSpudd extends TestCase {
	/**
	 * 
	 * Test the Spudd planning method on the Coffee Robot problem
	 * 
	 */
	public void testSpuddOnCoffeeRobot() {
		StaticForest forest = new StaticForest(XmlUtils
				.getRootFromFile("data/forests/coffeerobot.forest"));
		Spudd spudd = new Spudd(forest);
		spudd.solve();
		TestSpiSvi.testOptimalPolicy(spudd.getOptimalPolicy());
		// animatlab.kodamat.core.io.DotUtils.show(
		// spudd.getOptimalValueFunction(), new IToDot.IToDotPolicy<Double>() {
		// public String toDotLabel(Double data) {
		// NumberFormat numberFormat = NumberFormat.getInstance();
		// numberFormat.setMinimumFractionDigits(1);
		// numberFormat.setMaximumFractionDigits(1);
		// return numberFormat.format(Math.round(data * 50.0) / 100.0);
		// }
		// });
	}

	/**
	 * 
	 * Test the Spudd planning method on the Coffee Robot problem
	 * 
	 */
	public void testSpuddOnCoffeeRobotWithTransitionLeaves() {
		StaticForest forest = new StaticForest(XmlUtils
				.getRootFromFile("data/forests/coffeerobot.forest"));
		Variable variable = Variable.getVariable("huc");
		Tree<IProbabilities<String>> tree = forest.getTransitionFunction(variable,
				"move");
		tree.getRoot().remove();
		Leaf<IProbabilities<String>> leaf = (Leaf<IProbabilities<String>>) tree
				.getRoot();
		StaticCPD proba = new StaticCPD();
		proba.setProba("1", 1);
		leaf.set(proba);
		assert (forest.getTransitionFunction(variable, "move").numberOfNodes() == 1);
		Spudd spudd = new Spudd(forest);
		spudd.solve();
	}

}
