package animatlab.kodamat.reinforcementlearning.planning.stats;

import java.util.Map;

import animatlab.kodamat.core.utils.Chrono;
import animatlab.kodamat.reinforcementlearning.agent.IAgentFactory;
import animatlab.kodamat.reinforcementlearning.agent.ModelBasedAgent;
import animatlab.kodamat.reinforcementlearning.core.stats.Measure;
import animatlab.kodamat.reinforcementlearning.core.stats.MeasureWriter;
import animatlab.kodamat.reinforcementlearning.model.forest.AbstractLearningForest;

/**
 * 
 * Measure used to compute the quality of policy from a trajectory
 * 
 * @author Thomas Degris
 * 
 */
public class BatchLearningMeasure extends BatchMeasure {
	/**
	 * 
	 * Constructor
	 * 
	 * 
	 */
	public BatchLearningMeasure() {
		super();
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.core.stats.Measure#computeMeasures(java.util.Map)
	 */
	@Override
	public void computeMeasures(Map<String, Double> data) {
		for (IAgentFactory agentFactory : agentsFactory) {
			ModelBasedAgent thisAgent = (ModelBasedAgent) agentFactory.createAgent(
					null, problem, solution);
			MeasureWriter measureWriter = agentWriters.get(agentFactory);
			Chrono chrono = new Chrono();
			((AbstractLearningForest) thisAgent.getModel()).learnFromTrajectory(
					trajectory, null);
			measureWriter.addToNextStep("tl", chrono.getCurrentChrono());
			for (Measure measure : measures) {
				measure.setMeasureData(problem, solution, thisAgent, resultsFolder);
			}
			measureWriter.computeMeasures(data);
		}
	}
}
