/*
 * 
 * Created on 15 nov. 2005
 *
 */
package animatlab.kodamat.reinforcementlearning.planning.stats;

import java.util.Map;

import animatlab.kodamat.core.functions.IFunction;
import animatlab.kodamat.reinforcementlearning.agent.ModelBasedAgent;
import animatlab.kodamat.reinforcementlearning.core.stats.Measure;
import animatlab.kodamat.reinforcementlearning.planning.IPlanning;

/**
 * 
 * Measure logging the size of the model learn by the agent
 * 
 * @author Thomas Degris
 * 
 */
public class FunctionSizeMeasure extends Measure {
	/**
	 * 
	 * Constructor
	 * 
	 */
	public FunctionSizeMeasure() {
		super();
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.core.stats.Measure#computeMeasures(Map)
	 */
	@Override
	public void computeMeasures(Map<String, Double> data) {
		if (!(agent instanceof ModelBasedAgent)) {
			return;
		}
		ModelBasedAgent modelBasedAgent = (ModelBasedAgent) agent;
		data.put("modelsize", (double) modelBasedAgent.getModel()
				.getTransitionsModelSize());
		data.put("rewardmodelsize", (double) modelBasedAgent.getModel()
				.getRewardModelSize());
		IPlanning planning = modelBasedAgent.getPlanning();
		if (planning != null) {
			IFunction<Double> optimalValueFunction = planning
					.getOptimalValueFunction();
			double numberOfNode = 0;
			if (optimalValueFunction != null) {
				numberOfNode = optimalValueFunction.numberOfNodes();
			}
			data.put("valfunsize", numberOfNode);
		}
	}

	/**
	 * @see animatlab.kodamat.reinforcementlearning.core.stats.Measure#reset()
	 */
	@Override
	public void reset() {
		// Nothing to reset
	}
}
