package animatlab.kodamat.supervisedlearning.core;

import java.util.List;

/**
 * Example
 * 
 * Example of instance to learn in the tree. An example is composed of a list of
 * value for each variable and the class of the example
 * 
 * @param <T>
 *          type of object classified
 * 
 * @author Thomas Degris
 * 
 */
public class Example<T> {
	/**
	 * 
	 * List of values for each variables
	 * 
	 */
	final public List<String> values;

	/**
	 * 
	 * Class of this example to classify
	 * 
	 */
	final public T className;

	/**
	 * 
	 * We compute the hash code only one time by example, so the value is cashed
	 * 
	 */
	protected int hashCodeComputed;

	/**
	 * 
	 * True if the hash code has already been computed
	 * 
	 */
	protected boolean hashFlag = false;

	/**
	 * 
	 * Constructor
	 * 
	 * @param className
	 *          class of this example
	 * @param values
	 *          values of the variable for this example
	 * 
	 */
	public Example(T className, List<String> values) {
		assert (className != null);
		this.className = className;
		this.values = values;
	}

	/**
	 * @see java.lang.Object#equals(java.lang.Object)
	 */
	@SuppressWarnings("unchecked")
	@Override
	public boolean equals(Object obj) {
		Example<T> example = (Example<T>) obj;
		for (int index = 0; index < values.size(); index++) {
			if (((values.get(index) != null) && (example.values.get(index) != null) && !values
					.get(index).equals(example.values.get(index)))) {
				return false;
			}
		}
		return className.equals(example.className);
	}

	/**
	 * @see java.lang.Object#hashCode()
	 */
	@Override
	public int hashCode() {
		if (!hashFlag) {
			StringBuilder buffer = new StringBuilder();
			for (int index = 0; index < values.size(); index++) {
				buffer.append(values.get(index));
			}
			hashCodeComputed = buffer.toString().hashCode();
			hashFlag = true;
		}
		return hashCodeComputed;
	}

	/**
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		return String.valueOf(values) + ">" + String.valueOf(className);
	}
}
