/*
 * 
 * Created on 31 mars 2005
 *
 */
package animatlab.kodamat.supervisedlearning.core;

import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.Map;

/**
 * 
 * Database to store the example that occured (and the number of time they
 * occured)
 * 
 * @author Thomas Degris
 * @param <T>
 *          type of the example to count
 * 
 */
public class ExampleDatabase<T> extends LinkedHashMap<Example<T>, Integer> {
	/**
	 * 
	 * SerialVersionUID
	 * 
	 */
	private static final long serialVersionUID = 7390258788220776066L;

	/**
	 * 
	 * Add an example to this database
	 * 
	 * @param example
	 *          the example to add
	 * 
	 */
	public void add(Example<T> example) {
		int count = 0;
		Integer counter = get(example);
		if (counter != null) {
			count = counter;
		}
		put(example, count + 1);
	}

	/**
	 * 
	 * Test if the same example has been added with another class as result
	 * 
	 * @param example
	 *          the example to compare with
	 * @return true if all the example previously added to this database are from
	 *         the same class
	 * 
	 */
	public boolean isConsistant(Example<T> example) {
		Iterator<Example<T>> iterator = keySet().iterator();
		while (iterator.hasNext()) {
			Example<T> previouslyAddedExample = iterator.next();
			if (previouslyAddedExample.values.equals(example.values)
					&& !example.className.equals(previouslyAddedExample.className)) {
				return false;
			}
		}
		return true;
	}

	/**
	 * 
	 * Merge a database to this one
	 * 
	 * @param exampleDatabase
	 *          the database to merge
	 * 
	 */
	public void merge(ExampleDatabase<T> exampleDatabase) {
		for (Map.Entry<Example<T>, Integer> entry : exampleDatabase.entrySet()) {
			for (int index = 0; index < entry.getValue(); index++) {
				add(entry.getKey());
			}
		}
	}

	/**
	 * 
	 * Convert this database to a dot string
	 * 
	 * @return a dot string representing this database
	 * 
	 */
	public String toDotString() {
		StringBuilder result = new StringBuilder();
		for (Map.Entry<Example<T>, Integer> entry : entrySet()) {
			result.append(entry.getKey().toString());
			result.append(" x ");
			result.append(entry.getValue());
			result.append("\\n");
		}
		return result.toString();
	}

	/**
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		StringBuilder result = new StringBuilder();
		for (Map.Entry<Example<T>, Integer> entry : entrySet()) {
			result.append(entry.getKey().toString());
			result.append("x");
			result.append(entry.getValue());
			result.append("\n");
		}
		return result.toString();
	}
}
