package animatlab.kodamat.supervisedlearning.core;

import java.util.List;

import animatlab.kodamat.core.functions.IPartition;
import animatlab.kodamat.core.functions.IPartitionFunction;
import animatlab.kodamat.core.io.ILog;
import animatlab.kodamat.core.io.Log;
import animatlab.kodamat.core.trees.Tree;
import animatlab.kodamat.core.utils.Parameters;
import animatlab.kodamat.core.utils.Context;
import animatlab.kodamat.core.utils.VariableOrdering;
import animatlab.kodamat.supervisedlearning.entropy.VariableMeasure;
import animatlab.kodamat.supervisedlearning.learningtree.ID3Tree;

/**
 * 
 * Convert a partionned function into a tree
 * 
 * @author Thomas Degris
 * @param <T>
 *          output of the function
 * 
 */
public class PartitionsToTree<T> implements ILog {
	/**
	 * 
	 * Measure to use to reorganize the tree
	 * 
	 */
	final private VariableMeasure<T> variableMeasure;

	/**
	 * 
	 * Log used to print reordering results
	 * 
	 */
	private Log log = Log.sysout;

	/**
	 * 
	 * Constructor
	 * 
	 * @param variableMeasure
	 *          measure to use to select the best test
	 * 
	 */
	public PartitionsToTree(VariableMeasure<T> variableMeasure) {
		this.variableMeasure = variableMeasure;
	}

	/**
	 * 
	 * Build a tree from a partionned function
	 * 
	 * @param function
	 *          the function to learn
	 * @return the tree representing the function
	 * 
	 */
	public Tree<T> build(IPartitionFunction<T> function) {
		VariableOrdering variableOrdering = VariableOrdering.buildOrdering(function
				.scope());
		ID3Tree.Plain<T> learningTree = new ID3Tree.Plain<T>(variableMeasure,
				variableOrdering);
		for (IPartition<T> partition : function.partitions()) {
			List<String> state = Context.contextToList(variableOrdering, partition
					.getContext());
			Example<T> example = new Example<T>(partition.get(), state);
			learningTree.learn(example);
		}
		learningTree.flush();
		log.out().print(
				"o["
						+ Parameters.format((double) learningTree.numberOfNodes()
								/ function.numberOfNodes()) + "]");
		return learningTree;
	}

	/**
	 * 
	 * @return the measure used to reorganize the tree
	 * 
	 */
	public VariableMeasure<T> getMeasure() {
		return variableMeasure;
	}

	/**
	 * @see animatlab.kodamat.core.io.ILog#setLog(animatlab.kodamat.core.io.Log)
	 */
	public void setLog(Log log) {
		this.log = log;
	}
}
