package animatlab.kodamat.supervisedlearning.entropy;

import static animatlab.kodamat.core.proba.Chi2Utils.chi2Utils;

import java.util.Set;

import animatlab.kodamat.core.utils.Prototype;
import animatlab.kodamat.core.utils.Variable;
import animatlab.kodamat.supervisedlearning.core.ExampleDatabase;

/**
 * 
 * Variable measure using Chi2 tests to compute the measure
 * 
 * @author Thomas Degris
 * 
 */
public class Chi2Measure extends VariableMeasure<String> {
	/**
	 * 
	 * Ki square threshold to determin if H0 is rejected or not
	 * 
	 */
	static public double DEFAULT_KI_SQUARE_VALUE = 30.0;

	/**
	 * 
	 * Difference between removing and selecting a test
	 * 
	 */
	static public double DEFAULT_KI_SQUARE_DIFF = 0.0;

	/**
	 * 
	 * Constructor
	 * 
	 */
	public Chi2Measure() {
		this(DEFAULT_KI_SQUARE_VALUE);
	}

	/**
	 * 
	 * Constructor
	 * 
	 * @param threshold
	 *          the threshold to set
	 * 
	 */
	public Chi2Measure(Double threshold) {
		super(threshold);
	}

	/**
	 * @see animatlab.kodamat.supervisedlearning.entropy.VariableMeasure#computeMeasure(animatlab.kodamat.core.utils.Variable,
	 *      animatlab.kodamat.supervisedlearning.entropy.VariableValueStatistics,
	 *      ExampleDatabase, Set)
	 */
	@Override
	protected double computeMeasure(Variable variable,
			VariableValueStatistics<String> variableValueStatistics,
			ExampleDatabase<String> exampleDatabase, Set<Variable> ignoredVariables) {
		return chi2Utils.computeChi2(variableValueStatistics.getStats(variable),
				variableValueStatistics);
	}

	/**
	 * @see animatlab.kodamat.core.utils.Prototype#newInstance()
	 */
	public Prototype newInstance() {
		Chi2Measure result = new Chi2Measure();
		result.set(this);
		return result;
	}

	/**
	 * 
	 * Set the theshold used for stochastic problem
	 * 
	 * @param chiSquareValue
	 *          The kiSquareValue to set.
	 * 
	 */
	public void setThreshold(double chiSquareValue) {
		this.threshold = chiSquareValue;
	}

	/**
	 * @see animatlab.kodamat.supervisedlearning.entropy.VariableMeasure#isSignificant(animatlab.kodamat.supervisedlearning.entropy.VariableMeasure.VariableMeasureData,
	 *      animatlab.kodamat.supervisedlearning.entropy.VariableValueStatistics,
	 *      animatlab.kodamat.supervisedlearning.core.ExampleDatabase,
	 *      java.util.Set, animatlab.kodamat.core.utils.Variable)
	 */
	@Override
	protected boolean isSignificant(VariableMeasureData bestTest,
			VariableValueStatistics<String> variableValueStatistics,
			ExampleDatabase<String> exampleDatabase, Set<Variable> ignoredVariables,
			Variable currentTest) {
		assert (!Double.isNaN(bestTest.measure));
		if (bestTest.variable.equals(currentTest)) {
			return (bestTest.measure > threshold - DEFAULT_KI_SQUARE_DIFF);
		}
		return (bestTest.measure > threshold);
	}

}
