package animatlab.kodamat.supervisedlearning.learningtree;

import animatlab.kodamat.core.trees.DNode;
import animatlab.kodamat.core.trees.Leaf;
import animatlab.kodamat.core.trees.Vertex;
import animatlab.kodamat.core.utils.VariableOrdering;
import animatlab.kodamat.supervisedlearning.core.Example;
import animatlab.kodamat.supervisedlearning.core.ExampleDatabase;
import animatlab.kodamat.supervisedlearning.entropy.VariableMeasure;
import animatlab.kodamat.supervisedlearning.entropy.VariableValueStatistics;

/**
 * 
 * Tree used to clone a learning tree
 * 
 * @author Thomas Degris
 * @param <T>
 *          type of the examples to learn
 * @param <U>
 *          type contained at the leaves of the tree
 * 
 */
public class ClonedTree<T, U> extends LearningTree<T, U> {
	/**
	 * 
	 * Example database associated with this tree
	 * 
	 */
	private ExampleDatabase<T> exampleDatabase = null;

	/**
	 * 
	 * The prototype to use as leaf
	 * 
	 */
	private Leaf<U> leafPrototype = null;

	/**
	 * 
	 * Constructor
	 * 
	 * @param variableMeasure
	 *          measure to use to select the best test
	 * @param variableOrdering
	 *          variable ordering to use with the examples
	 * 
	 */
	public ClonedTree(VariableMeasure<T> variableMeasure,
			VariableOrdering variableOrdering) {
		super(variableMeasure, variableOrdering);
	}

	/**
	 * @see animatlab.kodamat.supervisedlearning.learningtree.LearningTree#learn(animatlab.kodamat.core.trees.Vertex,
	 *      animatlab.kodamat.supervisedlearning.core.Example)
	 */
	@Override
	protected void learn(Vertex<U> node, Example<T> example) {
		assert (false);
	}

	/**
	 * @see animatlab.kodamat.supervisedlearning.learningtree.LearningTree#setDataToLeaf(animatlab.kodamat.core.trees.Leaf,
	 *      animatlab.kodamat.supervisedlearning.entropy.VariableValueStatistics,
	 *      ExampleDatabase)
	 */
	@Override
	protected void setDataToLeaf(Leaf<U> leaf, VariableValueStatistics<T> stats,
			ExampleDatabase<T> exampleDatabase) {
		assert (false);
	}

	/**
	 * @see animatlab.kodamat.supervisedlearning.learningtree.LearningTree#newInstance()
	 */
	@Override
	public LearningTree<T, U> newInstance() {
		assert (false);
		return null;
	}

	/**
	 * 
	 * Set the example database of this tree
	 * 
	 * @param exampleDatabase
	 *          the example database to set
	 * 
	 */
	protected void setExampleDatabase(ExampleDatabase<T> exampleDatabase) {
		this.exampleDatabase = exampleDatabase;
	}

	/**
	 * @see animatlab.kodamat.supervisedlearning.learningtree.LearningTree#getExampleDatabase()
	 */
	@Override
	public ExampleDatabase<T> getExampleDatabase() {
		return exampleDatabase;
	}

	/**
	 * @see animatlab.kodamat.core.trees.Tree#newLeaf(animatlab.kodamat.core.trees.DNode,
	 *      java.lang.String)
	 */
	@Override
	protected Leaf<U> newLeaf(DNode<U> parent, String value) {
		if (leafPrototype == null) {
			return super.newLeaf(parent, value);
		}
		return leafPrototype.newInstance(this, parent, value);
	}

	// /**
	// * @see animatlab.kodamat.core.trees.Tree#newClonedTree()
	// **/
	// @Override
	// protected LearningTree<T, U> newClonedTree() {
	// ClonedTree<T, U> result = new ClonedTree<T, U>(variableMeasure, legend);
	// result.leafPrototype = leafPrototype;
	// return result;
	// }
	//
	/**
	 * 
	 * Set the leaf to use as prototype
	 * 
	 * @param prototype
	 *          the new prototype
	 * 
	 */
	public void setLeafPrototype(Leaf<U> prototype) {
		leafPrototype = prototype;
	}
}
