package animatlab.kodamat.supervisedlearning.learningtree;

import animatlab.kodamat.core.trees.DNode;
import animatlab.kodamat.core.trees.Tree;
import animatlab.kodamat.core.utils.Variable;
import animatlab.kodamat.supervisedlearning.core.Example;
import animatlab.kodamat.supervisedlearning.core.ExampleDatabase;

/**
 * 
 * Learning decision node with an example database
 * 
 * @param <T>
 *          type of the examples to learn
 * @param <U>
 *          type contained at the leaves of the tree
 * 
 * @author Thomas Degris
 * 
 */
public class DBLearningDNode<T, U> extends LearningDNode<T, U> {
	/**
	 * 
	 * Database of the examples learned
	 * 
	 */
	private ExampleDatabase<T> exampleDatabase = null;

	/**
	 * 
	 * Constructor
	 * 
	 * @param tree
	 *          the tree owning this decision node
	 * @param value
	 *          value of the tested variable
	 * @param parent
	 *          Parent in the tree of this node
	 * @param testedVariable
	 *          tested variable at this decision node
	 * 
	 */
	public DBLearningDNode(Tree<U> tree, DNode<U> parent, String value,
			Variable testedVariable) {
		super(tree, parent, value, testedVariable);
	}

	/**
	 * @see animatlab.kodamat.supervisedlearning.learningtree.ILearningVertex#getExampleDatabase()
	 */
	@Override
	public ExampleDatabase<T> getExampleDatabase() {
		if (exampleDatabase == null) {
			exampleDatabase = new ExampleDatabase<T>();
		}
		return exampleDatabase;
	}

	/**
	 * @see animatlab.kodamat.supervisedlearning.learningtree.ILearningVertex#setData(animatlab.kodamat.supervisedlearning.learningtree.ILearningVertex)
	 */
	@Override
	public void setData(ILearningVertex<T> learningVertex) {
		super.setData(learningVertex);
		assert (exampleDatabase == null);
		exampleDatabase = learningVertex.getExampleDatabase();
	}

	/**
	 * @see animatlab.kodamat.supervisedlearning.learningtree.ILearningVertex#addExample(animatlab.kodamat.supervisedlearning.core.Example)
	 */
	@Override
	public void addExample(Example<T> example) {
		super.addExample(example);
		getExampleDatabase().add(example);
	}
}
