package animatlab.kodamat.supervisedlearning.learningtree;

import animatlab.kodamat.core.proba.IProbabilities;
import animatlab.kodamat.core.trees.DNode;
import animatlab.kodamat.core.trees.EqualPolicy;
import animatlab.kodamat.core.trees.MergeLeavesPolicy;
import animatlab.kodamat.core.trees.Tree;
import animatlab.kodamat.core.utils.Variable;
import animatlab.kodamat.supervisedlearning.core.Example;

/**
 * 
 * ID5 decision node
 * 
 * @author Thomas Degris
 * 
 */
public class ID5DNode extends LearningDNode<String, IProbabilities<String>> {
	/**
	 * 
	 * Stale flag used to mark when a test has been checked for this node
	 * 
	 */
	protected boolean fresh = false;

	/**
	 * 
	 * Constructor
	 * 
	 * @param tree
	 *          the tree owning this decision node
	 * @param value
	 *          value of the tested variable
	 * @param parent
	 *          Parent in the tree of this node
	 * @param testedVariable
	 *          tested variable at this decision node
	 * 
	 */
	public ID5DNode(Tree<IProbabilities<String>> tree,
			DNode<IProbabilities<String>> parent, String value,
			Variable testedVariable) {
		super(tree, parent, value, testedVariable);
	}

	/**
	 * 
	 * @return Returns the fresh.
	 * 
	 */
	public final boolean isFresh() {
		return fresh;
	}

	/**
	 * 
	 * Set fresh
	 * 
	 * @param fresh
	 *          The fresh to set.
	 * 
	 */
	public final void setFresh(boolean fresh) {
		this.fresh = fresh;
	}

	/**
	 * @see animatlab.kodamat.core.trees.Vertex#simplify(EqualPolicy,
	 *      MergeLeavesPolicy)
	 */
	@Override
	public void simplify(EqualPolicy<IProbabilities<String>> equalPolicy,
			MergeLeavesPolicy<IProbabilities<String>> mergePolicy) {
		if (fresh) {
			return;
		}
		super.simplify(equalPolicy, mergePolicy);
		if (children().size() == 1) {
			replaceWithChild(children().iterator().next());
		}
		fresh = true;
	}

	/**
	 * @see animatlab.kodamat.core.trees.DNode#localInfoToString()
	 */
	@Override
	protected String localInfoToString() {
		if (fresh) {
			return "";
		}
		return "~";
	}

	/**
	 * 
	 * Add a new example to this vertex
	 * 
	 * @param example
	 *          the example to add
	 * 
	 */
	@Override
	public void addExample(Example<String> example) {
		super.addExample(example);
		fresh = false;
	}
}
