package animatlab.kodamat.supervisedlearning.learningtree;

import animatlab.kodamat.core.utils.Context;
import animatlab.kodamat.supervisedlearning.core.Example;
import animatlab.kodamat.supervisedlearning.core.ExampleDatabase;
import animatlab.kodamat.supervisedlearning.entropy.VariableValueStatistics;

/**
 * 
 * Interface for a node in the tree
 * 
 * @author Thomas Degris
 * @param <T>
 *          the type of the example to learn
 * 
 */
public interface ILearningVertex<T> {
	/**
	 * 
	 * Name of the attribute containing the value of the measure
	 * 
	 */
	public static final String ATTRIBUTE_MEASURE = "meas";

	/**
	 * 
	 * Name of the attribute containing the number of examples
	 * 
	 */
	public static final String NB_EXAMPLES_MEASURE = "nbex";

	/**
	 * 
	 * @return the database contained at this vertex (may be null)
	 * 
	 */
	public ExampleDatabase<T> getExampleDatabase();

	/**
	 * 
	 * @return the stats contained at the node
	 * 
	 */
	public VariableValueStatistics<T> getStats();

	/**
	 * 
	 * Add a new example to this vertex
	 * 
	 * @param example
	 *          the example to add
	 * 
	 */
	public void addExample(Example<T> example);

	/**
	 * 
	 * @return the context of the partition
	 * 
	 */
	public Context getContext();

	/**
	 * 
	 * Set the data of this vertex from another vertex
	 * 
	 * @param learningVertex
	 *          the vertex to get the data from
	 * 
	 */
	public void setData(ILearningVertex<T> learningVertex);

	/**
	 * 
	 * @return the tree owning this vertex
	 * 
	 */
	public LearningTree<T, ?> getTree();

	/**
	 * 
	 * Set the statistic for this decision node
	 * 
	 * @param stats
	 *          the stats to set
	 * 
	 */
	public void setStats(VariableValueStatistics<T> stats);
}
