package animatlab.kodamat.supervisedlearning.learningtree;

import animatlab.kodamat.core.proba.IProbabilities;
import animatlab.kodamat.core.trees.DNode;
import animatlab.kodamat.core.trees.Tree;
import animatlab.kodamat.core.trees.Vertex;
import animatlab.kodamat.core.utils.Variable;
import animatlab.kodamat.supervisedlearning.core.Example;

/**
 * 
 * Decision node for the ITI algorithm
 * 
 * @author Thomas Degris
 * 
 */
public class ItiDNode extends DBLearningDNode<String, IProbabilities<String>> {
	/**
	 * 
	 * Stale flag used to mark when a test has been checked for this node
	 * 
	 */
	protected boolean fresh = false;

	/**
	 * 
	 * Value corresponding to the left child
	 * 
	 */
	protected String selectedValue;

	/**
	 * 
	 * Constructor
	 * 
	 * @param tree
	 *          the tree owning this decision node
	 * @param value
	 *          value of the tested variable
	 * @param parent
	 *          Parent in the tree of this node
	 * @param testedVariable
	 *          tested variable at this decision node
	 * 
	 */
	public ItiDNode(Tree<IProbabilities<String>> tree,
			DNode<IProbabilities<String>> parent, String value,
			Variable testedVariable) {
		super(tree, parent, value, testedVariable);
	}

	/**
	 * 
	 * Get the child at the left of this node
	 * 
	 * @return the left child of this node
	 * 
	 */
	protected Vertex<IProbabilities<String>> left() {
		return super.getChild(selectedValue);
	}

	/**
	 * 
	 * Get the child at the right of this node
	 * 
	 * @return the right child of this node
	 * 
	 */
	protected Vertex<IProbabilities<String>> right() {
		return super.getChild(VALUE_WILDCARD);
	}

	/**
	 * 
	 * @return Returns the fresh.
	 * 
	 */
	public final boolean isFresh() {
		return fresh;
	}

	/**
	 * 
	 * Set fresh
	 * 
	 * @param fresh
	 *          The fresh to set.
	 * 
	 */
	public final void setFresh(boolean fresh) {
		this.fresh = fresh;
	}

	/**
	 * @see animatlab.kodamat.core.trees.DNode#localInfoToString()
	 */
	@Override
	protected String localInfoToString() {
		if (fresh) {
			return "";
		}
		return "~";
	}

	/**
	 * @see animatlab.kodamat.supervisedlearning.learningtree.ILearningVertex#addExample(animatlab.kodamat.supervisedlearning.core.Example)
	 */
	@Override
	public void addExample(Example<String> example) {
		super.addExample(example);
		fresh = false;
	}

}
