package animatlab.kodamat.supervisedlearning.learningtree;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import animatlab.kodamat.core.trees.DNode;
import animatlab.kodamat.core.trees.Tree;
import animatlab.kodamat.core.utils.Variable;
import animatlab.kodamat.supervisedlearning.core.Example;
import animatlab.kodamat.supervisedlearning.core.ExampleDatabase;
import animatlab.kodamat.supervisedlearning.entropy.VariableValueStatistics;
import animatlab.kodamat.supervisedlearning.entropy.VariableMeasure.VariableMeasureData;

/**
 * 
 * Learning decision node
 * 
 * @author Thomas Degris
 * @param <T>
 *          type of the examples to learn
 * @param <U>
 *          type contained at the leaves of the tree
 * 
 */
public class LearningDNode<T, U> extends DNode<U> implements ILearningVertex<T> {
	/**
	 * 
	 * Stats maintained at this node
	 * 
	 */
	private VariableValueStatistics<T> stats = null;

	/**
	 * 
	 * Constructor
	 * 
	 * @param tree
	 *          the tree owning this decision node
	 * @param value
	 *          value of the tested variable
	 * @param parent
	 *          Parent in the tree of this node
	 * @param testedVariable
	 *          tested variable at this decision node
	 * 
	 */
	public LearningDNode(Tree<U> tree, DNode<U> parent, String value,
			Variable testedVariable) {
		super(tree, parent, value, testedVariable);
	}

	/**
	 * @see animatlab.kodamat.core.trees.DNode#dataToString()
	 */
	@Override
	protected String dataToString() {
		// if (stats == null) {
		// return "";
		// }
		// return ((LearningTree<T, U>)tree).variableMeasure.toString(stats);
		return "";
	}

	/**
	 * @see animatlab.kodamat.supervisedlearning.learningtree.ILearningVertex#getExampleDatabase()
	 */
	public ExampleDatabase<T> getExampleDatabase() {
		return null;
	}

	/**
	 * @see animatlab.kodamat.supervisedlearning.learningtree.ILearningVertex#getStats()
	 */
	public VariableValueStatistics<T> getStats() {
		if (stats == null) {
			stats = new VariableValueStatistics<T>();
		}
		return stats;
	}

	/**
	 * @see animatlab.kodamat.supervisedlearning.learningtree.ILearningVertex#addExample(animatlab.kodamat.supervisedlearning.core.Example)
	 */
	@SuppressWarnings("unchecked")
	public void addExample(Example<T> example) {
		stats.add(((LearningTree<T, U>) tree).getVariableOrdering(), example);
	}

	/**
	 * @see animatlab.kodamat.supervisedlearning.learningtree.ILearningVertex#setData(animatlab.kodamat.supervisedlearning.learningtree.ILearningVertex)
	 */
	public void setData(ILearningVertex<T> learningVertex) {
		assert ((stats == null) || (stats == learningVertex.getStats()));
		stats = learningVertex.getStats();
	}

	/**
	 * @see animatlab.kodamat.supervisedlearning.learningtree.ILearningVertex#setStats(animatlab.kodamat.supervisedlearning.entropy.VariableValueStatistics)
	 */
	public void setStats(VariableValueStatistics<T> stats) {
		this.stats = stats;
	}

	/**
	 * @see animatlab.kodamat.core.trees.Vertex#getTree()
	 */
	@SuppressWarnings("unchecked")
	@Override
	public LearningTree<T, U> getTree() {
		return (LearningTree<T, U>) super.getTree();
	}

	/**
	 * @see animatlab.kodamat.core.io.IToXml#toXml(org.w3c.dom.Document,
	 *      animatlab.kodamat.core.io.IToXml.IToXmlPolicy)
	 */
	@Override
	public Element toXml(Document document, IToXmlPolicy<U> policy) {
		Element result = super.toXml(document, policy);
		VariableMeasureData measure = getTree().getMeasure().getBestMeasure(null,
				getStats(), getContext().keySet());
		if (measure != null) {
			result.setAttribute(ATTRIBUTE_MEASURE, String.valueOf(measure.measure));
		}
		result.setAttribute(NB_EXAMPLES_MEASURE, String.valueOf(getStats()
				.getNumberOfExamples()));
		return result;
	}
}
