package animatlab.kodamat.supervisedlearning.regression;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import animatlab.kodamat.core.proba.ILearningProbabilities;
import animatlab.kodamat.core.utils.Prototype;
import animatlab.kodamat.core.utils.Variable;
import animatlab.kodamat.supervisedlearning.core.ExampleDatabase;

/**
 * 
 * Variable measure using Chi2 tests to compute the measure
 * 
 * @author Thomas Degris
 * 
 */
public class LeastAbsoluteDeviationMeasure extends RegressionMeasure {
	/**
	 * 
	 * Constructor
	 * 
	 * @param threshold
	 *          the threshold to use
	 * 
	 */
	public LeastAbsoluteDeviationMeasure(Double threshold) {
		super(threshold);
	}

	/**
	 * @see animatlab.kodamat.core.utils.Prototype#newInstance()
	 */
	public Prototype newInstance() {
		LeastAbsoluteDeviationMeasure result = new LeastAbsoluteDeviationMeasure(
				null);
		result.set(this);
		return result;
	}

	/**
	 * @see animatlab.kodamat.supervisedlearning.regression.RegressionMeasure#computeValue(Variable,
	 *      String, ILearningProbabilities, ExampleDatabase)
	 */
	@Override
	protected double computeValue(Variable variable, String variableValue,
			ILearningProbabilities<Double> classNamesCounter,
			ExampleDatabase<Double> exampleDatabase) {
		int numberOfExamples = classNamesCounter.getNumberOfExamples();
		int medianNumberOfExamples = numberOfExamples / 2;
		if ((numberOfExamples % 2) == 1) {
			medianNumberOfExamples++;
		}
		List<Double> sortedValues = new ArrayList<Double>(classNamesCounter
				.getValueSet());
		Collections.sort(sortedValues);
		int currentNumberOfExamples = 0;
		int valueIndex = -1;
		do {
			valueIndex++;
			currentNumberOfExamples += classNamesCounter
					.getNumberOfExamples(sortedValues.get(valueIndex));
		} while (currentNumberOfExamples < medianNumberOfExamples);
		if ((numberOfExamples % 2) == 1) {
			return sortedValues.get(valueIndex);
		}
		if (currentNumberOfExamples == medianNumberOfExamples) {
			return (sortedValues.get(valueIndex + 1) - sortedValues.get(valueIndex)) / 2.0;
		}
		return sortedValues.get(valueIndex);
	}

	/**
	 * @see animatlab.kodamat.supervisedlearning.regression.RegressionMeasure#computeError(Variable,
	 *      String, double, ILearningProbabilities, ExampleDatabase)
	 */
	@Override
	public double computeError(Variable variable, String variableValue,
			double average, ILearningProbabilities<Double> classNamesCounter,
			ExampleDatabase<Double> exampleDatabase) {
		double result = 0;
		int counter = 0;
		for (Double value : classNamesCounter.getValueSet()) {
			int weight = classNamesCounter.getNumberOfExamples(value);
			result = result + weight * Math.abs(value - average);
			counter = counter + weight;
		}
		return result / counter;
	}

}
