/*
 * 
 * Created on 9 janv. 2006
 *
 */
package animatlab.kodamat.supervisedlearning.rmax;

import animatlab.kodamat.core.trees.DNode;
import animatlab.kodamat.core.trees.EqualPolicy;
import animatlab.kodamat.core.trees.MergeLeavesPolicy;
import animatlab.kodamat.core.trees.Vertex;
import animatlab.kodamat.core.utils.Variable;
import animatlab.kodamat.supervisedlearning.learningtree.LearningDNode;

/**
 * 
 * DNode used in RMax tree
 * 
 * @author Thomas Degris
 * 
 */
public class RMaxDNode extends LearningDNode<Double, Double> {
	/**
	 * 
	 * Stale flag used to mark when a test has been checked for this node
	 * 
	 */
	protected boolean fresh = true;

	/**
	 * 
	 * @return Returns the fresh.
	 * 
	 */
	public final boolean isFresh() {
		return fresh;
	}

	/**
	 * 
	 * Set fresh
	 * 
	 * @param fresh
	 *          The fresh to set.
	 * 
	 */
	public final void setFresh(boolean fresh) {
		this.fresh = fresh;
	}

	/**
	 * 
	 * Constructor
	 * 
	 * @param tree
	 *          the tree owning this decision node
	 * @param value
	 *          value of the tested variable
	 * @param parent
	 *          Parent in the tree of this node
	 * @param testedVariable
	 *          tested variable at this decision node
	 * 
	 */
	public RMaxDNode(RMaxTree tree, DNode<Double> parent, String value,
			Variable testedVariable) {
		super(tree, parent, value, testedVariable);
	}

	/**
	 * @see animatlab.kodamat.core.trees.DNode#getChild(java.lang.String)
	 */
	@Override
	public Vertex<Double> getChild(String value) {
		Vertex<Double> result = super.getChild(value);
		if (result == null) {
			result = addLeaf(value);
		}
		return result;
	}

	/**
	 * @see animatlab.kodamat.core.trees.Vertex#simplify(EqualPolicy,
	 *      MergeLeavesPolicy)
	 */
	@Override
	public void simplify(EqualPolicy<Double> equalPolicy,
			MergeLeavesPolicy<Double> mergePolicy) {
		if (fresh) {
			return;
		}
		if (children.size() == 0) {
			return;
		}
		for (Vertex<Double> child : children.values()) {
			child.simplify(equalPolicy, mergePolicy);
		}

		Vertex<Double> last = null;
		for (Vertex<Double> child : children.values()) {
			if ((last != null) && !last.equals(child, equalPolicy)) {
				return;
			}
			last = child;
		}
		Vertex<Double> selected = mergePolicy.merge(children.values());
		if (selected != null) {
			replaceWithChild(selected);
		}
		fresh = true;
	}

}
