/*
 * 
 * Created on Dec 29, 2005
 *
 */
package animatlab.kodamat.supervisedlearning.rmax;

import animatlab.kodamat.core.trees.ClonePolicy;
import animatlab.kodamat.core.trees.DNode;
import animatlab.kodamat.core.trees.Leaf;
import animatlab.kodamat.core.trees.Tree;
import animatlab.kodamat.core.utils.Context;
import animatlab.kodamat.supervisedlearning.core.Example;
import animatlab.kodamat.supervisedlearning.learningtree.LearningLeaf;
import animatlab.kodamat.supervisedlearning.learningtree.LearningTree;

/**
 * 
 * Leaf used in a learning tree
 * 
 * @author Thomas Degris
 * 
 */
public class RMaxLeaf extends LearningLeaf<Double, Double> {
	/**
	 * 
	 * Constructor
	 * 
	 * @param tree
	 *          the tree owning this decision node
	 * @param value
	 *          value of the tested variable
	 * @param parent
	 *          Parent in the tree of this node
	 * 
	 */
	protected RMaxLeaf(RMaxTree tree, DNode<Double> parent, String value) {
		super(tree, parent, value);
	}

	/**
	 * @see animatlab.kodamat.core.io.DataContainer#get()
	 */
	@Override
	public Double get() {
		RMaxTree parent = (RMaxTree) tree;
		if (getStats().getNumberOfExamples() < parent.numberOfTry) {
			return parent.rmax;
		}
		return super.get();
	}

	/**
	 * @see animatlab.kodamat.core.trees.Vertex#toString(int)
	 */
	@Override
	public String toString(int indentation) {
		String parentResult = super.toString(indentation);
		return parentResult.substring(0, parentResult.length() - 1) + "("
				+ getStats().getNumberOfExamples() + ")\n";
	}

	/**
	 * @see animatlab.kodamat.core.trees.Leaf#clone(animatlab.kodamat.core.trees.ClonePolicy,
	 *      animatlab.kodamat.core.trees.Tree, animatlab.kodamat.core.trees.DNode,
	 *      java.lang.String, animatlab.kodamat.core.utils.Context)
	 */
	@Override
	protected Leaf<Double> clone(ClonePolicy<Double> clonePolicy,
			Tree<Double> result, DNode<Double> parent, String value, Context testsList) {
		Leaf<Double> newLeaf = super.clone(clonePolicy, result, parent, value,
				testsList);
		newLeaf.set(get());
		return newLeaf;
	}

	/**
	 * 
	 * Merge to leaf together
	 * 
	 * @param leaf
	 *          the leaf to merge with this one
	 * 
	 */
	public void merge(RMaxLeaf leaf) {
		getExampleDatabase().merge(leaf.getExampleDatabase());
		getStats().merge(leaf.getStats());
	}

	/**
	 * 
	 * Add a new example to this vertex
	 * 
	 * @param example
	 *          the example to add
	 * 
	 */
	@SuppressWarnings("unchecked")
	@Override
	public void addExample(Example<Double> example) {
		super.addExample(example);
		set(((LearningTree<Double, Double>) tree).getMeasure().aggregate(
				getStats(), getExampleDatabase(), getContext().keySet()));
	}
}
