/*
 * 
 * Created on Jan 17, 2006
 *
 */
package animatlab.kodamat.supervisedlearning.tabular;

import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import animatlab.kodamat.core.functions.IFunction;
import animatlab.kodamat.core.proba.ILearningProbabilities;
import animatlab.kodamat.core.proba.IProbabilities;
import animatlab.kodamat.core.proba.ValueStatistics;
import animatlab.kodamat.core.utils.Context;
import animatlab.kodamat.core.utils.VariableOrdering;
import animatlab.kodamat.supervisedlearning.core.Example;
import animatlab.kodamat.supervisedlearning.core.ISupervisedLearning;

/**
 * 
 * Tabular supervised learning
 * 
 * @param <T>
 *          type of things to learn
 * 
 * @author Thomas Degris
 * 
 */
public class TabularLearning<T> implements IFunction<IProbabilities<T>>,
		ISupervisedLearning<T> {
	/**
	 * 
	 * Legend used to interprete examples
	 * 
	 */
	final private VariableOrdering variableOrdering;

	/**
	 * 
	 * Tabular registering examples
	 * 
	 */
	final private Map<List<String>, ValueStatistics<T>> tabular = new LinkedHashMap<List<String>, ValueStatistics<T>>();

	/**
	 * 
	 * Set of possible value
	 * 
	 */
	final private Set<T> valueSet = new LinkedHashSet<T>();

	/**
	 * 
	 * Constructor
	 * 
	 * @param variableOrdering
	 *          legend to interprete examples
	 * 
	 */
	public TabularLearning(VariableOrdering variableOrdering) {
		this.variableOrdering = variableOrdering;
	}

	/**
	 * @see animatlab.kodamat.supervisedlearning.core.ISupervisedLearning#learn(animatlab.kodamat.supervisedlearning.core.Example)
	 */
	public void learn(Example<T> example) {
		ValueStatistics<T> probabilities = tabular.get(example.values);
		if (probabilities == null) {
			probabilities = new ValueStatistics<T>();
			tabular.put(example.values, probabilities);
		}
		probabilities.add(example.className);
	}

	/**
	 * @see java.lang.Object#clone()
	 */
	@Override
	public TabularLearning<T> clone() {
		return new TabularLearning<T>(variableOrdering);
	}

	/**
	 * 
	 * @return the possible value predicted/classified by this structure
	 * 
	 */
	public Set<T> getValueSet() {
		return valueSet;
	}

	/**
	 * @see animatlab.kodamat.core.functions.IFunction#classify(Context)
	 */
	public IProbabilities<T> classify(Context state) {
		return tabular.get(variableOrdering.stateToList(state));
	}

	/**
	 * 
	 * @return the list of values of example
	 * 
	 */
	public Set<List<String>> getEntrySet() {
		return tabular.keySet();
	}

	/**
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		return tabular.toString();
	}

	/**
	 * @see animatlab.kodamat.core.functions.IFunction#numberOfNodes()
	 */
	public int numberOfNodes() {
		return tabular.size();
	}

	/**
	 * @see animatlab.kodamat.supervisedlearning.core.ISupervisedLearning#getStats()
	 */
	public ILearningProbabilities<T> getStats() {
		assert (false);
		return null;
	}

	/**
	 * @see animatlab.kodamat.supervisedlearning.core.ISupervisedLearning#flush()
	 */
	public void flush() {
		// Nothing to do in incremental learning algorithm
	}

}
