#!/usr/bin/python

import sys
import time

###################################
#
# Variable legend build method
#
###################################

def getVariableSet(problemDefinition):
    separator = "variables ("
    index = problemDefinition.find(separator)
    return problemDefinition[index + len(separator):index + problemDefinition.find(")", index)].split()

def getVariableIndex(problemDefinition, variable):
    """Return the index of a variable in the problem definition."""
    return getVariableSet(problemDefinition).index(variable)


def getVariableLegend(problemDefinition):
    """Parse problemDefinition to return the legend for the variable in xml.
"""
    print "Extracting legend..."
    separator = "variables ("
    index = problemDefinition.find(separator)
    legendSet = problemDefinition[index + len(separator):index + problemDefinition.find(")", index)].split()
    variableLegendSet = ""
    for variable in legendSet:
        variableLegendSet = variableLegendSet + "    <dtr:variable index='%i' legend='%s'/>\n" \
                            % (getVariableIndex(problemDefinition, variable), variable)
    result = """
  <dtr:legend>
%(variableLegendSet)s
  </dtr:legend>
""" % locals()
    return result

###################################
#
# Tree build method
#
###################################

def getVariableValueString(variableValue):
    if variableValue != None:
        return "value='%s' " % variableValue
    else:
        return ""
    
def findFirst(startIndex, string, item01, item02):
    firstItemIndex = string.find(item01, startIndex)
    if (firstItemIndex == -1):
        firstItemIndex = len(string)
    secondItemIndex = string.find(item02, startIndex)
    if (secondItemIndex == -1):
        secondItemIndex = len(string)
    return min(firstItemIndex, secondItemIndex)

def getRecursiveTree(treeDefinition, deepness, variableValue, leafPolicy):
    if (treeDefinition[0] != '('):
        print "Error: the first char of '" + treeDefinition + "' should be '('"; 
    indentation = "  " * deepness
    if treeDefinition[1].isalpha():   
        variableName = treeDefinition[1:findFirst(1, treeDefinition, "\t", " ")]
        leftTree = treeDefinition[treeDefinition.find("(", 1):]
        subtree = getRecursiveTree(leftTree, deepness + 1, 1, leafPolicy)
        bracketBalance = 0
        currentBracketIndex = 0
        while ((currentBracketIndex == 0) or (bracketBalance != 0)):
            currentBracketIndex = findFirst(currentBracketIndex + 1, treeDefinition, "(", ")")
            if (treeDefinition[currentBracketIndex] == ")"):
                bracketBalance = bracketBalance - 1
            else:
                bracketBalance = bracketBalance + 1
        rightTree = treeDefinition[treeDefinition.find("(", currentBracketIndex + 1):]
        subtree = subtree + "\n" + getRecursiveTree(rightTree, deepness + 1, 0, leafPolicy)
        return """%s<dtr:dnode %s variable='%s'>
%s
%s</dtr:dnode>""" % (indentation, getVariableValueString(variableValue), variableName, subtree, indentation)
    else: 
        leafValue = treeDefinition[1:treeDefinition.find(")", 1)]
        return leafPolicy(indentation, variableValue, leafValue)

###################################
#
# The reward tree build method
#
###################################

def leafRewardPolicy(indentation, variableValue, value):
    return "%s<dtr:leaf %s data='%s'/>" % (indentation, getVariableValueString(variableValue), value)

def getRewardTree(problemDefinition):
    """Parse problemDefinition to return the legend for the tree for the reward in xml.
"""
    print "Extracting reward..."
    separator = "reward "
    index = problemDefinition.find(separator) + len(separator)
    rewardDefinition = problemDefinition[index:problemDefinition.find("discount", index) - 1]
    result = """
  <fst:reward>
%s
  </fst:reward>
""" % getRecursiveTree(rewardDefinition, 2, None, leafRewardPolicy)
    return result

###################################
#
# The cpt tree build method
#
###################################

def leafProbaPolicy(indentation, variableValue, value):
    probabilityForValue01 = float(value)
    if probabilityForValue01 > 0:
        prob01 = "\n%(indentation)s  <spi:probability value='1' probability='%(probabilityForValue01)f'/>" % locals()
    else:
        prob01 = ""
    probabilityForValue00 = 1 - probabilityForValue01
    if probabilityForValue00 > 0:
        prob00 = "\n%(indentation)s  <spi:probability value='0' probability='%(probabilityForValue00)f'/>" % locals()
    else:
        prob00 = ""
    return "%s<dtr:leaf %s>%s%s\n%s</dtr:leaf>" % (indentation, getVariableValueString(variableValue), prob01, prob00, indentation)


def getCptTreeForOneAction(problemDefinition, actionDefinition):
    """Parse actionDefinition to return the legend for the cpt tree for one action in xml.
"""
    variableSet = getVariableSet(problemDefinition)
    result = ""
    for variableName in variableSet:
        variableSeparator= "\t%s" % variableName
        print "Variable: " + variableName
        variableDefinition = actionDefinition[actionDefinition.find(variableSeparator) + len(variableSeparator) + 1:]
        variableDefinition = variableDefinition[variableDefinition.find("("):]
        result = result + """      <fst:variable variable='%s'>
%s        
      </fst:variable>
""" % (variableName, getRecursiveTree(variableDefinition, 4, None, leafProbaPolicy))
    return result


def getCptTree(problemDefinition):
    """Parse problemDefinition to return the legend for the cpt tree for all the actions in xml.
"""
    separator = "action "
    index = problemDefinition.find(separator)
    result = """
  <fst:transitions>
"""        
    while index != -1:
        actionStringLength = problemDefinition.find("\n", index + len(separator)) - (index + len(separator))
        action = problemDefinition[index + len(separator):index + len(separator) + actionStringLength]        
        print "Extracting action " + action
        result = result + """    <fst:action name='%s'>
%s        
    </fst:action>        
""" % (action, getCptTreeForOneAction(problemDefinition, problemDefinition[index:problemDefinition.find("endaction", index)]))
        index = problemDefinition.find("action ", index + 1)
    result = result + """
  </fst:transitions>
"""
    return result


sys.setrecursionlimit(40)
fileHandle = open(sys.argv[1], 'r')

extensionPosition = sys.argv[1].rindex(".")
pathPosition = sys.argv[1].rindex("/")
outputFilePath = sys.argv[2] + "/" + sys.argv[1][pathPosition + 1:extensionPosition] + ".forest"
print "Saving in "+ outputFilePath + "..."

outputHandle = open(outputFilePath, 'w')

problemDefinition = fileHandle.read()

variableLegend = getVariableLegend(problemDefinition)
rewardTree = getRewardTree(problemDefinition)
cptTree = getCptTree(problemDefinition)


outputHandle.write("""<fst:forest xmlns:fst="http://animatlab.lip6.fr/spiti" xmlns:cnd="http://animatlab.lip6.fr/spiti" xmlns:dtr="http://animatlab.lip6.fr/spiti" xmlns:spi="http://animatlab.lip6.fr/spiti">
%(variableLegend)s
%(rewardTree)s
%(cptTree)s
</fst:forest>
""" % globals())

outputHandle.close()


