//( incmdp.cpp

#include <iostream>
#include <fstream>

#include "incmdp.hpp"

#define VALUE_FUNCTION 1000
#define POLICY 1001

//( IncMDP

IncMDP::IncMDP(char *infile, double badd, int maxIte) :
  MDP(infile, badd)
{
  this->maxIte = maxIte;
}


IncMDP::~IncMDP()
{
}

void IncMDP::writePolicyToFile(char *outputfile, bool dotFlag, char *spitifile) {
  MDP::writePolicyToFile(outputfile, dotFlag);
  if (spitifile == 0) {
    return;
  }
  std::ofstream out(spitifile);
  out << "<cud:spuddspiti>" << std::endl;
  if (OptimalValue != 0) {
    nodes.clear();
    out << "<cud:valfunsize size='" << getSize(OptimalValue) << "'/>" << std::endl;
  } else {
    nodes.clear();
    out << "<cud:valfunsize size='" << getSize(ApproximateValue) << "'/>" << std::endl;
  }
//   if (OptimalValue != 0) {
//     outputData(OptimalValue, "valuefunction", 0, 0, VALUE_FUNCTION, out);
//   } else {
//     outputData(ApproximateValue, "valuefunction", 0, 0, VALUE_FUNCTION, out);
//   }
  if (OptimalPolicy != 0) {
    outputData(OptimalPolicy, "policy", 0, 0, POLICY, out);
  } else {
    outputData(ApproximatePolicy, "policy", 0, 0, POLICY, out);
  }
  out << "</cud:spuddspiti>" << std::endl;
  out.flush();
  out.close();
}

int IncMDP::getSize(DdNode* node)
{
  assert(node != 0);
  DDNodeList::iterator ite = find(nodes.begin(), nodes.end(), node);
  if (ite != nodes.end()) {
    return 0;
  }
  nodes.push_back(node);
  if (Cudd_IsConstant(node)) {
    return 1;
  } else {
    return 1 + getSize(cuddT(node)) + getSize(cuddE(node));
  }
}


void IncMDP::outputData(DdNode* rootNode, char* info, char* attribute, char* attribute_name, int output, ostream &out)
{
  nodes.clear();

  if(attribute != NULL) {
    out << "<cud:" << info << " "<< attribute_name << "='" << attribute << "'>" << std::endl;
  }
  else {
    out << "<cud:" << info << ">" << std::endl;
  }
  out << "<cud:add>" << std::endl;
  printNode(rootNode, output, out);
  printEdge(rootNode, out);
  out << "</cud:add>" << std::endl;
  out << "</cud:" << info << ">" << std::endl;

}

void IncMDP::printNode(DdNode* node, int output, ostream &out)
{
  int rovar = 0;
  char namestr[1024];

  assert(node != 0);
  DDNodeList::iterator ite = find(nodes.begin(), nodes.end(), node);
  if (ite != nodes.end()) {
    return;
  }
  nodes.push_back(node);
  if (Cudd_IsConstant(node)) {
    if(output == VALUE_FUNCTION) {
      out << "<cud:leaf id='" << node << "' data='" << cuddV(node)->get_val() << "'/>" << std::endl;
    }
    else {
      strcpy(namestr,"");
      aconvert(namestr,actionnames,(*Cudd_V(node)).get_min());
      const std::string actionsPrefix = "act";
      const std::string insideActionsPrefix = "\\n" + actionsPrefix;
      std::string actions = namestr;
      if (actions.find(actionsPrefix) == 0) {
        actions = actions.substr(3);
      }
      int index = actions.find(insideActionsPrefix);
      while (index >= 0) {
        actions = actions.replace(index, insideActionsPrefix.length(), ":");
        index = actions.find(insideActionsPrefix);
      }
      out << "<cud:leaf id='" << node << "' data='" << actions << "'/>" << std::endl;
    }
  } else {
    while (Cudd_NodeReadIndex(vars[orig_vars[rovar].var1index].add_var) < Cudd_NodeReadIndex(node) &&
	   Cudd_NodeReadIndex(vars[orig_vars[rovar].var1index+orig_vars[rovar].nbvars-1].add_var) < Cudd_NodeReadIndex(node)) {
      rovar++;
    }
    std::string variableName = orig_vars[rovar].name;
    if (variableName.find("var") == 0) {
      variableName = variableName.substr(3);
    }
    out << "<cud:decision id='" << node << "' variable='" << variableName << "'/>" << std::endl;
    printNode(cuddT(node), output, out);
    printNode(cuddE(node), output, out);
  }
}


void IncMDP::printEdge(DdNode* node, ostream &out)
{
  assert(node != 0);
  if (Cudd_IsConstant(node)) {
  } else {
    out << "<cud:edge from='" << node << "' to='" << cuddT(node) << "' side='right'/>" << std::endl;
    out << "<cud:edge from='" << node << "' to='" << cuddE(node) << "' side='left'/>" << std::endl;
    printEdge(cuddT(node), out);
    printEdge(cuddE(node), out);
  }
}

int IncMDP::Spudd(DdNode *Reward)
{
  int test;
  int i,j,k,tempo,tempn,tempfoo,index,nodeCount;
  double temps0,temps1;
  long int usememory=0;
  DdNode *actionDD, *MidPoint_Add;
  DdNode *temp1,*temp2,*tempround;
  DdNode **Vcurrent,*Vpast;
  //DdNode *Reward;

  /* Structure to get the time of execution*/
  struct tms _t;
  long clk_tck = sysconf(_SC_CLK_TCK);

  FILE *value,*action,*both,*stats,*log,*log2,*reward,*actions;

  Vcurrent = (DdNode **)malloc(numactions*(sizeof(DdNode*)));

  for(i=0;i<numactions;i++) {
    Vcurrent[i] = RewardD;
    Cudd_Ref(Vcurrent[i]);
  }

  // Convergence test
  test = 0;
  //Number of iteration to convergence
  counter = 0;

  Vpast = RewardD;
  Cudd_Ref(Vpast);

  if (Reward == NULL) {
    Reward = RewardD;
    Cudd_Ref(Reward);
  }
  Merged_Add = Reward;
  Cudd_Ref(Merged_Add);


  actionDD = Zero;	/* Stupid ref to maintain loop invariant */
  Cudd_Ref(actionDD);

  MidPoint_Add = VerySmall;
  Cudd_Ref(MidPoint_Add);

  usememory = Cudd_ReadMemoryInUse(gbm);
  if (maxusememory < usememory)
    maxusememory = usememory;

  int lastiteration=0;
  if (horizon == 1)
    lastiteration = 1;
  /* We get the starting up time */
  times(&_t);
  temps0 = 1.0*_t.tms_utime/clk_tck;

  // swap action costs to primed variables
#ifdef COSTSPRIMED
  for(i=0;i<numactions;i++) {
    temp1 = Cudd_addSwapVariables(gbm,actionCost[i],Array1,Array2,numvars);
    Cudd_Ref(temp1);
    Cudd_RecursiveDeref(gbm,actionCost[i]);
    actionCost[i]= temp1;
  }
#endif
  /* VALUE ITERATION LOOP COMMENCES */
  // while ((maxIte == -1 && (test==0 || lastiteration == 1)) || (maxIte >= 0 && counter < maxIte)){
  while (test==0 || lastiteration == 1) {

    /* KEEP TRACK OF THE NUMBER OF ITERATIONS */
    counter = counter + 1;
    fprintf(stderr,"\niteration %d ",counter);
    if (counter == maxIte) {
      lastiteration = 0;
      test = 1;
    }

    /* update past Value function for convergence check */
    if (maxIte == -1) {
      Cudd_RecursiveDeref(gbm,Vpast);
      Vpast = Merged_Add;
      Cudd_Ref(Vpast);
    }

    // print out stats
    //printStats(One,Merged_Add,stderr);
    /* Reward is Merged_Add with variables primed */
    Cudd_RecursiveDeref(gbm,Reward);
    Reward = Cudd_addSwapVariables(gbm,Merged_Add,Array1,Array2,numvars);
    Cudd_Ref(Reward);

    Cudd_RecursiveDeref(gbm,actionDD);
    actionDD = Zero;
    Cudd_Ref(actionDD);

    /* Merged_Add maintains the current Value function*/
    Cudd_RecursiveDeref(gbm,Merged_Add);
    Merged_Add = VerySmall;
    Cudd_Ref(Merged_Add);

    /*Loop over the states */
    for(i=0;i<numactions;i++) {

      /* do the sum_t(Pr(s,a,t)*V(t)) */
#ifdef OPTIMIZED
      temp1 = primeReward(Reward,&(Allprime[i][0]),prime_vars,0,&(allPlist[i][0]),numvars,i);
#else
      temp1 = multiplySumSet(Reward,NewPrime[i]);
#endif
      //fprintf(stderr,"iteration %d action %d\n",counter,i);
      //Cudd_PrintDebug(gbm,temp1,4,100);
      /* multiply by the discount factor */
      temp2 = Cudd_addApply(gbm,Cudd_addTimes,temp1,discount);
      Cudd_Ref(temp2);

      /* add the reward to get the Vcurrent for this action*/
      Cudd_RecursiveDeref(gbm,temp1);
      temp1 = Cudd_addApply(gbm,Cudd_addPlus,temp2,RewardD);
      Cudd_Ref(temp1);

      Cudd_RecursiveDeref(gbm,temp2);

      // add the action costs.
      temp2 = Cudd_addApply(gbm,Cudd_addPlus,temp1,actionCost[i]);
      Cudd_Ref(temp2);

      Cudd_RecursiveDeref(gbm,temp1);

#ifdef DEBUGP
      FILE *fileh = fopen("shit.dot","w");
      DumpDot_p(gbm, temp2, vars, prime_vars, numvars, orig_vars, numorigvars, NULL, fileh);
      fclose(fileh);
#endif

      /* maximize over actions */
      if (approxMeth == APPROX_NONE) {
	temp1=Cudd_addApply(gbm,Cudd_addMaximum,Merged_Add,temp2);
      } else {
	temp1=Cudd_addApply(gbm,My_addMaximum,Merged_Add,temp2);
      }
      Cudd_Ref(temp1);

#ifdef DEBUGP
      FILE *ffileh = fopen("shit.dot","w");
      DumpDot_p(gbm, temp1, vars, prime_vars, numvars, orig_vars, numorigvars, NULL, ffileh);
      fclose(ffileh);
#endif

      // get the policy during the final iteration
      if ((lastiteration) || (counter == maxIte)) {
        std::cout << "p";
	if (approxMeth == APPROX_NONE) {
	  // add this action to actionDD at each node where temp2 == temp1
	  // replace it when temp2 > temp1
	  //fprintf(stdout,"calling recover policy at %d\n",i);
	  recoverPolicy(temp2,temp1,Merged_Add,&actionDD,i);
	  // save the Qfunction
	  Qf[i] = temp2;
	  Cudd_Ref(temp2);
	  Cudd_RecursiveDeref(gbm,temp2);
	} else {
	  // get MidPoint ADD
	  // temp3 is the midpoint Q add for this action
	  DdNode *temp3 = Cudd_addApply(gbm, addMean, temp2, One);
	  Cudd_Ref(temp3);
	  Cudd_RecursiveDeref(gbm,temp2);

	  // get new MidPoint_Add as those maxima
	  temp2=Cudd_addApply(gbm,Cudd_addMaximum,MidPoint_Add,temp3);
	  Cudd_Ref(temp2);

	  // add this action to actionDD policy if  temp3 > temp2 (current MidPoint_Add)
	  recoverPolicy(temp3,temp2,MidPoint_Add,&actionDD,i);
	  Cudd_RecursiveDeref(gbm,temp3);
	  Cudd_RecursiveDeref(gbm,MidPoint_Add);
	  MidPoint_Add = temp2;
	}
      } else {
	Cudd_RecursiveDeref(gbm,temp2);
      }

      Cudd_RecursiveDeref(gbm,Merged_Add);
      Merged_Add = temp1;


      usememory = Cudd_ReadMemoryInUse(gbm);
      if (maxusememory < usememory)
	maxusememory = usememory;
      if (usememory > MAXMEMHARD) {
	fprintf(stderr,"Memory required exceeds availabe %ld bytes. \nTry a smaller bigadd limit constant\n",MAXMEMHARD);
	Cudd_Quit(gbm);
	return 0;
      }

    } // end of loop over actions

    //test = Cudd_DagSize(Merged_Add);
    //fprintf(stderr,"size of function is now %d\n",test);
    //Cudd_PrintInfo(gbm,stderr);


    //fprintf(stderr,"memory in use is %ld\n",usememory);

      // attempt a reorder?
    if (reorderApplyMeth ==REORDERAPPLY_ALL || (reorderApplyMeth == REORDERAPPLY_FIRSTFIVE && counter < 5))
      reorder(Merged_Add);

    // approximation takes place here
    // could modify this so approximate takes Merged_Add as an argument - then
    // Merged_Add would not have to be a member variable
    if (!lastiteration)
      approximate();

    /* COMPARE THE TWO MOST RECENT ADDs IN ORDER TO VERIFY THE CONVERGENCE  */
    if (maxIte == -1) {
      if (approxMeth == APPROX_NONE) {
        test = Cudd_EqualSupNorm(gbm,Merged_Add,Vpast,toleranceP,0);
      } else {
        DdNode *testD = Cudd_addApply(gbm,Convergence_Test,Vpast,Merged_Add);
        Cudd_Ref(testD);
        if((*Cudd_V(testD)).get_min() == 1.0)
          test = 1;
        Cudd_RecursiveDeref(gbm,testD);
      }
      if (test>0 || lastiteration ||  (horizon >= 0 && counter == horizon-1))
        lastiteration++;
    }
    std::cout << endl;
  }   /* END OF VALUE ITERATION */

  /* We get the time of completion*/
  times(&_t);
  temps1 = 1.0*_t.tms_utime/clk_tck;

  temps = temps1- temps0;
  usememory = Cudd_ReadMemoryInUse(gbm);
  if (maxusememory < usememory)
    maxusememory = usememory;

  // possibly deref the old optimal policy
  if (approxMeth == APPROX_NONE) {
    if (OptimalPolicy)
      Cudd_RecursiveDeref(gbm,OptimalPolicy);
    if (OptimalValue)
      Cudd_RecursiveDeref(gbm,OptimalValue);

    // set global OptimalPolicy and OptimalValue to new values
    OptimalPolicy = actionDD;
    Cudd_Ref(OptimalPolicy);
    OptimalValue = Merged_Add;
    Cudd_Ref(OptimalValue);
  } else {
    if (ApproximateMidPointValue)
      Cudd_RecursiveDeref(gbm,ApproximateMidPointValue);
    ApproximateMidPointValue = MidPoint_Add;
    Cudd_Ref(ApproximateMidPointValue);

    if (ApproximatePolicy)
      Cudd_RecursiveDeref(gbm,ApproximatePolicy);
    ApproximatePolicy = actionDD;
    Cudd_Ref(ApproximatePolicy);

    if (ApproximateValue)
      Cudd_RecursiveDeref(gbm,ApproximateValue);
    ApproximateValue =  Merged_Add;
    Cudd_Ref(ApproximateValue);
  }
  // success
  return 1;
}


//) IncMDP


//) incmdp.cpp
