//( incmdp.hpp

#ifndef INCMDP_HPP
#define INCMDP_HPP

#include <vector>

#include "MDP.h"


//( IncMDP

/**
 * MDP with an output for the spiti framework
 *
 * @author Thomas Degris
 **/
class IncMDP :
  public MDP
{
public:
  /**
   *
   * Constructs and reads in an ADD
   *
   **/
  IncMDP(char *infile, double badd, int maxIte);

  /**
   * Default destructor
   *
   **/
  ~IncMDP();

public:
  // writes policy and value functdion to Dual Optimal file
  // also writes to dot file if dotFlag = true
  // also writes spiti data file if not null
  void writePolicyToFile(char *, bool dotFlag, char *);

private:
  /**
   *
   * Output on the standard output the value function
   *
   * @param rootNode root of the function to represent
   *
   * @return  void
   *
   **/
  void outputData(DdNode* rootNode, char* info, char* atribute, char* atribute_name, int output, ostream &out);

  /**
   *
   * Print a node on the standard output
   *
   * @param   node  node to print out
   *
   * @return  void
   *
   **/
  void printNode(DdNode* node, int output, ostream &out);

  /**
   *
   * Print the edge in the graph
   *
   * @param   node  the node to print
   *
   * @return  void
   *
   **/
  void printEdge(DdNode* node, ostream &out);

  /**
   *
   * The number of nodes of an ADD
   *
   * @param   node  the root node of the add to count
   *
   * @return  the number of nodes of an ADD
   *
   **/
  int getSize(DdNode* node);

private:
  // double-timing Spudd loop is to replace both
  // optimalSpudd and approximateSpudd in a single
  // super-efficient piece of code
  // optional argument specifies the value function to
  // start with
  virtual int Spudd(DdNode *Reward=NULL);

private:
  /**
   *
   * Type: list to save the nodes we printed out
   *
   **/
  typedef std::vector<DdNode*> DDNodeList;


  /**
   *
   * List to save the nodes we printed out
   *
   **/
  DDNodeList nodes;

  /**
   *
   * Maximum number of iterations
   *
   **/
  int maxIte;

}
;

//) IncMDP


#endif

//) incmdp.hpp
